#include "Currents.h"
#include "RndGenerator.h"
#include "Utils.h"

Currents::Vector	Currents::g_currents;

Currents::Currents(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const float strength,const float radius) : Object3d("current",NULL)
{
	Material* mat=new Material;
	mat->set_color_vertex(true,false);
	mat->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
	mat->set_shader("basic");

	_mesh=new Mesh<VertexPC>(true);
	_mesh->alloc_vert((int)radius*100);
	_mesh->set_primitive(D3DPT_LINELIST);
	_mesh->set_render_verts(0);
	_mesh->set_material(mat,true);

	_radius=radius;
	_p[0]=p1;
	_p[1]=p2;
	_p[2]=p3;
	_p[3]=p4;

	_choose[0][0]=1.0f;
	_choose[1][0]=1.0f;
	_choose[1][1]=1.0f;
	for (int i=2; i<=3; ++i)
	{
		_choose[i][0]=1.0f;
		_choose[i][i]=1.0f;
		for (int j=1; j<i; ++j)
		{
			_choose[i][j]=_choose[i-1][j-1]+_choose[i-1][j];
		}
	}

	_strength=strength;

	for (size_t i=0; i<(size_t)(_radius*50); i++)
	{
		Particle	p;
		p.t=g_rnd.get_random_f(0.0f,1.0f);
		p.offset_pos.set(g_rnd.get_random_f(-_radius,_radius),g_rnd.get_random_f(-_radius,_radius),0.0f);
		p.pos=p.old_pos=get_pos(p.t)+p.offset_pos;

		_particles.push_back(p);
	}

	_color[0].set(1.0f,1.0f,1.0f,0.5f);
	_color[1].set(1.0f,1.0f,1.0f,0.0f);

	for (int i=0; i<10; i++)
	{
		_approx[i]=get_pos((float)i/9.0f);
		if (i>0)
		{
			_approx_dir[i-1]=_approx[i]-_approx[i-1];
			_approx_len[i-1]=_approx_dir[i-1].normalize();
		}
	}

	g_currents.push_back(this);
}

Currents::~Currents(void)
{
	if (_mesh)
	{
		delete _mesh;
		_mesh=NULL;
	}

	auto it=g_currents.begin();
	while (it!=g_currents.end())
	{
		if ((*it)==this)
		{
			g_currents.erase(it);
			break;
		}

		++it;
	}
}

void Currents::draw()
{
	if (!_draw) return;
	if (!_mesh) return;
	if (_particles.empty()) return;
	
	if (GFX->get_pass()==PASS_SOLID) return;

	GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0.0f,0.0f,0.0f,1.0f));

	_mesh->draw();
}

void Currents::anim(const float t)
{
	if (_particles.empty()) return;

	VertexPC*	vp=_mesh->get_write_pointer_v();
	UINT		n=0;

	unsigned long	color1=float_to_colorref(_color[0]);
	unsigned long	color2=float_to_colorref(_color[1]);
	auto			it=_particles.begin();
	while (it!=_particles.end())
	{
		it->old_pos=it->pos;
		it->t+=_strength*0.005f*t;
		if (it->t>1.0f)
		{
			it->t=1.0f;
			if (it->old_pos==it->pos)
			{
				// Reset this particle
				it->t=0.0f;
				it->old_pos=get_pos(0)+it->offset_pos;
			}
		}

		it->pos=get_pos(it->t)+it->offset_pos;

		// Update the mesh
		vp->pos=it->pos;
		vp->color=color1;

		vp++;

		vp->pos=it->old_pos;
		vp->color=color2;

		vp++;

		n++;
		++it;
	}
	_mesh->set_render_verts(n*2);
	_mesh->release_write_pointer_v();
}

Vec3f Currents::get_pos(const float t) const
{
	float	oneMinusT=(float)1.0f-t;
	float	powT=t;
	
	Vec3f result=oneMinusT*_p[0];

	for (int i=1; i<3; ++i)
	{
		float coeff=_choose[3][i]*powT;
		result=(result+coeff*_p[i])*oneMinusT;
		powT*=t;
	}

	result=result+powT*_p[3];

	return result;
}

Vec3f Currents::get_derivative(const float t) const
{
	float	oneMinusT=1.0f-t;
	float	powT=t;
	Vec3f	result=oneMinusT*_p[0];

	int degreeM1=2;
	for (int i=1; i<degreeM1; ++i)
	{
		float coeff=_choose[degreeM1][i]*powT;
		result=(result+coeff*_p[i])*oneMinusT;
		powT*=t;
	}

	result=result+powT*_p[degreeM1];
	result*=3.0f;

	return result;
}

bool Currents::get_force(const Vec3f& pos,Vec3f& force)
{
	float	min_dist=FLT_MAX;
	Vec3f	min_dir;
	float	t=FLT_MAX;
	Vec3f	diff;
	float	lp,d;
	for (int i=0; i<9; i++)
	{
		diff=pos-_approx[i];
		lp=dot_product(diff,_approx_dir[i]);
		if (lp<0) d=diff.norm_squared();
		else if (lp>_approx_len[i]) d=(pos-_approx[i+1]).norm_squared();
		else
		{
			diff=(_approx[i]+lp*_approx_dir[i])-pos;
			d=diff.norm_squared();
		}
		if ((d<min_dist) && (d<Sqr(_radius)))
		{
			t=(i/0.9f)+lp*0.1f;
			min_dist=d;
			min_dir=_approx_dir[i];
		}
	}

	if (t==FLT_MAX) return false;

	force=min_dir*_strength;

	return true;
}
