
#include "Defs.h"
#include "Gfx.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "Keyboard.h"
#include "Mouse.h"
#include "Timers.h"
#include "Sound.h"
#include "ModelManager.h"
#include "RndGenerator.h"
#include "Launcher.h"
#include "Screen.h"
#include "ImageScreen.h"
#include "XML.h"
#include "Utils.h"

#include "GameScreen.h"

bool init_gfx(HINSTANCE hInst)
{
	new Gfx;

	GFX->set_app_name("Pressure");
	GFX->set_window_mode(options_get_window_mode());
	GFX->set_resolution(options_get_res_x(),options_get_res_y(),options_get_bitdepth(),options_get_refresh_rate());
	GFX->set_virtual_resolution(1280,720);
	if (!GFX->init(hInst)) return false;

	new TextureManager();
	new BlitManager();
	new BlitManager3d();
	new ModelManager;

	return true;
}

bool init_input(HINSTANCE hInst)
{
	new Keyboard;
	if (!KEY->init(hInst)) return false;
	new Mouse;
	if (!MOUSE->init()) return false;

	return true;
}

bool init_sound()
{
	new Sound;

	if (!SOUND->init()) 
	{
		ERROR_MSG("Couldn't start sound... Game will be silent...");
	}
	

	return true;
}

bool init(HINSTANCE hInst)
{
	if (!init_gfx(hInst)) return false;
	if (!init_input(hInst)) return false;
	if (!init_sound()) return false;

	new Timers;

	return true;
}

void shutdown()
{
	delete SOUND;
	delete TIMERS;
	delete KEY;
	delete BLIT;
	delete BLIT3D;
	delete TEXTURES;
	delete MODEL_MANAGER;
	delete GFX;
}

#define N_ATMOSPHERE_SCHEMES	7
Color g_atmosphere_colors[N_ATMOSPHERE_SCHEMES]=
{
	Color(0.7f,0.1f,0.1f,1.0f),
	Color(0.7f,0.7f,0.1f,1.0f),
	Color(0.1f,0.7f,0.1f,1.0f),
	Color(0.1f,0.7f,0.7f,1.0f),
	Color(0.1f,0.1f,0.7f,1.0f),
	Color(0.7f,0.1f,0.7f,1.0f),
	Color(0.7f,0.7f,0.7f,1.0f),
};

void generate_level(const int level,LevelDef* dest)
{
	g_rnd.set_seed(dest->_next_seed);

	dest->_menu_mode=false;
	dest->_n_lives=3;
	dest->_level=level;
	dest->_rock_count=level*5;
	dest->_gravity=1.0f;
	dest->_density=1.0f;
	dest->_height=250.0f+level*10.0f;
	dest->_mineral_count=10+2*level;
	dest->_current_count=(level/2)*2;
	dest->_jellyfish_count=((level%2)==0)?(level*10):(0);
	if (level>2) dest->_pirate_count=((level%2)==1)?(level*5):(0);
	else dest->_pirate_count=0;

	int idx=g_rnd.get_random_i(0,N_ATMOSPHERE_SCHEMES-1);
	dest->_atmosphere_color[0]=g_atmosphere_colors[idx];
	idx=g_rnd.get_random_i(0,N_ATMOSPHERE_SCHEMES-1);
	dest->_atmosphere_color[1]=g_atmosphere_colors[idx]*0.2f; dest->_atmosphere_color[1].a=1.0f;
}

void new_game(LevelDef* def)
{
	def->_level=1;
	def->_xp=0;
	def->_available_upgrades=0;
	def->_loaded=false;
	def->_next_seed=timeGetTime();
	for (int i=0; i<UPGRADE_MAX; i++) def->_upgrades[i]=0;
	for (int i=0; i<TUTORIAL_MAX; i++) def->_tutorials[i]=false;
}

void save(const std::string& filename,LevelDef* def)
{
	scXMLWriterNode* root=new scXMLWriterNode(NULL,"UTF-8");
	scXMLWriterNode* save=new scXMLWriterNode(root,"savegame");
	save->add_property("clean",!def->_loaded);
	save->add_property("level",def->_level);
	save->add_property("xp",def->_xp);
	save->add_property("available_upgrades",def->_available_upgrades);
	save->add_property("seed",def->_next_seed);

	for (int i=0; i<UPGRADE_MAX; i++)
	{
		scXMLWriterNode* upgrade=new scXMLWriterNode(save,"upgrade");
		upgrade->add_property("type",i);
		upgrade->add_property("count",def->_upgrades[i]);
	}

	for (int i=0; i<TUTORIAL_MAX; i++)
	{
		scXMLWriterNode* tutorial=new scXMLWriterNode(save,"tutorial");
		tutorial->add_property("type",i);
		tutorial->add_property("state",def->_tutorials[i]);
	}
	
	root->output(filename);

	def->_loaded=true;
}

void load(const std::string& filename,LevelDef* def)
{
	new_game(def); 

	XMLNode node=XMLNode::openFileHelper(filename.c_str());
	if (node.isEmpty()) return;
	node=node.getChildNode("savegame");
	if (node.isEmpty()) return;

	def->_loaded=!node.getBoolAttribute("clean",NULL,true);
	def->_level=node.getIntAttribute("level",NULL,1);
	def->_xp=node.getIntAttribute("xp",NULL,1);
	def->_available_upgrades=node.getIntAttribute("available_upgrades",NULL,0);
	def->_next_seed=node.getIntAttribute("seed",NULL,timeGetTime());

	for (int i=0; i<node.nChildNode("upgrade"); i++)
	{
		XMLNode upgrade_node=node.getChildNode("upgrade",i);
		int		idx=upgrade_node.getIntAttribute("type",NULL,-1);
		if (idx==-1) continue;
		int		count=upgrade_node.getIntAttribute("count",NULL,0);

		def->_upgrades[idx]=count;
	}
	for (int i=0; i<node.nChildNode("tutorial"); i++)
	{
		XMLNode tutorial_node=node.getChildNode("tutorial",i);
		int		idx=tutorial_node.getIntAttribute("type",NULL,-1);
		if (idx==-1) continue;

		def->_tutorials[idx]=tutorial_node.getBoolAttribute("state",NULL,false);
	}

	def->_loaded=true;
}

INT WINAPI WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,INT)
{
	options_load();
	if (options_show_launcher())
	{
		if (!launch_launcher()) return -1;
		options_save();
	}

	if (!init(hInst)) return -1;

	Screen::load_assets("assets.dat");

	long music_id=-1;
	if (options_get_sound())
	{
		music_id=SOUND->play("music",options_get_music_volume()*0.5f);
	}

	ImageScreen* img_scr=new ImageScreen;
	img_scr->init("assets/spellcaster.dds");
	img_scr->run();
	delete img_scr;

	int			ret=0;
	GameScreen* game_screen;
	LevelDef	def;

	load("savegame.xml",&def);

	while (ret!=NEXT_SCREEN_EXIT)
	{
		generate_level(def._level,&def);
		def._menu_mode=true;
		def._rock_count*=3;

		if (music_id!=-1) SOUND->set_volume(music_id,options_get_music_volume()*0.5f);

		game_screen=new GameScreen;
		game_screen->init(&def);
		//ret=NEXT_SCREEN_NEWGAME;
		ret=game_screen->run();
		delete game_screen;

		if (ret==NEXT_SCREEN_NEWGAME)
		{
			new_game(&def);
		}
		else if (ret==NEXT_SCREEN_CONTINUE)
		{
			load("savegame.xml",&def);
		}
		
		if (ret!=NEXT_SCREEN_EXIT)
		{
			while (ret!=NEXT_SCREEN_MENU)
			{
				generate_level(def._level,&def);

				if (music_id!=-1) SOUND->set_volume(music_id,options_get_music_volume()*0.25f);

				game_screen=new GameScreen;
				game_screen->init(&def);
				ret=game_screen->run();
				delete game_screen;

				if (ret==NEXT_SCREEN_GAMEOVER)
				{
					new_game(&def);

					ret=NEXT_SCREEN_MENU;
				}

				// Decide next level seed
				def._next_seed=g_rnd.get_random();
				save("savegame.xml",&def);
			}
		}
	}

	if (music_id!=-1)
	{
		SOUND->stop(music_id);
	}

	shutdown();

	return 0;
}