#include "GameScreen.h"
#include "Gfx.h"
#include "Keyboard.h"
#include "Mouse.h"
#include "Object3d.h"
#include "Light.h"
#include "VoxelWorld.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "RndGenerator.h"
#include "Sparks.h"
#include "Font.h"
#include "MouseCursor.h"
#include "UIButton.h"
#include "Sound.h"
#include "Launcher.h"
#include "FSEffect.h"
#include "Utils.h"

#include "Player.h"
#include "btsExplosion.h"
#include "Minerals.h"
#include "Currents.h"
#include "Jellyfish.h"
#include "Pirate.h"

GameScreen::GameScreen()
	: _camera("main_camera")
{
	_xp_sound=0;
}

GameScreen::~GameScreen()
{
	for (size_t i=0; i<_objects3d.size(); i++)
	{
		if (_objects3d[i]) delete _objects3d[i];
	}
	_objects3d.clear();

	for (size_t i=0; i<_objects2d.size(); i++)
	{
		if (_objects2d[i]) delete _objects2d[i];
	}
	_objects2d.clear();

	for (size_t i=0; i<_lights.size(); i++)
	{
		if (_lights[i]) delete _lights[i];
	}
	_lights.clear();

/*	if (_effect)
	{
		delete _effect;
		_effect=NULL;
	}*/
	
	delete _voxels; _voxels=NULL;
}

const char* GameScreen::g_tooltips[UPGRADE_MAX] =
{
	"Armor:\nIncreases armor of the ship",
	"Tractor Beam:\nIncreases the speed of mineral harvesting",
	"Antigrav Plates:\nImproves vertical maneuverability",
	"Thrusters:\nImproves horizontal maneuverability",
	"Ammo:\nIncreases ammo carried",
	"Damage:\nIncreases damage of shots"
};

void GameScreen::init(LevelDef* def)
{
	_font=new Font;
	if (!_font->init(_font_size,"gomarice_rocks"))
	{
		delete _font;
		_font=NULL;
		ERROR_MSG("Can't create font!");
		return;
	}
	// Setup level
	_def=def;

	// Setup camera
	_camera._pos.set(0.0f,0.0f,-40.0f);
	_camera._ori.identity_quaternion();

	_camera_wander_inc.set(0.0f,0.0f,0.0f);
	_camera_target_pos.set(0.0f,0.0f,0.0f);
	_camera_wander_time=0.0f;
	_camera_start=true;

	// Setup player
	if (!_def->_menu_mode)
	{
		_player=new Player(100.0f+_def->_upgrades[UPGRADE_AMMO]*50.0f);
		_player->_pos.set(0.0f,_def->_height-13.0f,0.0f);
		_player->set_max_health(1000.0f+250.0f*_def->_upgrades[UPGRADE_ARMOR]);

		_objects3d.push_back(_player);
	}
	else _player=NULL;

	// Sparks
	_sparks=new Sparks;

	_objects3d.push_back(_sparks);

	// Setup light
	Light* light=new Light("main_light");
	light->_ori.rotation_quaternion(1.0f,0.0f,0.0f,80.0f);
	light->set_type(LIGHT_DIR);
	
	_lights.push_back(light);

	_voxels=new VoxelWorld;
	_voxels->create(128,256,4,16,16,4,2.0f,_def->_height/256.0f,4.0f);
	_voxels->set_material_texture(TEXTURES->find_by_name("voxel_materials"),4,4);
	_voxels->set_material_alpha_enable(false);

	for (int i=0; i<2; i++)
	{
		_voxels->set_uv_scale(i,Vec3f(1.0f,1.0f,1.0f));
		_voxels->set_color(i,Color(1.0f,1.0f,1.0f,1.0f));
		_voxels->set_emissive_color(i,Color(0.0f,0.0f,0.0f,1.0f));
	}

	for (int x=0; x<128; x++)
	{
		for (int z=0; z<4; z++)
		{
			int h=g_rnd.get_random_i(1,8);
			if ((x>58) && (x<70)) h=1;
			for (int y=0; y<h; y++)
			{
				_voxels->set(x,y,z,1);
			}
		}
	}

	// Rocks
	for (int i=0; i<def->_rock_count; i++)
	{
		int xx=g_rnd.get_random_i(14,114);
		int yy=(int)(256.0f*g_rnd.get_random_f(0.1f,0.9f));
		int sx=g_rnd.get_random_i(3,10);
		int sy=Max(4,sx/2);

		spawn_rock(xx,yy,sx,sy);
	}

	if (!cube_colliding(64,230,0,16))
	{
		spawn_rock(64+g_rnd.get_random_i(-8,8),220+g_rnd.get_random_i(-8,8),g_rnd.get_random_i(4,8),g_rnd.get_random_i(4,8));
	}

	_voxels->_pos.set(-128.0f,0.0f,-8.0f);
	_voxels->rebuild();

	for (int i=0; i<def->_mineral_count; i++)
	{
		bool	found_pos=false;
		Vec3f	p;
		while (!found_pos)
		{
			int px=g_rnd.get_random_i(4,120);
			int py=g_rnd.get_random_i(12,240);

			if (!cube_colliding(px,py,2,4)) 
			{
				found_pos=true;
				p.set(px*2.0f-128.0f,(float)py*(_def->_height/256.0f),0.0f);
			}
		}

		Minerals* minerals=new Minerals;
		minerals->_pos=p;

		_objects3d.push_back(minerals);
	}

	for (int i=0; i<def->_current_count; i++)
	{
		Vec3f p1(g_rnd.get_random_f(-128.0f,128.0f),g_rnd.get_random_f(10.0f,_def->_height-20.0f),0.0f);
		Vec3f p2=p1+((g_rnd.get_random_i(0,100)<50)?(-1.0f):(1.0f))*Vec3f(g_rnd.get_random_f(10.0f,20.0f),g_rnd.get_random_f(0.0f,20.0f),0.0f);
		Vec3f p3=p2+((g_rnd.get_random_i(0,100)<50)?(-1.0f):(1.0f))*Vec3f(g_rnd.get_random_f(10.0f,20.0f),g_rnd.get_random_f(0.0f,20.0f),0.0f);
		Vec3f p4=p3+((g_rnd.get_random_i(0,100)<50)?(-1.0f):(1.0f))*Vec3f(g_rnd.get_random_f(10.0f,20.0f),g_rnd.get_random_f(0.0f,20.0f),0.0f);
		Currents* current=new Currents(p1,p2,p3,p4,
									   g_rnd.get_random_f(0.05f,Min(0.3f,0.05f+0.05f*def->_level)),g_rnd.get_random_f(2.0f,5.0f));
		_objects3d.push_back(current);
	}

	for (int i=0; i<def->_jellyfish_count; i++)
	{
		bool	found_pos=false;
		Vec3f	p;
		while (!found_pos)
		{
			int px=g_rnd.get_random_i(4,120);
			int py=g_rnd.get_random_i(12,240);

			if (!cube_colliding(px,py,2,4)) 
			{
				found_pos=true;
				p.set(px*2.0f-128.0f,(float)py*(_def->_height/256.0f),0.0f);
			}
		}

		Jellyfish* fish=new Jellyfish;
		fish->_pos=p;
		fish->set_max_health(1000.0f+(200*def->_level));
		_objects3d.push_back(fish);
	}

	for (int i=0; i<def->_pirate_count; i++)
	{
		bool	found_pos=false;
		Vec3f	p;
		while (!found_pos)
		{
			int px=g_rnd.get_random_i(4,120);
			int py=g_rnd.get_random_i(12,240);

			if (!cube_colliding(px,py,2,4)) 
			{
				found_pos=true;
				p.set(px*2.0f-128.0f,(float)py*(_def->_height/256.0f),0.0f);
			}
		}

		float speed=Min(12.0f,6.0f+_def->_level*0.5f);
		Pirate* pirate=new Pirate(g_rnd.get_random_f(speed-2.0f,speed));
		pirate->_pos=p;
		pirate->set_max_health(1000.0f+(200*def->_level));
		_objects3d.push_back(pirate);
	}

	// 2d Elements
	MouseCursor* cursor=new MouseCursor("mouse_cursor");
	_objects2d.push_back(cursor);

	if (_def->_menu_mode)
	{
		Vec2f vres=GFX->get_virtual_resolution();
		UIButton* button=new UIButton("ui_continue",Vec3f(vres.x*0.5f-128.0f,300.0f,0.5f),"","continue","blop");
		if (!_def->_loaded) button->set_available(false);
		_objects2d.push_back(button);
		button=new UIButton("ui_newgame",Vec3f(vres.x*0.5f-128.0f,384.0f,0.5f),"","new_game","blop");
		_objects2d.push_back(button);
		button=new UIButton("ui_exit",Vec3f(vres.x*0.5f-128.0f,468.0f,0.5f),"","exit","blop");
		_objects2d.push_back(button);
	}

	Vec2f vres=GFX->get_virtual_resolution();
	float x=vres.x*0.5f-(UPGRADE_MAX/2)*70.0f;
	for (int i=0; i<UPGRADE_MAX; i++)
	{
		std::string img_name=string_format("ui_upgrade%02i",i+1);
		UIButton* button=new UIButton(img_name.c_str(),Vec3f(x,440.0f,0.2f),g_tooltips[i],string_format("u%i",i),"blop");
		
		button->set_available(false);
		button->set_draw(false);

		x+=70.0f;

		_upgrades.push_back(button);
		_objects2d.push_back(button);
	}

	UIButton* button=new UIButton("ui_next",Vec3f(vres.x*0.5f-70.0f,600.0f,0.2f),"Next Planet...","next","blop");
	button->set_available(false);
	button->set_draw(false);
	_upgrades.push_back(button);
	_objects2d.push_back(button);

	button=new UIButton("ui_menu",Vec3f(vres.x*0.5f+6.0f,600.0f,0.2f),"Save and Exit to Menu...","menu","blop");
	button->set_available(false);
	button->set_draw(false);
	_upgrades.push_back(button);
	_objects2d.push_back(button);

	_landing_icon_anim_index=0;
	_landing_icon_anim_time=0;

	// XP stuff
	_total_xp=0;
	_xp_factor=1.0f+(def->_level*0.25f);
	_xp_mode=false;
	_surveyed=false;
	_exit_warning_time=0.0f;
	_time_to_respawn=0.0f;
	_gameover=false;
	_paused_anim=false;

//	_effect=new FSEffect("main",2,false,0.5f);
}

void GameScreen::spawn_rock(const int xx,const int yy,const int sx,const int sy)
{
	for (int y=yy-sy; y<yy+sy; y++)
	{
		float	tmp=((float)abs(y-yy))/(float)sy;
		int		new_sx=(int)(Max(2,sx*(1.0f-tmp)));
		int		new_sz=(int)(Max(1,2*(1.0f-tmp)));
		for (int x=xx-new_sx; x<xx+new_sx; x++)
		{
			for (int z=Max(0,2-new_sz); z<Min(4,2+new_sz); z++)
			{
				if (((rand()%100)>10) || (z==2)) _voxels->set(x,y,z,1);
			}
		}
	}
}

void GameScreen::anim(const float t)
{
#ifdef DEV
	if (KEY->is_pressed(DIK_ESCAPE)) _exit=true;
#endif

	anim_camera(t);

	if (_paused_anim) 
	{
		CURSOR->anim(t);
		if (_tutorial_text!="")
		{
			if ((timeGetTime()-_time_of_last_tutorial)>100.0f)
			{
				if ((MOUSE->button_clicked(0)) ||
					(MOUSE->button_clicked(1)) ||
					(KEY->any_pressed()))
				{
					_tutorial_text="";
					_paused_anim=false;
				}
			}
		}
		auto mineral_it=Minerals::g_minerals.begin();
		while (mineral_it!=Minerals::g_minerals.end())
		{
			(*mineral_it)->anim(t);
			++mineral_it;
		}
		auto current_it=Currents::g_currents.begin();
		while (current_it!=Currents::g_currents.end())
		{
			(*current_it)->anim(t);
			++current_it;
		}
		return;
	}

	// Animate objects
	auto it=_objects3d.begin();
	while (it!=_objects3d.end())
	{
		(*it)->anim(t);
		++it;
	}

	auto it2d=_objects2d.begin();
	while (it2d!=_objects2d.end())
	{
		(*it2d)->anim(t);
		++it2d;
	}


	if (!_def->_menu_mode)
	{
		if ((_player->is_landed()) && (!_surveyed))
		{
			if (_player->_pos.y<3.0f)
			{
				_surveyed=true;
				int bonus_xp=(int)(200.0f*_xp_factor);
				spawn_combat_text(_player->_pos+Vec3f(0.0f,2.0f,0.0f),string_format("Survey Complete: +%i XP!",bonus_xp),Color(0.0f,1.0f,1.0f,0.7f),4000.0f);
				if (options_get_sound()) SOUND->play("survey",options_get_fx_volume());
				_total_xp+=bonus_xp;
			}
		}

		if (_gameover)
		{
			if ((MOUSE->button_clicked(0)) ||
				(MOUSE->button_clicked(1)) ||
				(KEY->is_pressed(DIK_RETURN)) ||
				(KEY->is_pressed(DIK_SPACE)) ||
				(KEY->is_pressed(DIK_ESCAPE)))
			{
				set_exit(true,NEXT_SCREEN_GAMEOVER);
			}
		}
		else
		{
			if (_player->get_health()<=0.0f)
			{
				if (_time_to_respawn==0.0f) _time_to_respawn=2000.0f;
				_time_to_respawn-=t;
				if (_time_to_respawn<=0.0f)
				{
					_def->_n_lives--;
					if (_def->_n_lives<1)
					{
						_gameover=true;
						_camera_target_pos=_camera._pos; _camera_target_pos.z=0.0f;
						_camera_wander_time=0.0f;
					}
					else
					{
						_player->set_draw(true);
						_camera_start=true;
						_player->set_inc_alpha(0.01f);
						_player->_pos.set(0.0f,_def->_height-13.0f,0.0f);
						_player->_velocity.set(0.0f,0.0f,0.0f);
						_player->set_max_health(1000.0f+250.0f*_def->_upgrades[UPGRADE_ARMOR]);
						_player->set_health(1000.0f+250.0f*_def->_upgrades[UPGRADE_ARMOR]);
						_time_to_respawn=0.0f;
					}
				}
			}
		}

		if (!_xp_mode)
		{
			if (_exit_warning_time!=0.0f) 
			{
				float prev_time=_exit_warning_time;
				_exit_warning_time-=t;

				if ((_exit_warning_time<=5000.0f) && (prev_time>5000.0f)) { spawn_combat_text(_player->_pos+Vec3f(0.0f,2.0f,0.0f),"Leaving planet in 5 seconds...",Color(0.0f,1.0f,1.0f,0.7f),750.0f); if (options_get_sound()) SOUND->play("leaving",options_get_fx_volume()); }
				if ((_exit_warning_time<=4000.0f) && (prev_time>4000.0f)) { spawn_combat_text(_player->_pos+Vec3f(0.0f,2.0f,0.0f),"Leaving planet in 4 seconds...",Color(0.0f,1.0f,1.0f,0.7f),750.0f); if (options_get_sound()) SOUND->play("leaving",options_get_fx_volume()); }
				if ((_exit_warning_time<=3000.0f) && (prev_time>3000.0f)) { spawn_combat_text(_player->_pos+Vec3f(0.0f,2.0f,0.0f),"Leaving planet in 3 seconds...",Color(0.0f,1.0f,1.0f,0.7f),750.0f); if (options_get_sound()) SOUND->play("leaving",options_get_fx_volume()); }
				if ((_exit_warning_time<=2000.0f) && (prev_time>2000.0f)) { spawn_combat_text(_player->_pos+Vec3f(0.0f,2.0f,0.0f),"Leaving planet in 2 seconds...",Color(0.0f,1.0f,1.0f,0.7f),750.0f); if (options_get_sound()) SOUND->play("leaving",options_get_fx_volume()); }
				if ((_exit_warning_time<=1000.0f) && (prev_time>1000.0f)) { spawn_combat_text(_player->_pos+Vec3f(0.0f,2.0f,0.0f),"Leaving planet in 1 second...",Color(0.0f,1.0f,1.0f,0.7f),750.0f); if (options_get_sound()) SOUND->play("leaving",options_get_fx_volume()); }
				if (_exit_warning_time<=0.0f)
				{
					// Move to tally
					_player->set_input(false);
					_xp_mode=true;
					_xp_time=0.0f;
				}
			}

			// Exit planet?
			if (_player->_pos.y>(_def->_height-12.0f))
			{
				if (_exit_warning_time==0.0f) _exit_warning_time=6000.0f;
			}
			else
			{
				_exit_warning_time=0.0f;
			}
		}
	}

	if (_xp_mode)
	{
		_xp_time+=t;
		
		if (_xp_time<1000.0f)
		{
			_player->_pos.y=_player->_pos.y+0.05f*t;
		}
		else if (_xp_time>1500.0f)
		{
			_player->stop_thruster();

			float minerals=_player->get_minerals();
			
			if (minerals>0.0f)
			{
				float m=Min(minerals,100.0f);
				_def->_xp+=(int)(m/10.0f);
				minerals-=m;
				_player->set_minerals(minerals);
				_xp_time=1500.0f;

				if (timeGetTime()-_xp_sound>50)
				{
					if (options_get_sound()) SOUND->play("xp",options_get_fx_volume());
					_xp_sound=timeGetTime();
				}
			}
			else if (_total_xp>0)
			{
				int m=Min(_total_xp,10);
				_def->_xp+=m;
				_total_xp-=m;
				_xp_time=1500.0f;

				if (timeGetTime()-_xp_sound>50)
				{
					if (options_get_sound()) SOUND->play("xp",options_get_fx_volume());
					_xp_sound=timeGetTime();
				}
			}
		}
		if (MOUSE->button_clicked(0))
		{
			for (size_t i=0; i<_upgrades.size(); i++)
			{
				if (_upgrades[i]->is_inside())
				{
					if (i>=UPGRADE_MAX) 
					{
						if (options_get_sound()) SOUND->play("blip",options_get_fx_volume());

						float minerals=_player->get_minerals();
						_def->_xp+=(int)(minerals/10.0f);
						_def->_xp+=_total_xp;

						if (_upgrades[i]->get_callback_id()=="next")
						{
							set_exit(true,NEXT_SCREEN_CONTINUE);
						}
						else if (_upgrades[i]->get_callback_id()=="menu")
						{
							set_exit(true,NEXT_SCREEN_MENU);
						}
					}
					else
					{
						if (_upgrades[i]->is_available()) 
						{
							if (_def->_available_upgrades>0)
							{
								if (options_get_sound()) SOUND->play("upgrade",options_get_fx_volume());

								_def->_available_upgrades--;
								_def->_upgrades[i]++;
							}
						}
					}
				}
			}
		}
	}

	// Add objects
	it=_to_add_objects3d.begin();
	while (it!=_to_add_objects3d.end())
	{
		_objects3d.push_back(*it);
		it=_to_add_objects3d.erase(it);
	}

	it2d=_to_add_objects2d.begin();
	while (it2d!=_to_add_objects2d.end())
	{
		_objects2d.push_back(*it2d);
		it2d=_to_add_objects2d.erase(it2d);
	}

	// Remove objects
	it=_to_del_objects3d.begin();
	while (it!=_to_del_objects3d.end())
	{
		actual_remove_object3d(*it);
		it=_to_del_objects3d.erase(it);
	}

	it2d=_to_del_objects2d.begin();
	while (it2d!=_to_del_objects2d.end())
	{
		actual_remove_object2d(*it2d);
		it2d=_to_del_objects2d.erase(it2d);
	}

	// Cleanup list
	it=_objects3d.begin();
	while (it!=_objects3d.end())
	{
		if (!(*it))
		{
			it=_objects3d.erase(it);
		}
		else ++it;
	}

	it2d=_objects2d.begin();
	while (it2d!=_objects2d.end())
	{
		if (!(*it2d))
		{
			it2d=_objects2d.erase(it2d);
		}
		else ++it2d;
	}

	// Animate landing icon
	_landing_icon_anim_time+=t;
	if (_landing_icon_anim_time>500.0f)
	{
		_landing_icon_anim_index=(_landing_icon_anim_index+1)%4;
		_landing_icon_anim_time-=500.0f;
	}

	// Check for clicks (menu)
	if (_def->_menu_mode)
	{
		if (MOUSE->button_clicked(0))
		{
			// Check if any button clicked
			UIButton*	sel_button=NULL;

			it2d=_objects2d.begin();
			while (it2d!=_objects2d.end())
			{
				UIButton* button=dynamic_cast<UIButton*>(*it2d);
				if (button)
				{
					if (button->is_inside())
					{
						sel_button=button;
					}
				}
				++it2d;
			}
			if (sel_button)
			{
				std::string callback=sel_button->get_callback_id();
				if (callback=="continue")
				{
					if (options_get_sound()) SOUND->play("blip",options_get_fx_volume());
					set_exit(true,NEXT_SCREEN_CONTINUE);
				}
				else if (callback=="new_game")
				{
					if (options_get_sound()) SOUND->play("blip",options_get_fx_volume());
					set_exit(true,NEXT_SCREEN_NEWGAME);
				}
				else if (callback=="exit")
				{
					if (options_get_sound()) SOUND->play("blip",options_get_fx_volume());
					set_exit(true,NEXT_SCREEN_EXIT);
				}
			}
		}
	}

	auto text_it=_texts.begin();
	while (text_it!=_texts.end())
	{
		(*text_it)._life-=t;
		if ((*text_it)._life<=0.0f)
		{
			text_it=_texts.erase(text_it);
		}
		else 
		{
			(*text_it)._pos.y+=0.001f*t;
			float t=(*text_it)._life/(*text_it)._max_life;

			(*text_it)._color.a=t;

			++text_it;
		}
	}

	trigger_tutorials();
}

void GameScreen::render()
{
	GFX->begin_scene();

	render3d();
	render2d();

	GFX->end_scene();
}

void GameScreen::render2d()
{
	Vec2f res=GFX->get_real_resolution();
	Vec2f vres=GFX->get_virtual_resolution();

	GFX->clear(D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
	GFX->set_ambient_light(Color(1.0f,1.0f,1.0f,1.0f));
	GFX->set_2d_render();
	GFX->set_camera2d(vres.x*0.5f,vres.y*0.5f,0.0f);
	GFX->clear_lights();
	
	GFX->set_pass(PASS_SOLID);

	auto it=_objects2d.begin();
	while (it!=_objects2d.end())
	{
		(*it)->draw();
		++it;
	}

	if (_def->_menu_mode)
	{
		BLIT->add(0.0f,0.0f,0.9f,vres.x,vres.y,Color(0.0f,0.0f,0.0f,0.5f));
		BLIT->add(TEXTURES->find_by_name("menu_assets"),vres.x*0.5f-158.0f*2.0f,100.0f,0.2f,328.0f*2.0f,78.0f*2.0f,1.0f,1.0f,1.0f,1.0f,0.0f,0.0f,328.0f/512.0f,78.0f/512.0f,D3DTEXF_POINT);

		_font->write_text(vres.x*0.5f,580.0f,0.25f,0.7f,0.7f,0.7f,0.7f,"A game by\nDiogo 'Covenant' Andrade\n(http://www.shadowcovenant.com | Twitter: shadowcovenant)\nLudum Dare 48-Hour Competition 29",TEXT_CENTER|TEXT_SHADOW);
	}
	else
	{
		float health=_player->get_health_percentage();
		render_icon(vres.x-210.0f,16.0f,0.1f,"ui_icons01",Rect(0.0f,0.0f,0.5f,0.5f),health);
		float ammo=_player->get_ammo_percentage();
		render_icon(vres.x-140.0f,16.0f,0.1f,"ui_icons01",Rect(0.5f,0.0f,1.0f,0.5f),ammo);
		float pressure=get_pressure_factor(_player->_pos.y);
		render_icon(vres.x-70.0f,16.0f,0.1f,"ui_icons01",Rect(0.0f,0.5f,0.5f,1.0f),pressure);
		BLIT->add(TEXTURES->find_by_name("ui_icons01"),vres.x-210.0f,84.0f,0.1f,64.0f,64.0f,1.0f,1.0f,1.0f,1.0f,0.5f,0.5f,1.0f,1.0f);
		std::string txt=string_format("x %i",(int)_player->get_minerals());
		_font->write_text(vres.x-140.0f,104.0f,0.1f,1.0f,1.0f,1.0f,1.0f,txt.c_str(),TEXT_SHADOW);
		BLIT->add(TEXTURES->find_by_name("ui_icons02"),vres.x-210.0f,128.0f,0.1f,64.0f,64.0f,1.0f,1.0f,1.0f,1.0f,0.0f,0.0f,0.5f,0.5f);
		txt=string_format("x %i",(int)_def->_n_lives);
		_font->write_text(vres.x-140.0f,168.0f,0.1f,1.0f,1.0f,1.0f,1.0f,txt.c_str(),TEXT_SHADOW);

		if ((_player->close_to_landing()) && (_player->get_health_percentage()>0.0f))
		{
			if (!_landing_icon_was_visible) { _landing_icon_anim_index=3; _landing_icon_anim_time=0.0f; }

			_landing_icon_was_visible=true;

			Rect	icon_uv[4]= { Rect(0.5f,0.5f,1.0f,1.0f),Rect(0.0f,0.5f,0.5f,1.0f),Rect(0.5f,0.0f,1.0f,0.5f),Rect(0.0f,0.0f,0.5f,0.5f) };
			Vec3f	p=convert_3d_to_2d(_player->_pos);
			Color	c(1.0f,1.0f,1.0f,0.5f);
			if (!_player->is_landing_speed())
			{
				c.set(1.0f,0.2f,0.2f,0.5f);
			}
			Rect	uv=icon_uv[_landing_icon_anim_index];
			BLIT->add(TEXTURES->find_by_name("ui_icons02"),p.x-32.0f,p.y-128.0f,0.1f,64.0f,64.0f,c.r,c.g,c.b,c.a,uv.x1,uv.y1,uv.x2,uv.y2);
		}
		else _landing_icon_was_visible=false;

		auto it=_texts.begin();
		while (it!=_texts.end())
		{
			Vec3f pos=convert_3d_to_2d((*it)._pos);
			Color c=(*it)._color;

			_font->write_text(pos.x,pos.y,0.05f,c.r,c.g,c.b,c.a,(*it)._text.c_str(),TEXT_CENTER|TEXT_SHADOW);

			++it;
		}
	}

	if (_gameover)
	{
		BLIT->add(TEXTURES->find_by_name("menu_assets"),vres.x*0.5f-183.0f,300.0f,0.2f,366.0f,160.0f,1.0f,1.0f,1.0f,1.0f,0.0f,96.0f/512.0f,366.0f/512.0f,0.5f);
	}

	if (_xp_mode)
	{
		if (_xp_time>1000.0f)
		{
			float next_level_xp=fibonacci(_def->_level+1)*1000.0f;
			// Write XP bar
			float p=_def->_xp/next_level_xp;
			if (p>=1.0f)
			{
				_def->_xp-=(int)next_level_xp;
				_def->_level++;
				_def->_available_upgrades++;
				next_level_xp=fibonacci(_def->_level+1)*1000.0f;
				p=_def->_xp/next_level_xp;

				if (options_get_sound()) SOUND->play("levelup",options_get_fx_volume());
				spawn_combat_text(Vec3f(_camera._pos.x,_camera._pos.y,0.0f),"LEVEL UP!",Color(1.0f,1.0f,0.0f,1.0f),2000.0f);
			}

			// Display XP
			_font->write_text(vres.x*0.5f,110.0f,0.2f,0.6f,1.0f,1.0f,0.8f,string_format("Current Level: %i",_def->_level).c_str(),TEXT_CENTER|TEXT_SHADOW);
			_font->write_text(vres.x*0.5f,160.0f,0.2f,0.6f,1.0f,1.0f,0.8f,string_format("Minerals: %i",(int)_player->get_minerals()).c_str(),TEXT_CENTER|TEXT_SHADOW);
			_font->write_text(vres.x*0.5f,210.0f,0.2f,0.6f,1.0f,1.0f,0.8f,string_format("XP: %i",_total_xp).c_str(),TEXT_CENTER|TEXT_SHADOW);
			_font->write_text(vres.x*0.5f,260.0f,0.2f,0.6f,1.0f,1.0f,0.8f,string_format("Total XP: %i / %i",_def->_xp,(int)next_level_xp).c_str(),TEXT_CENTER|TEXT_SHADOW);
			_font->write_text(vres.x*0.5f,400.0f,0.2f,0.6f,1.0f,1.0f,0.8f,string_format("Upgrades: %i",_def->_available_upgrades).c_str(),TEXT_CENTER|TEXT_SHADOW);

			BLIT->add(vres.x*0.5f-200.0f,310.0f,0.2f,400.0f,60.0f,Color(0.0f,0.0f,0.0f,1.0f));
			BLIT->add(vres.x*0.5f-196.0f,314.0f,0.2f,392.0f,52.0f,Color(0.5f,0.3f,0.0f,1.0f));
			BLIT->add(vres.x*0.5f-196.0f,314.0f,0.2f,392.0f*p,52.0f,Color(1.0f,1.0f,0.0f,1.0f));

			for (size_t i=0; i<_upgrades.size(); i++)
			{
				if (_upgrades[i]->is_inside())
				{
					// Render tooltip
					_font->write_text(CURSOR->_pos.x+16.0f,CURSOR->_pos.y+16.0f,0.05f,0.2f,0.6f,0.6f,0.8f,_upgrades[i]->get_tooltip().c_str(),TEXT_SHADOW);
				}

				Color text_color(0.5f,1.0f,1.0f,0.9f);

				_upgrades[i]->set_draw(true);
				if ((_def->_available_upgrades>0) || (i>=UPGRADE_MAX)) { _upgrades[i]->set_available(true); text_color.a=0.9f; }
				else { _upgrades[i]->set_available(false); text_color.a=0.5f; }

				if (i<UPGRADE_MAX) _font->write_text(_upgrades[i]->_pos.x+32.0f,_upgrades[i]->_pos.y+70.0f,0.2f,text_color.r,text_color.g,text_color.b,text_color.a,string_format("%i",_def->_upgrades[i]).c_str(),TEXT_CENTER|TEXT_SHADOW);
			}
		}	
	}

	if (_tutorial_text!="")
	{
		Vec2f vres=GFX->get_virtual_resolution();
		Rect r=_font->get_text_rect(vres.x*0.5f,284.0f,_tutorial_text.c_str(),TEXT_CENTER|TEXT_SHADOW);
		_font->write_text(vres.x*0.5f+_tutorial_pos.x,384.0f-(r.y2-r.y1)*0.5f+_tutorial_pos.y,0.2f,0.8f,1.0f,1.0f,0.9f,_tutorial_text.c_str(),TEXT_CENTER|TEXT_SHADOW);

		if (_tutorial_arrow!=ARROW_NONE)
		{
			render_arrow(Vec3f(vres.x*0.5f+_tutorial_pos.x,384.0f-(r.y2-r.y1)*0.5f+_tutorial_pos.y,0.5f),Vec3f(vres.x*0.5f+_tutorial_pos.x,384.0f+(r.y2-r.y1)*0.5f+_tutorial_pos.y,0.5f),Color(1.0f,1.0f,0.8f,0.9f));
		}
	}

	BLIT->flush();
}

void GameScreen::render_icon(const float x,const float y,const float z,const char* img_name,const Rect& uv,const float p)
{
	float	val=Clamp(p,0.0f,1.0f);
	int		img=TEXTURES->find_by_name(img_name);
	float	img_size_x=(float)TEXTURES->get_size_x(img)*(uv.x2-uv.x1)*0.5f;
	float	img_size_y=(float)TEXTURES->get_size_y(img)*(uv.y2-uv.y1)*0.5f;

	float	wh=img_size_y*(1.0f-val);
	BLIT->add(img,x,y,z,img_size_x,wh,1.0f,1.0f,1.0f,0.25f,uv.x1,uv.y1,uv.x2,uv.y1+(uv.y2-uv.y1)*(1.0f-val));
	BLIT->add(img,x,y+wh,z,img_size_x,img_size_y-wh,1.0f,1.0f,1.0f,1.0f,uv.x1,uv.y1+(uv.y2-uv.y1)*(1.0f-val),uv.x2,uv.y2);
}

void GameScreen::render3d()
{
	GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0.0f,0.0f,0.2f,1.0f,1.0f,0);
	GFX->set_ambient_light(Color(0.1f,0.1f,0.1f,1.0f));

	// Setup camera
	_camera.set();

	// Setup lights
	GFX->clear_lights();
	for (size_t i=0; i<_lights.size(); i++)
	{
		GFX->add_light(_lights[i]);
	}
	// Draw background
	Vec3f ld; _lights[0]->_ori.rotate_point(ld,Vec3f(0.0f,0.0f,1.0f));
	BLIT3D->add_quad(Vec3f(-160.0f,-10.0f,8.0f),Vec3f(160.0f,-10.0f,8.0f),Vec3f(140.0f,_def->_height+20.0f,8.0f),Vec3f(-140.0f,_def->_height+20.0f,8.0f),
					 _def->_atmosphere_color[1],_def->_atmosphere_color[1],_def->_atmosphere_color[0],_def->_atmosphere_color[0],
					 -ld,BLIT_NORMAL);
	BLIT3D->flush();

	// Setup fog
	Vec3f	fog_center;
	if ((_def->_menu_mode) || (_gameover)) fog_center.set(_camera._pos.x,_camera._pos.y,0.0f);
	else 
	{
		if ((_tutorial_text!="") && (_tutorial_focus)) fog_center=_tutorial_focus_point;
		else fog_center=_player->_pos;
	}

	float sfog=Max(0.0f,5.0f+((1.0f-_def->_density)*10));
	float efog=50.0f+((1.0f-_def->_density)*30);
	float position_y=fog_center.y/_def->_height;
	Color fog_color=_def->_atmosphere_color[0]*position_y+_def->_atmosphere_color[1]*(1.0f-position_y);
	GFX->set_fog(true,fog_color,sfog,efog,fog_center);
	// Render
	draw_pass(PASS_SOLID);
	draw_pass(PASS_ALPHA);

	GFX->set_fog(false,fog_color,10.0f,20.0f,_player->_pos);

/*	_effect->begin();
	GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0.0f,0.0f,0.0f,0.0f,1.0f,0);
	draw_pass(PASS_SOLID_ONLY_EMISSIVE);


	// Do glow effect
	Vec2f	r=GFX->get_real_resolution();

	float inv_x=1.0f/r.x;
	float inv_y=1.0f/r.y;

	_effect->set_constant("v_offset0",Vec4f(0.0f,1.3366f*inv_y,0.0f,0.0f));
	_effect->set_constant("v_offset1",Vec4f(0.0f,3.4295f*inv_y,0.0f,0.0f));
	_effect->set_constant("v_offset2",Vec4f(0.0f,5.4264f*inv_y,0.0f,0.0f));
	_effect->set_constant("p_offset0",Vec4f(0.0f,7.4359f*inv_y,0.0f,0.0f));
	_effect->set_constant("p_offset1",Vec4f(0.0f,9.4436f*inv_y,0.0f,0.0f));
	_effect->set_constant("p_offset2",Vec4f(0.0f,11.4401f*inv_y,0.0f,0.0f));
	_effect->set_constant("glow_str",Vec4f(1.5f,0.0f,0.0f,0.0f));
	_effect->run("fx_blur");

	_effect->set_constant("v_offset0",Vec4f(1.3366f*inv_y,0.0f,0.0f,0.0f));
	_effect->set_constant("v_offset1",Vec4f(3.4295f*inv_y,0.0f,0.0f,0.0f));
	_effect->set_constant("v_offset2",Vec4f(5.4264f*inv_y,0.0f,0.0f,0.0f));
	_effect->set_constant("p_offset0",Vec4f(7.4359f*inv_y,0.0f,0.0f,0.0f));
	_effect->set_constant("p_offset1",Vec4f(9.4436f*inv_y,0.0f,0.0f,0.0f));
	_effect->set_constant("p_offset2",Vec4f(11.4401f*inv_y,0.0f,0.0f,0.0f));
	_effect->set_constant("glow_str",Vec4f(1.5f,0.0f,0.0f,0.0f));
	_effect->run("fx_blur");

	_effect->end(D3DBLEND_ONE,D3DBLEND_ONE);*/
}

void GameScreen::draw_pass(const PassType pass)
{
	GFX->set_pass(pass);

	_voxels->render(&_camera);

	auto it=_objects3d.begin();
	while (it!=_objects3d.end())
	{
		(*it)->draw();
		++it;
	}

	BLIT3D->flush();
}

bool GameScreen::check_collision(const Vec3f& p,Vec3f& collision_point)
{
	float sy=(_def->_height/256.0f);
	// Check collision with voxels
	int x=(int)((p.x+128.0f+1.0f)/2.0f); if ((x<0) || (x>=128)) return false;
	int y=(int)((p.y+sy*0.5f)/sy); if ((y<0) || (y>=256)) return false;
	int z=2;
	if (_voxels->get(x,y,z)==0) return false;

	collision_point.x=(x*2.0f)-128.0f;
	collision_point.y=(y*sy);
	collision_point.z=p.z;

	return true;
}

void GameScreen::spawn_sparks(const Vec3f& pos,const float speed,const int n_particles,const float max_life,const unsigned long color1,const unsigned long color2)
{
	_sparks->new_explosion(pos,speed,n_particles,max_life,color1,color2);
}

void GameScreen::spawn_explosion(const Vec3f& pos,const float radius,Object3d* source)
{
	btsExplosion* explosion=new btsExplosion(pos,Vec3f(0.0f,1.0f,0.0f),radius,false,Vec3f(0.0f,0.0f,0.0f),source);

	_to_add_objects3d.push_back(explosion);
}

void GameScreen::remove_object3d(Object3d* obj)
{
	_to_del_objects3d.push_back(obj);
}

void GameScreen::actual_remove_object2d(Drawable* obj)
{
	auto it=_objects2d.begin();
	while (it!=_objects2d.end())
	{
		if ((*it)==obj)
		{
			delete (*it);
			(*it)=NULL;
		}
		++it;
	}
}

void GameScreen::actual_remove_object3d(Object3d* obj)
{
	auto it=_objects3d.begin();
	while (it!=_objects3d.end())
	{
		if ((*it)==obj)
		{
			delete (*it);
			(*it)=NULL;
		}
		++it;
	}
}

float GameScreen::get_pressure_factor(const float y)
{
	float f=Clamp(1.0f-y/_def->_height,0.0f,1.0f);
	f=f*_def->_density;

	return f;
}

bool GameScreen::cube_colliding(const int x,const int y,const int z,const int r)
{
	for (int xx=Max(0,x-r); xx<Min(128,x+r); xx++)
	{
		for (int yy=Max(0,y-r); yy<Min(256,y+r); yy++)
		{
			for (int zz=Max(0,z-r); zz<Min(4,z+r); zz++)
			{
				if (_voxels->get(xx,yy,zz)!=0) return true;
			}
		}
	}

	return false;
}

Vec3f GameScreen::get_mouse_pos()
{
	Vec3f	mp=CURSOR->_pos;
	Vec3f	origin,dir;
	
	_camera.extract_ray((unsigned long)mp.x,(unsigned long)mp.y,origin,dir);

	float t=origin.z/dir.z;
	
	return origin-t*dir;
}

void GameScreen::spawn_combat_text(const Vec3f& pos,const std::string& str,const Color& c,const float life)
{
	Text	txt;
	txt._text=str;
	txt._pos=pos;
	txt._life=life;
	txt._max_life=life;
	txt._color=c;
	txt._value=FLT_MAX;

	_texts.push_back(txt);
}

Vec3f GameScreen::convert_3d_to_2d(const Vec3f& pos)
{
	return GFX->transform_vertex(_camera._fov,_camera._aspect,_camera._near_plane,_camera._far_plane,_camera._pos.x,_camera._pos.y,_camera._pos.z,_camera._ori,pos);
}

void GameScreen::add_xp(const int xp,const Vec3f& pos)
{
	spawn_combat_text(pos,string_format("+%i XP",xp),Color(0.0f,1.0f,1.0f,0.7f),1000.0f);
	_total_xp+=xp;
}

bool GameScreen::has_los(const Vec3f& p1,const Vec3f& p2)
{
	float t;
	Vec3f dir=p2-p1;
	float max_t=dir.normalize();
	if (_voxels->raycast(p1-_voxels->_pos,dir,t,max_t))
	{
		return false;
	}
	return true;
}

void GameScreen::trigger_tutorials()
{
	if (_def->_menu_mode) return;

	Vec2f vres=GFX->get_virtual_resolution();
	
	if (_xp_mode)
	{
		if ((!_def->_tutorials[TUTORIAL_XP_MODE]) && (_xp_time>1600.0f))
		{
			trigger_tutorial(Vec3f(-320.0f,-256.0f,0.0f),"When you leave a planet, you'll be able\nto convert your experience into upgrades.\nPlanets will be more difficult the higher\nyour level, so gear up.\nHover the mouse cursor on top of these\nbuttons to learn more.",false,Vec3f(0.0f,0.0f,0.0f),ARROW_2D,Vec3f(vres.x*0.5f,400.0f,0.0f));
			_def->_tutorials[TUTORIAL_XP_MODE]=true;
			return;
		}
	}
	else
	{
		if (!_def->_tutorials[TUTORIAL_INTRO])
		{
			trigger_tutorial(Vec3f(-256.0f,128.0f,0.0f),"Welcome to 'Pressure'!\nIn this game, you control a mining vessel that\ngather minerals from the atmosphere of gas giants.\n \nPress a key or click the\nmouse to continue!",false,Vec3f(0.0f,0.0f,0.0f),ARROW_3D,_player->_pos);
			_def->_tutorials[TUTORIAL_INTRO]=true;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_CONTROLS])
		{
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"The gravity of gas giants is huge,\nso you'll have to keep using your thrusters.\nTo fire your thrusters, use the arrow keys\nor WASD.",false,Vec3f(0.0f,0.0f,0.0f),ARROW_NONE,_player->_pos);
			_def->_tutorials[TUTORIAL_CONTROLS]=true;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_UI_ARMOR])
		{
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"Avoid the rocky formations and enemies,\n or else you'll take damage.\nYou can see the status of your hull here.",false,Vec3f(0.0f,0.0f,0.0f),ARROW_2D,Vec3f(vres.x-210.0f,48.0f,0.5f));
			_def->_tutorials[TUTORIAL_UI_ARMOR]=true;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_UI_PRESSURE])
		{
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"You also take damage from the\npressure. On this dial, you can see\nhow much pressure you're whitstanding!\nThe deeper you go, the more damage you take!",false,Vec3f(0.0f,0.0f,0.0f),ARROW_2D,Vec3f(vres.x-70.0f,48.0f,0.5f));
			_def->_tutorials[TUTORIAL_UI_PRESSURE]=true;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_UI_LIVES])
		{
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"If you take too much damage, your ship will explode!\nYou can see here how many lives you still have.\nIf you loose all of them, you'll have to start\nover again!",false,Vec3f(0.0f,0.0f,0.0f),ARROW_2D,Vec3f(vres.x-210.0f,182.0f,0.5f));
			_def->_tutorials[TUTORIAL_UI_LIVES]=true;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_UI_MINERALS])
		{
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"Your objective is to gather as much\nminerals as you can. You can see how many you got on this display.",false,Vec3f(0.0f,0.0f,0.0f),ARROW_2D,Vec3f(vres.x-210.0f,112.0f,0.5f));
			_def->_tutorials[TUTORIAL_UI_MINERALS]=true;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_VEINS])
		{
			float	min_dist=FLT_MAX;
			Vec3f	mineral_pos;
			auto it=Minerals::g_minerals.begin();
			while (it!=Minerals::g_minerals.end())
			{
				float d=((*it)->_pos-_player->_pos).norm_squared();
				if (d<min_dist)
				{
					min_dist=d;
					mineral_pos=(*it)->_pos;
				}
				++it;
			}
			mineral_pos=mineral_pos+Vec3f(0.0f,-5.0f,0.0f);
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"To gather minerals, move your ship slightly above\nthese free minerals, and hold it here\nwhile it harvests them.",true,mineral_pos,ARROW_NONE,Vec3f(vres.x-210.0f,112.0f,0.5f));
			_def->_tutorials[TUTORIAL_VEINS]=true;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_REPAIR])
		{
			// Find a flat place where the ship can land
			Vec3f	old_pos=_player->_pos;
			while (!_player->close_to_landing())
			{
				_player->_pos.y-=1.0f;
				if (_player->_pos.y<0.0f)
				{
					_def->_tutorials[TUTORIAL_REPAIR]=true;
					return;
				}
			}

			Vec3f flat=_player->_pos;
			Vec3f flat_focus=flat+Vec3f(0.0f,15.0f,0.0f);
			_def->_tutorials[TUTORIAL_REPAIR]=true;
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"You can land your ship in flat rock.\nYou have to drop the ship down very slowly, or\nelse you'll take damage!\nWhen landed, the ship will automatically repair\nyour hull with minerals, as well as replenish\nyour ammo!",true,flat_focus,ARROW_3D,flat);
			_player->_pos=old_pos;
			return;
		}
		if (!_def->_tutorials[TUTORIAL_SURVEY])
		{
			// Find a flat place where the ship can land
			Vec3f flat=Vec3f(0.0f,0.0f,0.0f);
			_def->_tutorials[TUTORIAL_SURVEY]=true;
			trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"You'll get extra point for upgrades if you\nsurvey the rocky core.\nTo do so, you have to land here!",true,flat,ARROW_3D,flat);
			return;
		}
		if (!_def->_tutorials[TUTORIAL_LEAVING])
		{
			if (_player->get_minerals()>2000.0f)
			{
				// Find a flat place where the ship can land
				_def->_tutorials[TUTORIAL_LEAVING]=true;
				trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"When you think you've harvested enough, you\ncan leave the planet by moving your ship to the outer\natmosphere!!",true,Vec3f(0.0f,_def->_height,0.0f),ARROW_NONE,Vec3f(0.0f,0.0f,0.0f));
				return;
			}
		}
		if (!_def->_tutorials[TUTORIAL_ENEMIES])
		{
			if (!GameEnemy::g_enemies.empty())
			{
				float	min_dist=FLT_MAX;
				Vec3f	enemy_pos;
				auto it=GameEnemy::g_enemies.begin();
				while (it!=GameEnemy::g_enemies.end())
				{
					float d=((*it)->_pos-_player->_pos).norm_squared();
					if (d<min_dist)
					{
						min_dist=d;
						enemy_pos=(*it)->_pos;
					}
					++it;
				}

				// Find a flat place where the ship can land
				Vec3f enemy_pos_focus=enemy_pos+Vec3f(0.0f,-10.0f,0.0f);
				_def->_tutorials[TUTORIAL_ENEMIES]=true;
				trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"The gas giants are full of enemies!\nAvoid them at all cost, or destroy them with your\nlaser cannons.When\nTo fire, point your mouse cursor and click the left\nmouse button!",true,enemy_pos_focus,ARROW_3D,enemy_pos);
				return;
			}
		}
		if (!_def->_tutorials[TUTORIAL_UI_AMMO])
		{
			if (!GameEnemy::g_enemies.empty())
			{
				_def->_tutorials[TUTORIAL_UI_AMMO]=true;
				trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"You have limited ammo, but you can replenish\nit landing on flat ground.",false,Vec3f(0.0f,0.0f,0.0f),ARROW_2D,Vec3f(vres.x-140.0f,64.0f,0.5f));
				return;
			}
		}
		if (!_def->_tutorials[TUTORIAL_CURRENTS])
		{
			if (!Currents::g_currents.empty())
			{
				float	min_dist=FLT_MAX;
				Vec3f	current_pos;
				auto it=Currents::g_currents.begin();
				while (it!=Currents::g_currents.end())
				{
					float d=((*it)->get_center()-_player->_pos).norm_squared();
					if (d<min_dist)
					{
						min_dist=d;
						current_pos=(*it)->get_center();
					}
					++it;
				}

				// Find a flat place where the ship can land
				Vec3f current_pos_focus=current_pos+Vec3f(0.0f,-10.0f,0.0f);
				_def->_tutorials[TUTORIAL_CURRENTS]=true;
				trigger_tutorial(Vec3f(0.0f,0.0f,0.0f),"The atmospheric currents of a gas giant are also\nvery dangerous... Be careful with these treacherous\ncurrents!",true,current_pos_focus,ARROW_3D,current_pos);
				return;
			}
		}
	}
}

void GameScreen::trigger_tutorial(const Vec3f& tutorial_pos,const std::string& text,const bool focus_on,const Vec3f& focus_point,const ArrowType arrow_type,const Vec3f& point_to)
{
	_paused_anim=true;
	_tutorial_text=text;
	_tutorial_pos=tutorial_pos;
	_tutorial_arrow=arrow_type;
	_tutorial_arrow_target=point_to;
	_tutorial_focus=focus_on;
	_tutorial_focus_point=focus_point;
	_time_of_last_tutorial=timeGetTime();
}

void GameScreen::render_arrow(const Vec3f& sp1,const Vec3f& sp2,const Color& color)
{
	Vec3f sp,p,perp,delta1,delta2,delta;
	if (_tutorial_arrow==ARROW_3D)
	{
		p=convert_3d_to_2d(_tutorial_arrow_target); p.z=0.025f;
	}
	else
	{
		p=_tutorial_arrow_target; p.z=0.025f;
	}

	float d1,d2;
	delta1=p-Vec3f(sp1.x,sp1.y,0.025f); 
	d1=delta1.normalize();
	delta2=p-Vec3f(sp2.x,sp2.y,0.025f); 
	d2=delta2.normalize();
	if (d1<d2) { delta=delta1; sp.set(sp1.x,sp1.y,0.025f); }
	else { delta=delta2; sp.set(sp2.x,sp2.y,0.025f); }

	perp=delta; perp.normalize(); perp.set(-perp.y,perp.x,0.0f);

	BLIT->add_quad(sp-perp*10.0f,p-delta*40.0f-perp*10.0f,sp+perp*10.0f,p-delta*40.0f+perp*10.0f,color);
	BLIT->add_tri(p,p-perp*20.0f-delta*40.0f,p+perp*20.0f-delta*40.0f,color);
}

void GameScreen::anim_camera(const float t)
{
	// Animate camera
	float	speed_factor;
	Vec3f	camera_focus;
	if ((_def->_menu_mode) || (_gameover))
	{
		speed_factor=0.5f;

		// Move camera around
		_camera_wander_time-=t;
		if (_camera_wander_time<=0.0f)
		{
			Vec3f new_target(g_rnd.get_random_f(-120.0f,120.0f),g_rnd.get_random_f(0.0f,_def->_height),0.0f);
			_camera_wander_inc=new_target-_camera_target_pos;
			float d=_camera_wander_inc.normalize();

			float speed=250.0f;
			_camera_wander_time=d*speed;
			_camera_wander_inc=_camera_wander_inc/speed;
		}
		else
		{
			_camera_target_pos=_camera_target_pos+_camera_wander_inc*t;
		}

		camera_focus=_camera_target_pos;
	}
	else
	{
		Vec3f	velocity=_player->get_velocity();
		speed_factor=1.0f+Sqr(velocity.norm())*3.0f;
		camera_focus=_player->_pos;
		if (_tutorial_text!="")
		{
			if (_tutorial_focus)
			{
				camera_focus=_tutorial_focus_point;
			}
		}
	}

	Vec3f	camera_new_pos=_camera._pos;
	camera_new_pos.x=Max(Min(camera_focus.x,100.0f-(speed_factor*20)),-100.0f+(speed_factor*20));
	camera_new_pos.y=Max(camera_focus.y,13.0f+(speed_factor*10));
	camera_new_pos.y=Min(camera_new_pos.y,(_def->_height)-(speed_factor*10));
	camera_new_pos.z=-40.0f-(speed_factor*30.0f);

	if (_camera_start)
	{
		_camera._pos=camera_new_pos;
		_camera_start=false;
	}
	else _camera._pos=_camera._pos+(camera_new_pos-_camera._pos)*0.05f;
}
