#include "Camera.h"
#include "Gfx.h"

Camera*	CAMERA3D=NULL;

Camera::Camera(const char* name) : Object3d("camera3d",name)
{
	_fov=80.0f;

	Vec2f res=GFX->get_virtual_resolution();
	_aspect=float(res.y)/float(res.x);
	_near_plane=1.0f;
	_far_plane=1000.0f;
}

Camera::~Camera(void)
{
}

void Camera::set()
{ 
	GFX->set_perspective(_fov,_aspect,_near_plane,_far_plane);
	GFX->set_camera(_pos.x,_pos.y,_pos.z,_ori);

	_ori.rotate_point(_right_axis,Vec3f(1.0f,0.0f,0.0f));
	_y_plane_right_axis=_right_axis; _y_plane_right_axis.y=0.0f;
	_y_plane_right_axis.normalize();

	_ori.rotate_point(_up_axis,Vec3f(0.0f,1.0f,0.0f));

	_ori.rotate_point(_normal_axis,Vec3f(0.0f,0.0f,-1.0f));
	_y_plane_normal_axis=_normal_axis; _y_plane_normal_axis.y=0.0f;
	_y_plane_normal_axis.normalize();

	CAMERA3D=this;
}

void Camera::set_lookat(const Vec3f& f,const Vec3f& t)
{
	D3DXMATRIX	mat;
	D3DXVECTOR3	up=D3DXVECTOR3(0.0f,1.0f,0.0f);
	D3DXVECTOR3	from=D3DXVECTOR3(f.x,f.y,f.z);
	D3DXVECTOR3	to=D3DXVECTOR3(t.x,t.y,t.z);
	D3DXMatrixLookAtLH(&mat,&from,&to,&up);

	_pos=f;
	
	D3DXQUATERNION o;
	D3DXQuaternionIdentity(&o);
	D3DXQuaternionRotationMatrix(&o,&mat);
	D3DXQuaternionInverse(&o,&o);

	_ori=o;
}

void Camera::extract_ray(const unsigned long mpx,const unsigned long mpy,Vec3f& ray_origin,Vec3f& ray_dir)
{
	// Convert screen position into a raycast
	Vec2f	res=GFX->get_virtual_resolution();
	float	mx=mpx/res.x;
	float	my=mpy/res.y;
	float	aspect=float(res.x)/float(res.y);
	float	tan_fov_div_2_x=tanf(deg_to_rad(aspect*_fov/4.0f));
	float	tan_fov_div_2_y=tanf(deg_to_rad(_fov/4.0f));

	float	vx=(mx*2.0f)-1.0f;
	float	vy=1.0f-(my*2.0f);

	float	dx=tan_fov_div_2_x*vx;
	float	dy=tan_fov_div_2_y*vy;

	Vec3f	v(dx,dy,1.0f); 

	v=v*_far_plane;
	v.normalize();

	ray_origin=_pos;
	_ori.rotate_point(ray_dir,v);
}

bool Camera::inside(const Vec3f& pmin,const Vec3f& pmax)
{
	const Vec3f		box_axis[3]={ Vec3f(1.0f,0.0f,0.0f),Vec3f(0.0f,1.0f,0.0f),Vec3f(0.0f,0.0f,1.0f) };
	const float		box_extents[3]={ (pmax.x-pmin.x)*0.5f, (pmax.y-pmin.y)*0.5f, (pmax.z-pmin.z)*0.5f };

	Vec3f			kDiff=(pmax+pmin)*0.5f-_pos;

	float				afA[3],afB[3],afC[3],afD[3];
	float				afNA[3],afNB[3],afNC[3],afND[3];
	float				afNApLC[3],afNAmLC[3],afNBpUC[3],afNBmUC[3];
	float				afLC[3],afLD[3],afUC[3],afUD[3],afLBpUA[3],afLBmUA[3];
	float				fDdD,fR,fP,fMin,fMax,fMTwoLF,fMTwoUF,fLB,fUA,fTmp;
	unsigned short		i,j;

	Vec3f				frustum_dir; _ori.rotate_point(frustum_dir,Vec3f(0.0f,0.0f,1.0f));
	Vec3f				frustum_right; _ori.rotate_point(frustum_right,Vec3f(1.0f,0.0f,0.0f));
	Vec3f				frustum_up; _ori.rotate_point(frustum_up,Vec3f(0.0f,1.0f,0.0f));
	float				frustum_rdist=Tan(deg_to_rad(_fov*0.5f))*_near_plane;
	float				frustum_udist=Tan(deg_to_rad(_fov*_aspect*0.5f))*_near_plane;
	float				frustum_d_ratio=_far_plane/_near_plane;
	float				frustum_mt_wo_lf=-2.0f*frustum_rdist*_far_plane;
	float				frustum_mt_wo_rf=2.0f*frustum_rdist*_far_plane;
	float				frustum_mt_wo_uf=-2.0f*frustum_udist*_far_plane;

	// M = D
	afD[2]=dot_product(kDiff,frustum_dir);
	for (i=0; i<3; i++) afC[i]=dot_product(box_axis[i],frustum_dir);
	fR=(float)(box_extents[0]*fabs(afC[0])+
		box_extents[1]*fabs(afC[1])+
		box_extents[2]*fabs(afC[2]));
	if ((afD[2]+fR<_near_plane) || (afD[2]-fR>_far_plane)) return false;

	// M = n*L - l*D
	for (i=0; i<3; i++)
	{
		afA[i]=dot_product(box_axis[i],-frustum_right);
		afLC[i]=frustum_rdist*afC[i];
		afNA[i]=_near_plane*afA[i];
		afNAmLC[i]=afNA[i]-afLC[i];
	}
	afD[0]=dot_product(kDiff,-frustum_right);
	fR=(float)(box_extents[0]*fabs(afNAmLC[0])+
		box_extents[1]*fabs(afNAmLC[1])+
		box_extents[2]*fabs(afNAmLC[2]));
	afND[0]=_near_plane*afD[0];
	afLD[2]=frustum_rdist*afD[2];
	fDdD=afND[0]-afLD[2];
	fMTwoLF=frustum_mt_wo_lf;
	if ((fDdD+fR<fMTwoLF) || (fDdD>fR)) return false;

	// M = -n*L - l*D
	for (i=0; i<3; i++) afNApLC[i]=afNA[i]+afLC[i];
	fR=(float)(box_extents[0]*fabs(afNApLC[0])+
		box_extents[1]*fabs(afNApLC[1])+
		box_extents[2]*fabs(afNApLC[2]));
	fDdD=-(afND[0]+afLD[2]);
	if ((fDdD+fR<fMTwoLF) || (fDdD>fR)) return false;

	// M = n*U - u*D
	for (i=0; i<3; i++)
	{
		afB[i]=dot_product(box_axis[i],frustum_up);
		afUC[i]=frustum_udist*afC[i];
		afNB[i]=_near_plane*afB[i];
		afNBmUC[i]=afNB[i]-afUC[i];
	}
	afD[1]=dot_product(kDiff,frustum_up);
	fR=(float)(box_extents[0]*fabs(afNBmUC[0])+
		box_extents[1]*fabs(afNBmUC[1])+
		box_extents[2]*fabs(afNBmUC[2]));
	afND[1]=_near_plane*afD[1];
	afUD[2]=frustum_udist*afD[2];
	fDdD=afND[1]-afUD[2];
	fMTwoUF=frustum_mt_wo_uf;
	if ((fDdD+fR<fMTwoUF) || (fDdD>fR)) return false;

	// M = -n*U - u*D
	for (i=0; i<3; i++) afNBpUC[i]=afNB[i]+afUC[i];
	fR=(float)(box_extents[0]*fabs(afNBpUC[0])+
		box_extents[1]*fabs(afNBpUC[1])+
		box_extents[2]*fabs(afNBpUC[2]));
	fDdD=-(afND[1]+afUD[2]);
	if ((fDdD+fR<fMTwoUF) || (fDdD>fR)) return false;

	// M = A[i]
	for (i=0; i<3; i++)
	{
		fP=(float)(frustum_rdist*fabs(afA[i])+
			frustum_udist*fabs(afB[i]));
		afNC[i]=_near_plane*afC[i];
		fMin=afNC[i]-fP;
		if (fMin<0.0f) fMin*=frustum_d_ratio;
		fMax=afNC[i]+fP;
		if (fMax>0.0f) fMax*=frustum_d_ratio;
		fDdD=afA[i]*afD[0]+afB[i]*afD[1]+afC[i]*afD[2];
		if ((fDdD+box_extents[i]<fMin) || (fDdD-box_extents[i]>fMax)) return false;
	}

	// M = Cross(L,A[i])
	for (i=0; i<3; i++)
	{
		fP=(float)(frustum_udist*fabs(afC[i]));
		fMin=afNB[i]-fP;
		if (fMin<0.0f) fMin*=frustum_d_ratio;
		fMax=afNB[i]+fP;
		if (fMax>0.0f) fMax *= frustum_d_ratio;
		fDdD=-afC[i]*afD[1]+afB[i]*afD[2];
		fR=(float)(box_extents[0]*fabs(afB[i]*afC[0]-afB[0]*afC[i])+
			box_extents[1]*fabs(afB[i]*afC[1]-afB[1]*afC[i])+
			box_extents[2]*fabs(afB[i]*afC[2]-afB[2]*afC[i]));
		if ((fDdD+fR<fMin) || (fDdD-fR>fMax)) return false;
	}

	// M = Cross(U,A[i])
	for (i=0; i<3; i++)
	{
		fP=(float)(frustum_rdist*fabs(afC[i]));
		fMin=-afNA[i]-fP;
		if (fMin<0.0f )
			fMin*=frustum_d_ratio;
		fMax=-afNA[i] + fP;
		if (fMax>0.0f) fMax*=frustum_d_ratio;
		fDdD=afC[i]*afD[0]-afA[i]*afD[2];
		fR=(float)(box_extents[0]*fabs(afA[i]*afC[0]-afA[0]*afC[i])+
			box_extents[1]*fabs(afA[i]*afC[1]-afA[1]*afC[i])+
			box_extents[2]*fabs(afA[i]*afC[2]-afA[2]*afC[i]));
		if ((fDdD+fR<fMin) || (fDdD-fR>fMax)) return false;
	}

	// M = Cross(n*D+l*L+u*U,A[i])
	for (i=0; i<3; i++)
	{
		fLB=frustum_rdist*afB[i];
		fUA=frustum_udist*afA[i];
		afLBpUA[i]=fLB+fUA;
		afLBmUA[i]=fLB-fUA;
	}
	for (i=0; i<3; i++)
	{
		fP=(float)(frustum_rdist*fabs(afNBmUC[i])+
			frustum_udist*fabs(afNAmLC[i]));
		fTmp=_near_plane*afLBmUA[i];
		fMin=fTmp-fP;
		if (fMin<0.0f) fMin*=frustum_d_ratio;
		fMax=fTmp+fP;
		if (fMax>0.0f) fMax*=frustum_d_ratio;
		fDdD=-afD[0]*afNBmUC[i]+afD[1]*afNAmLC[i]+afD[2]*afLBmUA[i];
		fR=0.0f;
		for (j=0; j<3; j++)
		{
			fR+=(float)(box_extents[j]*fabs(-afA[j]*afNBmUC[i]+afB[j]*afNAmLC[i]+afC[j]*afLBmUA[i]));
		}
		if ((fDdD+fR<fMin) || (fDdD-fR>fMax)) return false;
	}

	// M = Cross(n*D+l*L-u*U,A[i])
	for (i=0; i<3; i++)
	{
		fP=(float)(frustum_rdist*fabs(afNBpUC[i])+
			frustum_udist*fabs(afNAmLC[i]));
		fTmp=_near_plane*afLBpUA[i];
		fMin=fTmp-fP;
		if (fMin<0.0f) fMin*=frustum_d_ratio;
		fMax=fTmp+fP;
		if (fMax>0.0f) fMax *= frustum_d_ratio;
		fDdD=-afD[0]*afNBpUC[i]+afD[1]*afNAmLC[i]+afD[2]*afLBpUA[i];
		fR=0.0f;
		for (j=0; j<3; j++)
		{
			fR+=(float)(box_extents[j]*fabs(-afA[j]*afNBpUC[i]+afB[j]*afNAmLC[i]+afC[j]*afLBpUA[i]));
		}
		if ((fDdD+fR<fMin) || (fDdD-fR>fMax)) return false;
	}

	// M = Cross(n*D-l*L+u*U,A[i])
	for (i=0; i<3; i++)
	{
		fP=(float)(frustum_rdist*fabs(afNBmUC[i])+
			frustum_udist*fabs(afNApLC[i]));
		fTmp=-_near_plane*afLBpUA[i];
		fMin=fTmp-fP;
		if (fMin<0.0f) fMin*=frustum_d_ratio;
		fMax=fTmp+fP;
		if (fMax>0.0f) fMax*=frustum_d_ratio;
		fDdD=-afD[0]*afNBmUC[i]+afD[1]*afNApLC[i]-afD[2]*afLBpUA[i];
		fR=0.0f;
		for (j=0; j<3; j++)
		{
			fR+=(float)(box_extents[j]*fabs(-afA[j]*afNBmUC[i]+afB[j]*afNApLC[i]-afC[j]*afLBpUA[i]));
		}
		if ((fDdD+fR<fMin) || (fDdD-fR>fMax)) return false;
	}

	// M = Cross(n*D-l*L-u*U,A[i])
	for (i=0; i<3; i++)
	{
		fP=(float)(frustum_rdist*fabs(afNBpUC[i])+
			frustum_udist*fabs(afNApLC[i]));
		fTmp=-_near_plane*afLBmUA[i];
		fMin=fTmp - fP;
		if (fMin<0.0f) fMin*=frustum_d_ratio;
		fMax=fTmp+fP;
		if (fMax>0.0f) fMax*=frustum_d_ratio;
		fDdD=-afD[0]*afNBpUC[i]+afD[1]*afNApLC[i]-afD[2]*afLBmUA[i];
		fR=0.0f;
		for (j=0; j<3; j++)
		{
			fR+=(float)(box_extents[j]*fabs(-afA[j]*afNBpUC[i]+afB[j]*afNApLC[i]-afC[j]*afLBmUA[i]));
		}
		if ((fDdD+fR<fMin) || (fDdD-fR>fMax)) return false;
	}

	return true;
}


void Camera::draw()
{
	if (_link)
	{
		_pos=_link->_pos;
		_ori=_link->_ori;
	}

	GFX->set_world_transformation(0,_pos.x,_pos.y,_pos.z,_ori,_scale.x,_scale.y,_scale.z);
}
