#include "FSEffect.h"
#include "TextureManager.h"
#include "Utils.h"

FSEffect::FSEffect(const std::string& effect_object_name,const int n_stages,const bool own_z,const float scale_rt)
{
	_n_stages=n_stages;
	_rt=new TextureID[n_stages];
	_rt_surface=new SurfaceID[n_stages];
	_rt_ids=new int[n_stages];

	Vec2f r=GFX->get_real_resolution(); 
	for (int i=0; i<n_stages; i++)
	{
		_rt[i]=GFX->new_texture((int)r.x,(int)r.y,D3DFMT_A8R8G8B8,true);
		_rt_surface[i]=GFX->get_surface(_rt[i]);
		_rt_ids[i]=TEXTURES->add(string_format("%s_%i",effect_object_name.c_str(),i).c_str(),_rt[i]);
	}
	if (own_z) _zbuffer=GFX->create_z_buffer((int)r.x,(int)r.y,D3DFMT_A8R8G8B8);
	else _zbuffer=NULL;

	_current_index=0;

	// Build mesh
	auto mesh=new Mesh<VertexPNCCT>(false);
	mesh->alloc_vert(4);
	mesh->alloc_index(6);

	VertexPNCCT* dest=mesh->get_write_pointer_v();
	dest->pos.set(-1.0f,-1.0f,0.5f); dest->normal.set(0.0f,0.0f,0.0f); dest->color1=dest->color2=0xFFFFFFFF; dest->uv.set(0.0f+0.375f/r.x,1.0f+0.375f/r.y); dest++;
	dest->pos.set( 1.0f,-1.0f,0.5f); dest->normal.set(0.0f,0.0f,0.0f); dest->color1=dest->color2=0xFFFFFFFF; dest->uv.set(1.0f+0.375f/r.x,1.0f+0.375f/r.y); dest++;
	dest->pos.set( 1.0f, 1.0f,0.5f); dest->normal.set(0.0f,0.0f,0.0f); dest->color1=dest->color2=0xFFFFFFFF; dest->uv.set(1.0f+0.375f/r.x,0.0f+0.375f/r.y); dest++;
	dest->pos.set(-1.0f, 1.0f,0.5f); dest->normal.set(0.0f,0.0f,0.0f); dest->color1=dest->color2=0xFFFFFFFF; dest->uv.set(0.0f+0.375f/r.x,0.0f+0.375f/r.y); dest++;
	WORD* idest=mesh->get_write_pointer_i();
	*idest=0; idest++;
	*idest=1; idest++;
	*idest=2; idest++;
	*idest=0; idest++;
	*idest=2; idest++;
	*idest=3; idest++;
	mesh->close(true);

	_mesh=mesh;

	// Build material
	_material=new Material;
	_material->set_cull_mode(D3DCULL_NONE);
	_material->set_z_write(false);
	for (int i=0; i<_n_stages; i++) _material->set_texture_sampling(i,D3DTEXF_POINT);
	
	_mesh->set_material(_material,true);
}

FSEffect::~FSEffect(void)
{
	for (int i=0; i<_n_stages; i++)
	{
		TEXTURES->remove(_rt_ids[i]);
		_rt_surface[i]->Release();
		_rt[i]->Release();
	}
	if (_zbuffer) 
	{
		_zbuffer->Release(); 
		_zbuffer=NULL;
	}
	delete[] _rt; _rt=NULL;
	delete[] _rt_surface; _rt_surface=NULL;
	delete[] _rt_ids; _rt_ids=NULL;

	delete _mesh; _mesh=NULL; _material=NULL;
}

void FSEffect::begin()
{
	// Set this as render target
	_current_index=0;
	GFX->set_render_target(0,_rt_surface[0],_zbuffer);
}

void FSEffect::end(const D3DBLEND src_blend,const D3DBLEND dest_blend)
{
	GFX->unset_render_target(0);
	
	_current_index=(_current_index+1)%_n_stages;
	_material->set_alpha(true,src_blend,dest_blend);

	actual_run_effect("fx_passthrough");

	_material->set_alpha(true,D3DBLEND_ONE,D3DBLEND_ZERO);

	GFX->enable_z(true);
}

void FSEffect::run(const std::string& effect_name)
{
	// Push the new render target
	_current_index=(_current_index+1)%_n_stages;
	GFX->set_render_target(0,_rt_surface[_current_index],_zbuffer);

	// Render effect
	actual_run_effect(effect_name);
}

void FSEffect::actual_run_effect(const std::string& effect_name)
{
	// Disable Z test
	GFX->enable_z(false);

	// Setup the "LAST" map on the material
	int n=_current_index-1; 
	if (n<0) n=_n_stages-1;
	for (int i=0; i<_n_stages; i++)
	{
		_material->set_texture(i,_rt_ids[n]);
		n=_current_index-1; if (n<0) n=_n_stages-1;
	}

	// Set the shader on the material
	_material->set_shader(effect_name.c_str());

	// Render effect
	_mesh->draw();
}

void FSEffect::set_constant(const std::string& constant_name,const Vec4f& val)
{
	_material->add_custom_param(constant_name,val);
}
