#include "Font.h"
#include "Gfx.h"
#include "BlitManager.h"
#include "Image_A8R8G8B8.h"
#include "TextureManager.h"
#include "Utils.h"

#ifdef USE_FREETYPE2
FT_Library	Font::g_freetype2_lib=NULL;
int			Font::g_freetype2_lib_ref_count=0;
#endif

Font::Font()
{
#ifdef USE_D3DFONT
	_font=NULL;
	_sprite=NULL;
#endif

#ifdef USE_FREETYPE2
	if (!g_freetype2_lib)
	{
		FT_Error error=FT_Init_FreeType(&g_freetype2_lib);
		if (error)
		{
			ERROR_MSG("Can't initialize FreeType2!");
		}
	}
	g_freetype2_lib_ref_count++;
	_font=NULL;
	_texture=-1;
	_texture_external=false;
#endif
	_text_outline_color.set(1.0f,1.0f,0.0f,1.0f);
	_text_background_color.set(1.0f,0.0f,0.0f,1.0f);
	_text_background_margin_x=0;
	_text_background_margin_y=0;
	_text_outline_margin_x=0;
	_text_outline_margin_y=0;
	_line_spacing=4.0f;
	_shadow_offset=1;
}

Font::~Font(void)
{
#ifdef USE_D3DFONT
	if (_font)
	{
		_font->Release();
		_font=NULL;
	}
	if (_sprite) 
	{
		_sprite->Release(); 
		_sprite=NULL;
	}
#endif

#ifdef USE_FREETYPE2
	FT_Done_Face(_font);
	_font=NULL;

	g_freetype2_lib_ref_count--;
	if (g_freetype2_lib_ref_count==0)
	{
		FT_Done_FreeType(g_freetype2_lib);
		g_freetype2_lib=NULL;
	}
	if (!_texture_external)
	{
		if (_texture!=-1)
		{
			TEXTURES->clear(_texture);
			_texture=-1;
		}
	}
#endif
}

bool Font::init(const int size,const char* name)
{
	_font_name=name;
	_font_size=size;

#ifdef USE_D3DFONT
	if (GFX->create_font(size,0,FW_BOLD,0,FALSE,ANSI_CHARSET,OUT_DEFAULT_PRECIS,DEFAULT_QUALITY,DEFAULT_PITCH | FF_DONTCARE,name,&_font)!=D3D_OK)
	{
		return false;
	}

	if (D3DXCreateSprite(GFX->get_device(),&_sprite)!=D3D_OK)
	{
		_font->Release();
		_font=NULL;
		return false;
	}
#endif

#ifdef USE_FREETYPE2
	char			buffer[512];
	sprintf_s((char*)&buffer,512,"%s:%i",name,size);
	_texture=TEXTURES->find_by_name(buffer);
	if (_texture!=-1) _texture_external=true;
	else _texture_external=false;

	std::string		filename=name;
	FILE*			file;
	fopen_s(&file,filename.c_str(),"rb");
	if (!file)
	{
		filename="fonts/"; filename+=name;
		fopen_s(&file,filename.c_str(),"rb");
		if (!file)
		{
			filename="fonts/"; filename+=name; filename+=".ttf";
			fopen_s(&file,filename.c_str(),"rb");
			if (!file)
			{
				filename="fonts/"; filename+=name; filename+=".otf";
				fopen_s(&file,filename.c_str(),"rb");
			}
		}
	}
	if (file) fclose(file);
	else
	{
		sprintf_s((char*)&buffer,512,"Can't open font \"%s\"!",name);
		ERROR_MSG(buffer);
		return false;
	}

	FT_Error error=FT_New_Face(g_freetype2_lib,filename.c_str(),0,&_font);
	if (error==FT_Err_Unknown_File_Format)
	{
		sprintf_s((char*)&buffer,512,"Unsupported font format \"%s\"!",name);
		ERROR_MSG(buffer);
		return false;
	}
	else if (error)
	{
		sprintf_s((char*)&buffer,512,"Error loading font \"%s\"!",name);
		ERROR_MSG(buffer);
		return false;
	}

	error=FT_Set_Pixel_Sizes(_font,0,size);
	if (error)
	{
		sprintf_s((char*)&buffer,512,"Error setting font \"%s\"'s size!",name);
		ERROR_MSG(buffer);
		return false;
	}

	// Build 1024x1024 atlas
	if (_texture_external)
	{
		int atlas_size=TEXTURES->get_texture_size_x(_texture);
		int atlas_margin=4;

		render_font_atlas(atlas_size,atlas_margin);
	}
	else
	{
		int atlas_size=256;
		int atlas_margin=4;

		while (!render_font_atlas(atlas_size,atlas_margin))
		{
			atlas_size*=2;
			if (atlas_size>2048)
			{
				return false;
			}
		}
	}

#endif
	return true;
}

#ifdef USE_FREETYPE2
bool Font::render_font_atlas(const int atlas_size,const int atlas_margin)
{
	Image_A8R8G8B8* img=NULL;

	if (!_texture_external)
	{
		img=new Image_A8R8G8B8;
		img->create(atlas_size,atlas_size);
		img->clear(1.0f,1.0f,1.0f,0.0f);
	}

	int atlas_x=atlas_margin;
	int atlas_y=atlas_margin;
	int max_y_line=0;
	_max_height=0;
	_max_top=0;
	for (int i=32; i<256; i++)
	{
		if (FT_Load_Char(_font,i,FT_LOAD_RENDER)==0)
		{
			FT_Render_Glyph(_font->glyph,FT_RENDER_MODE_NORMAL);

			// Copy buffer to the atlas
			if ((atlas_x+_font->glyph->bitmap.width)>atlas_size-atlas_margin)
			{
				atlas_x=atlas_margin;
				atlas_y+=max_y_line+atlas_margin;
				if (_max_height<max_y_line) _max_height=max_y_line;
				max_y_line=0;
			}

			if (max_y_line<_font->glyph->bitmap.rows) max_y_line=_font->glyph->bitmap.rows;

			if (!_texture_external)
			{
				if ((atlas_y+_font->glyph->bitmap.rows)>=atlas_size)
				{
					delete img;
					return false;
				}

				unsigned char*	dest;
				unsigned char*	line_start=_font->glyph->bitmap.buffer;
				unsigned char*	src;
				for (int yy=0; yy<_font->glyph->bitmap.rows; yy++)
				{
					src=line_start;
					dest=((unsigned char*)img->get_data())+(((atlas_y+yy)*atlas_size)+atlas_x)*4;

					for (int xx=0; xx<_font->glyph->bitmap.width; xx++)
					{
						*dest=0xFF; dest++; 
						*dest=0xFF; dest++; 
						*dest=0xFF; dest++; 
						*dest=(*src); dest++; 

						++src;
					}

					line_start=line_start+_font->glyph->bitmap.pitch;
				}
			}

			_uv[i].set((atlas_x)/(float)atlas_size,(atlas_y)/(float)atlas_size,(atlas_x+_font->glyph->bitmap.width+1)/(float)atlas_size,(atlas_y+_font->glyph->bitmap.rows+1)/(float)atlas_size);
			_delta_x[i]=_font->glyph->bitmap_left;
			_delta_y[i]=_font->glyph->bitmap_top;
			_advance_x[i]=_font->glyph->advance.x>>6;
			_advance_y[i]=_font->glyph->advance.y>>6;
			_size_x[i]=_font->glyph->bitmap.width+1;
			_size_y[i]=_font->glyph->bitmap.rows+1;

			if (_max_top<_font->glyph->bitmap_top) _max_top=_font->glyph->bitmap_top;

			atlas_x+=_font->glyph->bitmap.width+atlas_margin;
		}
		else 
		{
			_delta_x[i]=_delta_y[i]=_advance_x[i]=_advance_y[i]=_size_x[i]=_size_y[i]=0;
			_uv[i].set(0.0f,0.0f,0.0f,0.0f);
		}
	}

	if (_max_height<max_y_line) _max_height=max_y_line;

	if (!_texture_external)
	{
		char buffer[512];
		// Load image to texture
		sprintf_s((char*)&buffer,512,"%s:%i",_font_name.c_str(),_font_size);
		_texture=TEXTURES->load(buffer,img);

//		sprintf_s((char*)&buffer,512,"../%s_%i.dds",_font_name.c_str(),_font_size);
//		img->save_dds(buffer);

		delete img;
	}

	return true;
}
#endif

void Font::write_text(const float x,const float y,const float z,const float r,const float g,const float b,const float a,const char *string,const unsigned long flags)
{
#ifdef USE_D3DFONT
	if (!_font) return;

	RECT		rect;
	D3DCOLOR	color;
	Vec3f		c=GFX->get_color_modulation();

	color=D3DCOLOR_COLORVALUE(r*c.x,g*c.y,b*c.z,a);

	rect.left=long(x);
	rect.top=long(y);
	rect.bottom=long(y+1);
	rect.right=long(x+1);

	GFX->begin_scene();

	unsigned long f=DT_LEFT;
	if (flags&TEXT_CENTER) f=DT_CENTER;
	else f=DT_LEFT;

	if (flags&TEXT_BOTTOM) f|=DT_BOTTOM;

	_font->DrawText(NULL,string,-1,&rect,DT_CALCRECT|f,color);

	if (flags&TEXT_BACKGROUND) BLIT->add((float)rect.left-_text_background_margin_x,(float)rect.top-_text_background_margin_y,0.1f,(float)(rect.right-rect.left+_text_background_margin_x*2),(float)(rect.bottom-rect.top+_text_background_margin_y*2),Color(_text_background_color.r,_text_background_color.g,_text_background_color.b,_text_background_color.a));
	if (flags&TEXT_OUTLINE) BLIT->add((float)rect.left-_text_outline_margin_x,(float)rect.top-_text_outline_margin_y,0.11f,(float)(rect.right-rect.left+_text_outline_margin_x*2),(float)(rect.bottom-rect.top+_text_outline_margin_y*2),_text_outline_color);
	if (flags&TEXT_BACKGROUND) BLIT->add((float)rect.left-_text_background_margin_x,(float)rect.top-_text_background_margin_y,0.1f,(float)(rect.right-rect.left+_text_background_margin_x*2),(float)(rect.bottom-rect.top+_text_background_margin_y*2),_text_background_color);
	if ((flags&TEXT_BACKGROUND) || (flags&TEXT_OUTLINE)) BLIT->flush();

	D3DXMATRIX	matrix;
	_sprite->Begin(D3DXSPRITE_OBJECTSPACE|D3DXSPRITE_ALPHABLEND|D3DXSPRITE_SORT_TEXTURE|D3DXSPRITE_SORT_DEPTH_BACKTOFRONT);
	if (flags&TEXT_SHADOW) 
	{
		RECT rect2=rect;
		rect2.left+=_shadow_offset;
		rect2.right+=_shadow_offset;
		rect2.top+=_shadow_offset;
		rect2.bottom+=_shadow_offset;
		_font->DrawText(_sprite,string,-1,&rect2,f,color&0xFF000000);
	}
	_font->DrawText(_sprite,string,-1,&rect,f,color);
	_sprite->End();
#endif

#ifdef USE_FREETYPE2
	const char* ptr=string;
	const char* next_line=ptr;
	int			n_chars;
	float		px=(float)floorf(x+0.5f);
	float		py=(float)floorf(y+0.5f);
	float		ppx=px;
	float		ppy=py;
	char		c;
	int			baseline_y;
	Vec2f		s=get_text_size(string);

	if (flags&TEXT_BOTTOM)
	{
		py-=s.y;
	}

	Rect		rect(px,py,px+s.x,py+s.y);
	float		text_z=z;
	
	if (flags&TEXT_CENTER) rect.offset(-s.x*0.5f,0.0f);

	if (flags&TEXT_BACKGROUND) BLIT->add((float)rect.x1-_text_background_margin_x,(float)rect.y1-_text_background_margin_y,text_z+0.04f,(float)(s.x+_text_background_margin_x*2),(float)(s.y+_text_background_margin_y*2),Color(_text_background_color.r,_text_background_color.g,_text_background_color.b,_text_background_color.a));
	if (flags&TEXT_OUTLINE) BLIT->add((float)rect.x1-_text_outline_margin_x,(float)rect.y1-_text_outline_margin_y,text_z+0.05f,(float)(s.x+_text_outline_margin_x*2),(float)(s.y+_text_outline_margin_y*2),_text_outline_color);
	if (flags&TEXT_BACKGROUND) BLIT->add((float)rect.x1-_text_background_margin_x,(float)rect.y1-_text_background_margin_y,text_z+0.04f,(float)(s.x+_text_background_margin_x*2),(float)(s.y+_text_background_margin_y*2),_text_background_color);

	while (ptr!=NULL)
	{
		next_line=get_next_line(ptr,n_chars);

		s=get_text_line_size(ptr,n_chars,baseline_y);

		if (flags&TEXT_CENTER) px-=s.x*0.5f;
		py-=baseline_y;

		for (int i=0; i<n_chars; i++)
		{
			c=ptr[i];

			ppx=px+_delta_x[c];
			ppy=(float)(py-_delta_y[c]);
			if (flags&TEXT_SHADOW) BLIT->add(_texture,ppx+_shadow_offset,ppy+_shadow_offset,text_z+0.02f,(float)_size_x[c],(float)_size_y[c],0.0f,0.0f,0.0f,a,_uv[c].x1,_uv[c].y1,_uv[c].x2,_uv[c].y2,D3DTEXF_LINEAR);
			BLIT->add(_texture,ppx,ppy,text_z,(float)_size_x[c],(float)_size_y[c],r,g,b,a,_uv[c].x1,_uv[c].y1,_uv[c].x2,_uv[c].y2,D3DTEXF_LINEAR);

			px+=_advance_x[c];
		}

		py+=s.y+baseline_y+_line_spacing;
		px=(float)floorf(x+0.5f);

		ptr=next_line;
	}
#endif
}

Rect Font::get_text_rect(float x,float y,const char *string,const unsigned long flags)
{
#ifdef USE_D3DFONT
	if (!_font) return Rect(0.0f,0.0f,0.0f,0.0f);

	unsigned long f=DT_LEFT;
	if (flags&TEXT_CENTER) f=DT_CENTER;
	else f=DT_LEFT;

	if (flags&TEXT_BOTTOM) f|=DT_BOTTOM;

	RECT		rect;

	rect.left=long(x);
	rect.top=long(y);
	rect.bottom=long(y+1);
	rect.right=long(x+1);

	_font->DrawText(NULL,string,-1,&rect,DT_CALCRECT|f,0x00000000);

	return Rect((float)rect.left,(float)rect.top,(float)rect.right,(float)rect.bottom);
#endif
#ifdef USE_FREETYPE2
	Vec2f	s=get_text_size(string);
	Rect	r(x,y,x+s.x,y+s.y);

	if (flags&TEXT_CENTER) r.offset(-s.x*0.5f,0.0f);
	if (flags&TEXT_BOTTOM) r.offset(0.0f,-s.y);

	return r;
#endif
	return Rect(0.0f,0.0f,0.0f,0.0f);
}

#ifdef USE_FREETYPE2
const char*	Font::get_next_line(const char* start,int& n_chars)
{
	const char* ptr=start;
	
	n_chars=0;

	while ((*ptr)!=0)
	{
		if ((*ptr)=='\n')
		{
			return ptr+1;
		}
		n_chars++;
		ptr++;
	}

	return NULL;
}

Vec2f Font::get_text_line_size(const char* start,const int n_chars,int& baseline_y)
{
	Vec2f	s(0.0f,0.0f);
	char	c;
	float	dmin=0,dmax=0;
	for (int i=0; i<n_chars; i++)
	{
		c=start[i];

		s.x+=_advance_x[c];

		dmin=Min(dmin,-_delta_y[c]);
		dmax=Max(dmax,_size_y[c]-_delta_y[c]);
	}

	baseline_y=(int)dmin;
	s.y=dmax-dmin;

	return s;
}

Vec2f Font::get_text_size(const char* start)
{
	const char* next_line;
	const char* src=start;
	int			n;
	Vec2f		total_size(0.0f,0.0f);
	Vec2f		s;
	int			bl;
	while (src)
	{
		next_line=get_next_line(src,n);

		s=get_text_line_size(src,n,bl);
		total_size.x=Max(total_size.x,s.x);
		total_size.y+=s.y+((next_line!=NULL)?(_line_spacing):(0));

		src=next_line;
	}

	return total_size;
}
#endif

void Font::make_fixed_width()
{
#ifdef USE_FREETYPE2
	// Get max width
	long	max_w=0;
	int		max_char=0;
	for (int i=32; i<128; i++)
	{
		if (max_w<_advance_x[i])
		{
			max_w=_advance_x[i];
			max_char=i;
		}
	}
	float offset=0.0f;
	for (int i=32; i<256; i++)
	{
		offset=(float)(max_w-(_delta_x[i]+_size_x[i]));
		_delta_x[i]+=(int)floorf(offset*0.5f);
		_advance_x[i]=max_w;
	}
#endif
}
