// Gfx.cpp: implementation of the Gfx class.
//
//////////////////////////////////////////////////////////////////////

#include "Gfx.h"
#include "Defs.h"
#include "Screen.h"
#include "Light.h"
#include "Mesh.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include <stdlib.h>
#include "Utils.h"
#include "Logger.h"

Gfx*		GFX=NULL;

LRESULT CALLBACK WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch (uMsg)
	{
		case WM_ACTIVATE:
		case WM_ACTIVATEAPP:
			{
				MouseEventCallback	callback=get_mouse_event_callback();
				if (callback)
				{
					callback(uMsg,(unsigned long)wParam,(unsigned long)lParam);
				}
			}
			return 0L;
		case WM_CLOSE:
		case WM_DESTROY:
			// Clean up and close the app
			if (SCREEN)
			{
				SCREEN->set_exit(true,-1);
			}
			return 0L;
		case WM_MOUSEMOVE:
		case WM_LBUTTONUP:
		case WM_LBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_RBUTTONDOWN:
			{
				MouseEventCallback	callback=get_mouse_event_callback();
				if (callback)
				{
					callback(uMsg,(unsigned long)wParam,(unsigned long)lParam);
				}
			}
		case WM_ENTERSIZEMOVE:
			// Halt frame movement while the app is sizing or moving
			return 0;
			break;
		case WM_EXITSIZEMOVE:
			return 0;
			break;
		case WM_SETCURSOR:
			SetCursor(NULL);
			return TRUE;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Gfx::Gfx()
{
	if (!GFX) GFX=this;
	_adapter=-1;
	set_window_mode(false);
	set_resolution(1024,768,-1,-1);
	set_window(NULL);
	set_color_modulation(1.0f,1.0f,1.0f); 
	set_app_name("Game");
	for (int i=0; i<256; i++)
	{
		_renderstate[i]=0xFFFFFFFF;
	}
	_renderstate[D3DRS_AMBIENT]=0;
	for (int j=0; j<MAX_TEX; j++)
	{
		for (int i=0; i<256; i++)
		{
			_rss[j][i]=0xFFFFFFFF;
			_sampler[j][i]=0xFFFFFFFF;
		}
	}
	for (int i=0; i<MAX_TEX; i++) _texture[i]=NULL;
	_vertex_shader=(VertexProgram*)0x123456;
	_pixel_shader=(PixelShader*)0x123456;
	_vdecl=(VertexDeclaration*)0x123456;
	_fvf=0;
	_in_scene=false;
	_ambient_light.set(0.0f,0.0f,0.0f,0.0f);
	_refresh_clip_matrix=true;
	_clip_matrix.identity();
	_projection_matrix.identity();
	_view_matrix.identity();
	_world_matrix.identity();
	_pass=PASS_SOLID;
	_set_depth_buffer=NULL;
	_env.reset();
}

Gfx::~Gfx()
{
	if (GFX==this) GFX=NULL;
	shutdown();
}

bool Gfx::init(HINSTANCE instance)
{
    HRESULT hr;

    // Create the Direct3D object
    _d3d9=Direct3DCreate9(D3D_SDK_VERSION);
    if (_d3d9==NULL)
	{
        MessageBox(NULL,"Error creating D3D9 device!",_app_name.c_str(),MB_ICONERROR|MB_OK);
		return false;
	}

    // Build a list of Direct3D adapters, modes and devices. The
    // ConfirmDevice() callback is used to confirm that only devices that
    // meet the app's requirements are considered.
	_d3d_enumeration.AppUsesDepthBuffer=TRUE;
    _d3d_enumeration.SetD3D(_d3d9);
    _d3d_enumeration.ConfirmDeviceCallback=confirm_device_helper;
    if (FAILED(hr=_d3d_enumeration.Enumerate()))
    {
        MessageBox(NULL,"Can't enumerate device!",_app_name.c_str(),MB_ICONERROR|MB_OK);
        SAFE_RELEASE(_d3d9);
        return false;
    }

    // Register the windows class
    WNDCLASS wndClass = { 0, WndProc, 0, 0, instance,
                          LoadIcon( instance, MAKEINTRESOURCE(IDI_ASTERISK)),
                          LoadCursor( instance, MAKEINTRESOURCE(IDC_ARROW) ),
                          (HBRUSH)GetStockObject(WHITE_BRUSH),
                          NULL, _app_name.c_str() };
    RegisterClass(&wndClass);

    // Set the window's initial style
	DWORD	dwWindowStyle;
	int		wpx,wpy;
	if (_window_mode)
	{
		dwWindowStyle=WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | 
                     WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_VISIBLE;
		wpx=wpy=CW_USEDEFAULT;
	}
	else
	{
		dwWindowStyle=WS_EX_TOPMOST | WS_POPUP | WS_VISIBLE;
		wpx=wpy=0;
	}
//    DWORD dwWindowStyle=WS_POPUP|WS_VISIBLE;

    // Set the window's initial width
    RECT rc;
    SetRect(&rc,0,0,int(_resolution.x),int(_resolution.y));
    AdjustWindowRect(&rc,dwWindowStyle,false);

    // Create the render window
    _window= CreateWindow(_app_name.c_str(),_app_name.c_str(),dwWindowStyle,
						  wpx,wpy,
						  (rc.right-rc.left),(rc.bottom-rc.top),0,
                          NULL,instance,0);

    // Save window properties
    _window_style=GetWindowLong(_window,GWL_STYLE);
    GetWindowRect(_window,&_window_bounds);
    GetClientRect(_window,&_window_client);

	ShowWindow(_window,SW_SHOW);
	SetFocus(_window);
	
	RECT rect;
	GetWindowRect(GFX->get_window(),&rect);
	rect.right-=2; rect.left+=2;
	rect.top+=2; rect.bottom-=2;
#ifndef DEV
	ClipCursor(&rect);
#endif

	SetCursor(NULL);

    if (FAILED(hr=choose_initial_settings()))
    {
        MessageBox(NULL,"Can't choose initial D3D settings!",_app_name.c_str(),MB_ICONERROR|MB_OK);
        SAFE_RELEASE(_d3d9);
        return false;
    }

    // Initialize the 3D environment for the app
    if (FAILED(hr=init_d3d_environment()))
    {
        SAFE_RELEASE(_d3d9);
        MessageBox(NULL,"Can't initialize D3D environment!",_app_name.c_str(),MB_ICONERROR|MB_OK);
        return false;
    }

	set_renderstate(D3DRS_ALPHABLENDENABLE,TRUE);
	set_renderstate(D3DRS_LIGHTING,1);
	set_renderstate(D3DRS_ZENABLE,1);
	set_renderstate(D3DRS_ZWRITEENABLE,1);

    return true;
}

void Gfx::shutdown()
{
#ifdef USE_D3D9
	if (_vertex_shader) _device->SetVertexShader(NULL);
	if (_pixel_shader) _device->SetPixelShader(NULL);
	if (_vdecl) _device->SetVertexDeclaration(NULL);
	for (int i=0; i<MAX_TEX; i++)
	{
		_device->SetTexture(i,NULL);
	}
#ifdef USE_SHADERS
	Material::clear_shaders();
#endif
	SAFE_RELEASE(_device);
	SAFE_RELEASE(_d3d9);
#endif
}

bool Gfx::confirm_device_helper(D3DCAPS9* pCaps,VertexProcessingType vertexProcessingType, 
							   D3DFORMAT backBufferFormat,D3DFORMAT format)
{
    DWORD dwBehavior;

    if (vertexProcessingType == SOFTWARE_VP)
        dwBehavior = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    else if (vertexProcessingType == MIXED_VP)
        dwBehavior = D3DCREATE_MIXED_VERTEXPROCESSING;
    else if (vertexProcessingType == HARDWARE_VP)
        dwBehavior = D3DCREATE_HARDWARE_VERTEXPROCESSING;
    else if (vertexProcessingType == PURE_HARDWARE_VP)
        dwBehavior = D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE;
    else
        dwBehavior = 0; // TODO: throw exception
    
    return SUCCEEDED(GFX->confirm_device(pCaps,dwBehavior,backBufferFormat));
}

void Gfx::set_cursor_position(long x,long y)
{
	_device->SetCursorPosition(x,y,0);
}

HRESULT Gfx::choose_initial_settings()
{
	if (_window_mode)
	{
		if (!find_best_windowed_mode(false,false))
		{
			return E_FAIL;
		}
	}
	else
	{
		if (!find_best_fullscreen_mode(false,false))
		{
			return E_FAIL;
		}
	}

	return S_OK;
}

HRESULT Gfx::init_d3d_environment()
{
    HRESULT hr;

    D3DAdapterInfo* pAdapterInfo = _d3d_settings.PAdapterInfo();
    D3DDeviceInfo* pDeviceInfo = _d3d_settings.PDeviceInfo();

    _window_mode=_d3d_settings.IsWindowed;

    // Set up the presentation parameters
	build_present_params_from_settings();

    if( pDeviceInfo->Caps.PrimitiveMiscCaps & D3DPMISCCAPS_NULLREFERENCE )
    {
        // Warn user about null ref device that can't render anything
		MessageBox(NULL,"No device found!",_app_name.c_str(),MB_ICONERROR|MB_OK);
    }

    DWORD behaviorFlags;
    if (_d3d_settings.GetVertexProcessingType() == SOFTWARE_VP)
        behaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    else if (_d3d_settings.GetVertexProcessingType() == MIXED_VP)
        behaviorFlags = D3DCREATE_MIXED_VERTEXPROCESSING;
    else if (_d3d_settings.GetVertexProcessingType() == HARDWARE_VP)
        behaviorFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;
    else if (_d3d_settings.GetVertexProcessingType() == PURE_HARDWARE_VP)
        behaviorFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE;
    else
        behaviorFlags = 0; // TODO: throw exception

    // Create the device
	_d3dpp.PresentationInterval=D3DPRESENT_INTERVAL_ONE;
//	_d3dpp.PresentationInterval=D3DPRESENT_INTERVAL_IMMEDIATE;
//	_d3dpp.Flags=0;
//	_d3dpp.AutoDepthStencilFormat=D3DFMT_D24S8;
    hr = _d3d9->CreateDevice( _d3d_settings.AdapterOrdinal(), pDeviceInfo->DevType,
                               _window, behaviorFlags, &_d3dpp,
                               &_device );

    if( SUCCEEDED(hr) )
    {
        // When moving from fullscreen to windowed mode, it is important to
        // adjust the window size after recreating the device rather than
        // beforehand to ensure that you get the window size you want.  For
        // example, when switching from 640x480 fullscreen to windowed with
        // a 1000x600 window on a 1024x768 desktop, it is impossible to set
        // the window size to 1000x600 until after the display mode has
        // changed to 1024x768, because windows cannot be larger than the
        // desktop.

	    if( _window_mode )
        {
            SetWindowPos( _window, HWND_NOTOPMOST,
                          _window_bounds.left, _window_bounds.top,
                          ( _window_bounds.right - _window_bounds.left ),
                          ( _window_bounds.bottom - _window_bounds.top ),
                          SWP_SHOWWINDOW );
        }

		if (!_window_mode)
		{
			MSG msg;
			unsigned long t0=timeGetTime();
			while ((timeGetTime()-t0)<1000)
			{
				while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}

			_device->Reset(&_d3dpp);
		}

		// Store device Caps
        _device->GetDeviceCaps(&_caps);
        _create_flags=behaviorFlags;

        // Store render target surface desc
        LPDIRECT3DSURFACE9 pBackBuffer = NULL;
        _device->GetBackBuffer( 0, 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
        pBackBuffer->GetDesc(&_back_buffer_desc);
        pBackBuffer->Release();

		_default_render_target=NULL;
		_default_depth_buffer=NULL;
		_device->GetRenderTarget(0,&_default_render_target);
		_device->GetDepthStencilSurface(&_default_depth_buffer);


		return S_OK;
    }

    // If that failed, fall back to the reference rasterizer
    if (pDeviceInfo->DevType==D3DDEVTYPE_HAL)
    {
        if (find_best_windowed_mode(false,true))
        {
            _window_mode = true;
            // Make sure main window isn't topmost, so error message is visible
            SetWindowPos( _window, HWND_NOTOPMOST,
                          _window_bounds.left, _window_bounds.top,
                          ( _window_bounds.right - _window_bounds.left ),
                          ( _window_bounds.bottom - _window_bounds.top ),
                          SWP_SHOWWINDOW );

            // Let the user know we are switching from HAL to the reference rasterizer
            MessageBox(NULL,"No HAL, switching to REF!",_app_name.c_str(),MB_ICONERROR|MB_OK);

            hr=init_d3d_environment();
        }
    }
    return hr;
}

HRESULT Gfx::confirm_device(D3DCAPS9* pCaps,DWORD dwBehavior,D3DFORMAT Format)
{
    return S_OK;
}

bool Gfx::find_best_windowed_mode(bool bRequireHAL,bool bRequireREF)
{
    // Get display mode of primary adapter (which is assumed to be where the window 
    // will appear)
    D3DDISPLAYMODE primaryDesktopDisplayMode;
    _d3d9->GetAdapterDisplayMode(0, &primaryDesktopDisplayMode);

    D3DAdapterInfo* pBestAdapterInfo = NULL;
    D3DDeviceInfo* pBestDeviceInfo = NULL;
    D3DDeviceCombo* pBestDeviceCombo = NULL;

    for( UINT iai = 0; iai < _d3d_enumeration.m_pAdapterInfoList->Count(); iai++ )
    {
        D3DAdapterInfo* pAdapterInfo = (D3DAdapterInfo*)_d3d_enumeration.m_pAdapterInfoList->GetPtr(iai);
        for( UINT idi = 0; idi < pAdapterInfo->pDeviceInfoList->Count(); idi++ )
        {
            D3DDeviceInfo* pDeviceInfo = (D3DDeviceInfo*)pAdapterInfo->pDeviceInfoList->GetPtr(idi);
            if (bRequireHAL && pDeviceInfo->DevType != D3DDEVTYPE_HAL)
                continue;
            if (bRequireREF && pDeviceInfo->DevType != D3DDEVTYPE_REF)
                continue;
            for( UINT idc = 0; idc < pDeviceInfo->pDeviceComboList->Count(); idc++ )
            {
                D3DDeviceCombo* pDeviceCombo = (D3DDeviceCombo*)pDeviceInfo->pDeviceComboList->GetPtr(idc);
                bool bAdapterMatchesBB = (pDeviceCombo->BackBufferFormat == pDeviceCombo->AdapterFormat);
                if (!pDeviceCombo->IsWindowed)
                    continue;
                if (pDeviceCombo->AdapterFormat != primaryDesktopDisplayMode.Format)
                    continue;
                // If we haven't found a compatible DeviceCombo yet, or if this set
                // is better (because it's a HAL, and/or because formats match better),
                // save it
                if( pBestDeviceCombo == NULL || 
                    pBestDeviceCombo->DevType != D3DDEVTYPE_HAL && pDeviceCombo->DevType == D3DDEVTYPE_HAL ||
                    pDeviceCombo->DevType == D3DDEVTYPE_HAL && bAdapterMatchesBB )
                {
                    pBestAdapterInfo = pAdapterInfo;
                    pBestDeviceInfo = pDeviceInfo;
                    pBestDeviceCombo = pDeviceCombo;
                    if( pDeviceCombo->DevType == D3DDEVTYPE_HAL && bAdapterMatchesBB )
                    {
                        // This windowed device combo looks great -- take it
                        goto EndWindowedDeviceComboSearch;
                    }
                    // Otherwise keep looking for a better windowed device combo
                }
            }
        }
    }
EndWindowedDeviceComboSearch:
    if (pBestDeviceCombo == NULL )
        return false;

    _d3d_settings.pWindowed_AdapterInfo = pBestAdapterInfo;
    _d3d_settings.pWindowed_DeviceInfo = pBestDeviceInfo;
    _d3d_settings.pWindowed_DeviceCombo = pBestDeviceCombo;
    _d3d_settings.IsWindowed = true;
    _d3d_settings.Windowed_DisplayMode = primaryDesktopDisplayMode;
    _d3d_settings.Windowed_Width = _window_client.right - _window_client.left;
    _d3d_settings.Windowed_Height = _window_client.bottom - _window_client.top;
    if (_d3d_enumeration.AppUsesDepthBuffer)
        _d3d_settings.Windowed_DepthStencilBufferFormat = *(D3DFORMAT*)pBestDeviceCombo->pDepthStencilFormatList->GetPtr(0);
    _d3d_settings.Windowed_MultisampleType = *(D3DMULTISAMPLE_TYPE*)pBestDeviceCombo->pMultiSampleTypeList->GetPtr(0);
    _d3d_settings.Windowed_MultisampleQuality = 0;
    _d3d_settings.Windowed_VertexProcessingType = *(VertexProcessingType*)pBestDeviceCombo->pVertexProcessingTypeList->GetPtr(0);
    _d3d_settings.Windowed_PresentInterval = *(UINT*)pBestDeviceCombo->pPresentIntervalList->GetPtr(0);
    return true;
}

bool Gfx::find_best_fullscreen_mode(bool bRequireHAL,bool bRequireREF)
{
    // For fullscreen, default to first HAL DeviceCombo that supports the current desktop 
    // display mode, or any display mode if HAL is not compatible with the desktop mode, or 
    // non-HAL if no HAL is available
    D3DDISPLAYMODE adapterDesktopDisplayMode;
    D3DDISPLAYMODE bestAdapterDesktopDisplayMode;
    D3DDISPLAYMODE bestDisplayMode;
    bestAdapterDesktopDisplayMode.Width = 0;
    bestAdapterDesktopDisplayMode.Height = 0;
    bestAdapterDesktopDisplayMode.Format = D3DFMT_UNKNOWN;
    bestAdapterDesktopDisplayMode.RefreshRate = 0;

    D3DAdapterInfo* pBestAdapterInfo = NULL;
    D3DDeviceInfo* pBestDeviceInfo = NULL;
    D3DDeviceCombo* pBestDeviceCombo = NULL;

    for( UINT iai = 0; iai < _d3d_enumeration.m_pAdapterInfoList->Count(); iai++ )
    {
		if ((_adapter!=-1) && (iai!=_adapter)) continue;

        D3DAdapterInfo* pAdapterInfo = (D3DAdapterInfo*)_d3d_enumeration.m_pAdapterInfoList->GetPtr(iai);
        _d3d9->GetAdapterDisplayMode( pAdapterInfo->AdapterOrdinal, &adapterDesktopDisplayMode );
        for( UINT idi = 0; idi < pAdapterInfo->pDeviceInfoList->Count(); idi++ )
        {
            D3DDeviceInfo* pDeviceInfo = (D3DDeviceInfo*)pAdapterInfo->pDeviceInfoList->GetPtr(idi);
            if (bRequireHAL && pDeviceInfo->DevType != D3DDEVTYPE_HAL)
                continue;
            if (bRequireREF && pDeviceInfo->DevType != D3DDEVTYPE_REF)
                continue;
            for( UINT idc = 0; idc < pDeviceInfo->pDeviceComboList->Count(); idc++ )
            {
                D3DDeviceCombo* pDeviceCombo = (D3DDeviceCombo*)pDeviceInfo->pDeviceComboList->GetPtr(idc);
                bool bAdapterMatchesBB = (pDeviceCombo->BackBufferFormat == pDeviceCombo->AdapterFormat);
                bool bAdapterMatchesDesktop = (pDeviceCombo->AdapterFormat == adapterDesktopDisplayMode.Format);
                if (pDeviceCombo->IsWindowed)
                    continue;

				if (_bitdepth!=-1)
				{
					switch (_bitdepth)
					{
						case 16: if (pDeviceCombo->BackBufferFormat!=D3DFMT_R5G6B5) continue; break;
						case 32: if (pDeviceCombo->BackBufferFormat!=D3DFMT_X8R8G8B8) continue; break;
					}
				}

                // If we haven't found a compatible set yet, or if this set
                // is better (because it's a HAL, and/or because formats match better),
                // save it
                if (pBestDeviceCombo == NULL ||
                    pBestDeviceCombo->DevType != D3DDEVTYPE_HAL && pDeviceInfo->DevType == D3DDEVTYPE_HAL ||
                    pDeviceCombo->DevType == D3DDEVTYPE_HAL && pBestDeviceCombo->AdapterFormat != adapterDesktopDisplayMode.Format && bAdapterMatchesDesktop ||
                    pDeviceCombo->DevType == D3DDEVTYPE_HAL && bAdapterMatchesDesktop && bAdapterMatchesBB )
                {
                    bestAdapterDesktopDisplayMode = adapterDesktopDisplayMode;
                    pBestAdapterInfo = pAdapterInfo;
                    pBestDeviceInfo = pDeviceInfo;
                    pBestDeviceCombo = pDeviceCombo;
                    if (pDeviceInfo->DevType == D3DDEVTYPE_HAL && bAdapterMatchesDesktop && bAdapterMatchesBB)
                    {
                        // This fullscreen device combo looks great -- take it
                        goto EndFullscreenDeviceComboSearch;
                    }
                    // Otherwise keep looking for a better fullscreen device combo
                }
            }
        }
    }
EndFullscreenDeviceComboSearch:
    if (pBestDeviceCombo == NULL)
        return false;

    // Need to find a display mode on the best adapter that uses pBestDeviceCombo->AdapterFormat
    // and is as close to bestAdapterDesktopDisplayMode's res as possible
    bestDisplayMode.Width = 0;
    bestDisplayMode.Height = 0;
    bestDisplayMode.Format = D3DFMT_UNKNOWN;
    bestDisplayMode.RefreshRate = 0;
    for( UINT idm = 0; idm < pBestAdapterInfo->pDisplayModeList->Count(); idm++ )
    {
        D3DDISPLAYMODE* pdm = (D3DDISPLAYMODE*)pBestAdapterInfo->pDisplayModeList->GetPtr(idm);
		if (_bitdepth!=-1)
		{
			switch (_bitdepth)
			{
				case 16: if (pdm->Format!=D3DFMT_R5G6B5) continue; break;
				case 32: if (pdm->Format!=D3DFMT_X8B8G8R8) continue; break;
			}
		}
		if (_refresh_rate!=-1)
		{
			if (_refresh_rate!=pdm->RefreshRate) continue;
		}
		if (_resolution.x!=-1)
		{
			if (_resolution.x!=pdm->Width) continue;
		}
		if (_resolution.y!=-1)
		{
			if (_resolution.y!=pdm->Height) continue;
		}
		bestDisplayMode=*pdm;
/*        if( pdm->Format != pBestDeviceCombo->AdapterFormat )
            continue;
        if( pdm->Width == bestAdapterDesktopDisplayMode.Width &&
            pdm->Height == bestAdapterDesktopDisplayMode.Height && 
            pdm->RefreshRate == bestAdapterDesktopDisplayMode.RefreshRate )
        {
            // found a perfect match, so stop
            bestDisplayMode = *pdm;
            break;
        }
        else if( pdm->Width == bestAdapterDesktopDisplayMode.Width &&
                 pdm->Height == bestAdapterDesktopDisplayMode.Height && 
                 pdm->RefreshRate > bestDisplayMode.RefreshRate )
        {
            // refresh rate doesn't match, but width/height match, so keep this
            // and keep looking
            bestDisplayMode = *pdm;
        }
        else if( pdm->Width == bestAdapterDesktopDisplayMode.Width )
        {
            // width matches, so keep this and keep looking
            bestDisplayMode = *pdm;
        }
        else if( bestDisplayMode.Width == 0 )
        {
            // we don't have anything better yet, so keep this and keep looking
            bestDisplayMode = *pdm;
        }*/
    }

    _d3d_settings.pFullscreen_AdapterInfo = pBestAdapterInfo;
    _d3d_settings.pFullscreen_DeviceInfo = pBestDeviceInfo;
    _d3d_settings.pFullscreen_DeviceCombo = pBestDeviceCombo;
    _d3d_settings.IsWindowed = false;
    _d3d_settings.Fullscreen_DisplayMode = bestDisplayMode;
    if (_d3d_enumeration.AppUsesDepthBuffer)
        _d3d_settings.Fullscreen_DepthStencilBufferFormat = *(D3DFORMAT*)pBestDeviceCombo->pDepthStencilFormatList->GetPtr(0);
    _d3d_settings.Fullscreen_MultisampleType = *(D3DMULTISAMPLE_TYPE*)pBestDeviceCombo->pMultiSampleTypeList->GetPtr(0);
    _d3d_settings.Fullscreen_MultisampleQuality = 0;
    _d3d_settings.Fullscreen_VertexProcessingType = *(VertexProcessingType*)pBestDeviceCombo->pVertexProcessingTypeList->GetPtr(0);
    _d3d_settings.Fullscreen_PresentInterval = D3DPRESENT_INTERVAL_DEFAULT;
    return true;
}

void Gfx::build_present_params_from_settings()
{
    _d3dpp.Windowed               = _d3d_settings.IsWindowed;
    _d3dpp.BackBufferCount        = 1;
    _d3dpp.MultiSampleType        = _d3d_settings.MultisampleType();
    _d3dpp.MultiSampleQuality     = _d3d_settings.MultisampleQuality();
    _d3dpp.SwapEffect             = D3DSWAPEFFECT_DISCARD;
    _d3dpp.EnableAutoDepthStencil = _d3d_enumeration.AppUsesDepthBuffer;
    _d3dpp.hDeviceWindow          = _window;
    if( _d3d_enumeration.AppUsesDepthBuffer )
    {
        _d3dpp.Flags              = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
        _d3dpp.AutoDepthStencilFormat = _d3d_settings.DepthStencilBufferFormat();
    }
    else
    {
        _d3dpp.Flags              = 0;
    }

    if( _window_mode )
    {
        _d3dpp.BackBufferWidth=_window_client.right - _window_client.left;
        _d3dpp.BackBufferHeight=_window_client.bottom - _window_client.top;
        _d3dpp.BackBufferFormat=_d3d_settings.PDeviceCombo()->BackBufferFormat;
        _d3dpp.FullScreen_RefreshRateInHz=0;
        _d3dpp.PresentationInterval=_d3d_settings.PresentInterval();
    }
    else
    {
        _d3dpp.BackBufferWidth=(UINT)_resolution.x;
        _d3dpp.BackBufferHeight=(UINT)_resolution.y;
        _d3dpp.BackBufferFormat=_d3d_settings.PDeviceCombo()->BackBufferFormat;
        _d3dpp.FullScreen_RefreshRateInHz=_d3d_settings.Fullscreen_DisplayMode.RefreshRate;
        _d3dpp.PresentationInterval=_d3d_settings.PresentInterval();
    }
}

void Gfx::swap_buffers()
{
	HRESULT hRet=_device->Present(NULL,NULL,NULL,NULL);
	if (hRet!=D3D_OK)
	{
		switch (hRet)
		{
			case D3DERR_INVALIDCALL: 
				log("Couldn't swap buffers - Invalid call!");
				break;
			case D3DERR_DEVICELOST: 
				// Check if we have focus
				if (SCREEN)
				{
					while (!SCREEN->is_anim_enable())
					{
						MSG msg;
						while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
						{
							TranslateMessage(&msg);
							DispatchMessage(&msg);
						}
						Sleep(0);
					}
					// Clear everything
					for (int i=0; i<8; i++) _device->SetTexture(i,NULL);
					_default_render_target->Release();
					_default_depth_buffer->Release();
					hRet=_device->Reset(&_d3dpp);
					if (hRet!=D3D_OK)
					{
						__asm int 3
					}
				}
				break;
		}
	}
}

void Gfx::set_2d_render()
{
	_projection_matrix.identity();
	_projection_matrix._m[0][0]=2.0f/_vres.x;
	_projection_matrix._m[1][1]=-2.0f/_vres.y;
	_projection_matrix._m[3][0]=-1.0f;
	_projection_matrix._m[3][1]=1.0f;

	_world_matrix.identity();

	_refresh_clip_matrix=true;
#ifndef USE_SHADERS
//	D3DXMATRIX	persp( 2.0f/_vres.x, 0.0f, 0.0f, 0.0f,
//					   0.0f,-2.0f/_vres.y, 0.0f, 0.0f,
//					   0.0f, 0.0f, 1.0f, 0.0f,
//					  -1.0f, 1.0f, 0.0f, 1.0f);

	_device->SetTransform(D3DTS_PROJECTION,(D3DMATRIX*)&_projection_matrix);	

//	D3DXMATRIX	identity;
	
//	D3DXMatrixIdentity(&identity);
//	_device->SetTransform(D3DTS_WORLDMATRIX(0),&identity);
	GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0.0f,0.0f,0.0f,1.0f));
#endif
}

void Gfx::clear(DWORD clear_buffers,float r,float g,float b,float a,float zVal,unsigned long sVal)
{
	D3DVIEWPORT9 viewport;
	viewport.X=0; viewport.Y=0;
	viewport.Width=(DWORD)_resolution.x;
	viewport.Height=(DWORD)_resolution.y;
	viewport.MinZ=0.0f;
	viewport.MaxZ=1.0f;

	_device->SetViewport(&viewport);

	HRESULT		hRet=_device->Clear(0,NULL,clear_buffers,D3DCOLOR_COLORVALUE(r,g,b,a),zVal,sVal);

	if (hRet!=D3D_OK)
	{
		ERROR_MSG("Can't clear buffers!");
	}
}

void Gfx::begin_scene()
{
	if (_in_scene) return;
	HRESULT		hRet=_device->BeginScene();
	_in_scene=true;
}

void Gfx::end_scene()
{
	if (!_in_scene) return;
	HRESULT		hRet=_device->EndScene();
	_in_scene=false;
}

void Gfx::set_camera2d(float x,float y,float z)
{
	Quaternion	ori(0.0f,0.0f,0.0f,1.0f);
	Vec3f		pos(x/_resolution.x,y/_resolution.y,z); 
	pos=-pos; 
	ori.rotate_point(pos,pos);

	float xx=ori.x*ori.x;
	float xy=ori.x*ori.y;
	float xz=ori.x*ori.z;
	float xw=ori.x*ori.w;

	float yy=ori.y*ori.y;
	float yz=ori.y*ori.z;
	float yw=ori.y*ori.w;

	float zz=ori.z*ori.z;
	float zw=ori.z*ori.w;

	float ww=ori.w*ori.w;

	_view_matrix._m[0][0]=1.0f-2.0f*(yy+zz);
	_view_matrix._m[1][0]=     2.0f*(xy-zw);
	_view_matrix._m[2][0]=     2.0f*(xz+yw);
	_view_matrix._m[0][1]=     2.0f*(xy+zw);
	_view_matrix._m[1][1]=1.0f-2.0f*(xx+zz);
	_view_matrix._m[2][1]=     2.0f*(yz-xw);
	_view_matrix._m[0][2]=     2.0f*(xz-yw);
	_view_matrix._m[1][2]=     2.0f*(yz+xw);
	_view_matrix._m[2][2]=1.0f-2.0f*(xx+yy);
	_view_matrix._m[0][3]=0.0f;
	_view_matrix._m[1][3]=0.0f;
	_view_matrix._m[2][3]=0.0f;
	_view_matrix._m[3][0]=pos.x;
	_view_matrix._m[3][1]=pos.y;
	_view_matrix._m[3][2]=pos.z;
	_view_matrix._m[3][3]=1.0f;

	_refresh_clip_matrix=true;
#ifndef USE_SHADERS
//	D3DXMATRIX	view;
	
//	D3DXMatrixTranslation(&view,-x/_resolution.x+0.5f,-y/_resolution.y+0.5f,-z);
	_device->SetTransform(D3DTS_VIEW,(D3DXMATRIX*)&_view_matrix);
#endif
}

void Gfx::set_camera(float x,float y,float z,Quaternion rot)
{	
	Quaternion	ori; ori.invert(rot);
	Vec3f		pos(-x,-y,-z); 
	ori.rotate_point(pos,pos);

	float xx=ori.x*ori.x;
	float xy=ori.x*ori.y;
	float xz=ori.x*ori.z;
	float xw=ori.x*ori.w;

	float yy=ori.y*ori.y;
	float yz=ori.y*ori.z;
	float yw=ori.y*ori.w;

	float zz=ori.z*ori.z;
	float zw=ori.z*ori.w;

	float ww=ori.w*ori.w;

	_view_matrix._m[0][0]=1.0f-2.0f*(yy+zz);
	_view_matrix._m[1][0]=     2.0f*(xy-zw);
	_view_matrix._m[2][0]=     2.0f*(xz+yw);
	_view_matrix._m[0][1]=     2.0f*(xy+zw);
	_view_matrix._m[1][1]=1.0f-2.0f*(xx+zz);
	_view_matrix._m[2][1]=     2.0f*(yz-xw);
	_view_matrix._m[0][2]=     2.0f*(xz-yw);
	_view_matrix._m[1][2]=     2.0f*(yz+xw);
	_view_matrix._m[2][2]=1.0f-2.0f*(xx+yy);
	_view_matrix._m[0][3]=0.0f;
	_view_matrix._m[1][3]=0.0f;
	_view_matrix._m[2][3]=0.0f;
	_view_matrix._m[3][0]=pos.x;
	_view_matrix._m[3][1]=pos.y;
	_view_matrix._m[3][2]=pos.z;
	_view_matrix._m[3][3]=1.0f;

	_refresh_clip_matrix=true;
#ifndef USE_SHADERS
/*	D3DXMATRIX		mat;
	D3DXQUATERNION	quat=D3DXQUATERNION(rot.x,rot.y,rot.z,rot.w);
	D3DXQUATERNION	tmp;
	D3DXVECTOR3		vec(-x,-y,-z);

	D3DXQuaternionInverse(&tmp,&quat);
	D3DXMatrixAffineTransformation(&mat,1.0f,&(-vec),&tmp,&vec);*/
	_device->SetTransform(D3DTS_VIEW,(D3DMATRIX*)&_view_matrix);
#endif
}

void Gfx::set_perspective(float fov,float aspect,float near_plane,float far_plane)
{
	float fovY=deg_to_rad(fov*0.25f);
	float h=cosf(fovY)/sinf(fovY);
	float fovX=fovY/aspect;
	float w=cosf(fovX)/sinf(fovX);
	float dp=far_plane-near_plane;

	_projection_matrix.zero();
	_projection_matrix._m[0][0]=w;
	_projection_matrix._m[1][1]=h;
	_projection_matrix._m[2][3]=1.0f;
	_projection_matrix._m[2][2]=far_plane/dp;
	_projection_matrix._m[3][2]=-_projection_matrix._m[2][2]*(near_plane);
	_refresh_clip_matrix=true;
#ifndef USE_SHADERS
/*	D3DXMATRIX	persp;

//	D3DXMatrixPerspectiveFovLH(&persp,deg_to_rad(fov),aspect,near_plane,far_plane);

	float fovY=deg_to_rad(fov*0.25f);
	float h=cosf(fovY)/sinf(fovY);
	float fovX=fovY/aspect;
	float w=cosf(fovX)/sinf(fovX);
	//	float w=h*aspect_ratio;
	float dp=far_plane-near_plane;

	memset(persp,0,sizeof(D3DMATRIX));
	persp._11=w;
	persp._22=h;
	persp._34=1.0f;
	persp._33=far_plane/dp;
	persp._43=-persp._33*(near_plane);//-((far_plane)*(near_plane))/dp;*/

	_device->SetTransform(D3DTS_PROJECTION,(D3DMATRIX*)&_projection_matrix);
#endif
}

Vec3f Gfx::transform_vertex(const float fov,const float aspect,const float near_plane,const float far_plane,
							const float cam_x,const float cam_y,const float cam_z,const Quaternion& rot,
							const Vec3f& v)
{
	// Build matrixes
	Matrix persp;

	float fovY=deg_to_rad(fov*0.25f);
	float h=cosf(fovY)/sinf(fovY);
	float fovX=fovY/aspect;
	float w=cosf(fovX)/sinf(fovX);
	float dp=far_plane-near_plane;

	persp.zero();
	persp._m[0][0]=w;
	persp._m[1][1]=h;
	persp._m[2][3]=1.0f;
	persp._m[2][2]=far_plane/dp;
	persp._m[3][2]=-persp._m[2][2]*(near_plane);

	Matrix view;
	Quaternion	ori; ori.invert(rot);
	Vec3f		pos(-cam_x,-cam_y,-cam_z); 
	ori.rotate_point(pos,pos);

	float xx=ori.x*ori.x;
	float xy=ori.x*ori.y;
	float xz=ori.x*ori.z;
	float xw=ori.x*ori.w;

	float yy=ori.y*ori.y;
	float yz=ori.y*ori.z;
	float yw=ori.y*ori.w;

	float zz=ori.z*ori.z;
	float zw=ori.z*ori.w;

	float ww=ori.w*ori.w;

	view._m[0][0]=1.0f-2.0f*(yy+zz);
	view._m[1][0]=     2.0f*(xy-zw);
	view._m[2][0]=     2.0f*(xz+yw);
	view._m[0][1]=     2.0f*(xy+zw);
	view._m[1][1]=1.0f-2.0f*(xx+zz);
	view._m[2][1]=     2.0f*(yz-xw);
	view._m[0][2]=     2.0f*(xz-yw);
	view._m[1][2]=     2.0f*(yz+xw);
	view._m[2][2]=1.0f-2.0f*(xx+yy);
	view._m[0][3]=0.0f;
	view._m[1][3]=0.0f;
	view._m[2][3]=0.0f;
	view._m[3][0]=pos.x;
	view._m[3][1]=pos.y;
	view._m[3][2]=pos.z;
	view._m[3][3]=1.0f;

	Vec4f	ret(v.x,v.y,v.z,1.0f);
	(view*persp).pre_multiply_4x4(&ret);
	ret=ret/fabs(ret.w);

	ret.x=(ret.x*0.5f+0.5f)*_vres.x;
	ret.y=(1.0f-(ret.y*0.5f+0.5f))*_vres.y;

	return Vec3f(ret.x,ret.y,ret.z);
}

void Gfx::set_world_transformation(int idx,float x,float y,float z,Quaternion ori)
{
	float xx=ori.x*ori.x;
	float xy=ori.x*ori.y;
	float xz=ori.x*ori.z;
	float xw=ori.x*ori.w;

	float yy=ori.y*ori.y;
	float yz=ori.y*ori.z;
	float yw=ori.y*ori.w;

	float zz=ori.z*ori.z;
	float zw=ori.z*ori.w;

	float ww=ori.w*ori.w;

	_world_matrix._m[0][0]=(1.0f-2.0f*(yy+zz)); 
	_world_matrix._m[1][0]=(     2.0f*(xy-zw)); 
	_world_matrix._m[2][0]=(     2.0f*(xz+yw)); 
	_world_matrix._m[0][1]=(     2.0f*(xy+zw)); 
	_world_matrix._m[1][1]=(1.0f-2.0f*(xx+zz));  
	_world_matrix._m[2][1]=(     2.0f*(yz-xw)); 
	_world_matrix._m[0][2]=(     2.0f*(xz-yw)); 
	_world_matrix._m[1][2]=(     2.0f*(yz+xw)); 
	_world_matrix._m[2][2]=(1.0f-2.0f*(xx+yy)); 
	_world_matrix._m[0][3]=0.0f;
	_world_matrix._m[1][3]=0.0f;
	_world_matrix._m[2][3]=0.0f;
	_world_matrix._m[3][0]=x;
	_world_matrix._m[3][1]=y;
	_world_matrix._m[3][2]=z;
	_world_matrix._m[3][3]=1.0f;

	_refresh_clip_matrix=true;
#ifndef USE_SHADERS
//	D3DXMATRIX		w;
//	D3DXVECTOR3		vec=D3DXVECTOR3(x,y,z);
//	D3DXQUATERNION	rot=D3DXQUATERNION(r.x,r.y,r.z,r.w);

//	D3DXMatrixAffineTransformation(&w,1.0f,NULL,&rot,&vec);
	_device->SetTransform(D3DTS_WORLD,(D3DMATRIX*)&_world_matrix);
#endif
}

void Gfx::set_world_transformation(int idx,float x,float y,float z,Quaternion ori,const float sx,const float sy,const float sz)
{
	float xx=ori.x*ori.x;
	float xy=ori.x*ori.y;
	float xz=ori.x*ori.z;
	float xw=ori.x*ori.w;

	float yy=ori.y*ori.y;
	float yz=ori.y*ori.z;
	float yw=ori.y*ori.w;

	float zz=ori.z*ori.z;
	float zw=ori.z*ori.w;

	float ww=ori.w*ori.w;

	_world_matrix._m[0][0]=(1.0f-2.0f*(yy+zz))*sx; 
	_world_matrix._m[1][0]=(     2.0f*(xy-zw))*sy; 
	_world_matrix._m[2][0]=(     2.0f*(xz+yw))*sz; 
	_world_matrix._m[0][1]=(     2.0f*(xy+zw))*sx; 
	_world_matrix._m[1][1]=(1.0f-2.0f*(xx+zz))*sy;  
	_world_matrix._m[2][1]=(     2.0f*(yz-xw))*sz; 
	_world_matrix._m[0][2]=(     2.0f*(xz-yw))*sx; 
	_world_matrix._m[1][2]=(     2.0f*(yz+xw))*sy; 
	_world_matrix._m[2][2]=(1.0f-2.0f*(xx+yy))*sz; 
	_world_matrix._m[0][3]=0.0f;
	_world_matrix._m[1][3]=0.0f;
	_world_matrix._m[2][3]=0.0f;
	_world_matrix._m[3][0]=x;
	_world_matrix._m[3][1]=y;
	_world_matrix._m[3][2]=z;
	_world_matrix._m[3][3]=1.0f;

	_refresh_clip_matrix=true;
#ifndef USE_SHADERS
//	D3DXMATRIX		w;
//	D3DXVECTOR3		vec=D3DXVECTOR3(x,y,z);
//	D3DXQUATERNION	rot=D3DXQUATERNION(r.x,r.y,r.z,r.w);

//	D3DXMatrixAffineTransformation(&w,1.0f,NULL,&rot,&vec);

//	D3DXMATRIX		s; D3DXMatrixScaling(&s,sx,sy,sz);
//	w=s*w;

	_device->SetTransform(D3DTS_WORLD,(D3DMATRIX*)&_world_matrix);
#endif
}

void Gfx::unset_world_transformation(int idx)
{
	_world_matrix.identity();
	_refresh_clip_matrix=true;
#ifndef USE_SHADERS
/*	D3DXMATRIX w;
	
	D3DXMatrixIdentity(&w);*/

	_device->SetTransform(D3DTS_WORLD,(D3DMATRIX*)&_world_matrix);
#endif
}

void Gfx::clear_lights()
{
	for (int i=0; i<8; i++) 
	{
		_device->LightEnable(i,FALSE);
	}
	_env._props._light_count=0;
}

void Gfx::set_fog(const bool enable,const Color& color,const float start,const float end,const Vec3f& center)
{
#ifdef USE_SHADERS
	if (enable)
	{
		_env._props._fog_linear=1;
		_fog_color=color;
		_fog_range.set(start,end);
		_fog_center=center;
	}
	else
	{
		_env._props._fog_linear=0;
	}
#else
	if (enable)
	{
		_device->SetRenderState(D3DRS_FOGENABLE,TRUE);
		_device->SetRenderState(D3DRS_FOGCOLOR,float_to_colorref(color));
		_device->SetRenderState(D3DRS_FOGVERTEXMODE,D3DFOG_LINEAR);
		_device->SetRenderState(D3DRS_RANGEFOGENABLE,FALSE);
		_device->SetRenderState(D3DRS_FOGSTART,FtoDW(start));
		_device->SetRenderState(D3DRS_FOGEND,FtoDW(end));
		_fog_color=color;
		_fog_start=start;
		_fog_end=end;
	}
	else
	{
		_device->SetRenderState(D3DRS_FOGENABLE,FALSE);
	}
#endif
}

bool Gfx::get_available_resolutions(std::vector< Resolution >* dest)
{
	LPDIRECT3D9 d3d9=Direct3DCreate9(D3D_SDK_VERSION);
	if (d3d9==NULL) return false;

	CD3DEnumeration enumerate;
	enumerate.SetD3D(d3d9);
	enumerate.ConfirmDeviceCallback=confirm_device_helper;
	enumerate.Enumerate();

	d3d9->Release();

	for (UINT i=0; i<enumerate.m_pAdapterInfoList->Count(); i++)
	{
		D3DAdapterInfo* pAdapterInfo;
		pAdapterInfo = (D3DAdapterInfo*)enumerate.m_pAdapterInfoList->GetPtr(i);

		for (UINT j=0; j<pAdapterInfo->pDisplayModeList->Count(); j++)
		{
			D3DDISPLAYMODE dmode=*(D3DDISPLAYMODE*)pAdapterInfo->pDisplayModeList->GetPtr(j);

			Resolution	res;
			res._adapter_ordinal=pAdapterInfo->AdapterOrdinal;
			res._adapter=pAdapterInfo->AdapterIdentifier.DeviceName;
			res._width=dmode.Width;
			res._height=dmode.Height;
			res._bitdepth=get_bitcount(dmode.Format);
			res._refresh_rate=dmode.RefreshRate;

			dest->push_back(res);
		}
	}

	return true;
}

int	Gfx::get_bitcount(const D3DFORMAT format)
{
	switch (format)
	{
		case D3DFMT_R5G6B5: 
			return 16; break;
		case D3DFMT_X8R8G8B8: 
		case D3DFMT_A8R8G8B8: 
			return 32; break;
		default:
			return 0;
			break;
	}
}

HRESULT	Gfx::create_font(INT Height,UINT Width,UINT Weight,UINT MipLevels,BOOL Italic,DWORD CharSet,DWORD OutputPrecision,DWORD Quality,DWORD PitchAndFamily,LPCSTR pFaceName,LPD3DXFONT* ppFont)
{
#ifdef USE_D3D9
	UINT corrected_width=Width;//(UINT)ceilf((float)Width*(_resolution.x/_vres.x));
	UINT corrected_height=Height;//(UINT)ceilf((float)Height*(_resolution.y/_vres.y));
	return D3DXCreateFont(_device,corrected_height,corrected_width,Weight,MipLevels,Italic,CharSet,OutputPrecision,Quality,PitchAndFamily,pFaceName,ppFont);
#else
	return NULL;
#endif
}

void Gfx::add_light(Light* lgt)
{
	lgt->update();
	if (_env._props._light_count<MAX_LIGHT)
	{
#ifdef USE_SHADERS
		Vec3f d; lgt->_ori.rotate_point(d,Vec3f(0.0f,0.0f,1.0f));
		_lights_struct[_env._props._light_count].pos.set(lgt->_pos.x,lgt->_pos.y,lgt->_pos.z,1.0f);
		_lights_struct[_env._props._light_count].dir.set(d.x,d.y,d.z,0.0f);
		_lights_struct[_env._props._light_count].color=lgt->get_color();
		_lights_struct[_env._props._light_count].params.set(lgt->get_range_start(),lgt->get_range_end(),cosf(deg_to_rad(lgt->get_inner_cone()*0.5f)),cosf(deg_to_rad(lgt->get_outer_cone()*0.5f)));
#else
		_device->SetLight(_env._props._light_count,&(lgt->_light));
		_device->LightEnable(_env._props._light_count,TRUE);
#endif
		_lights[_env._props._light_count]=lgt;
		_env._props._lights[_env._props._light_count]._type=lgt->get_type();
		_env._props._light_count++;
	}
}

void Gfx::render(D3DPRIMITIVETYPE prim_type,VertexBuffer* vb,const UINT vcount,IndexBuffer* ib,const UINT icount)
{
#ifdef USE_D3D9
	UINT prim_count;
	
	_device->SetStreamSource(0,vb->_vb,0,vb->_vertex_size);
	if (ib) 
	{
		if (prim_type==D3DPT_TRIANGLELIST) prim_count=icount/3;
		else if (prim_type==D3DPT_POINTLIST) prim_count=icount;
		else if (prim_type==D3DPT_LINELIST) prim_count=icount/2;
		_device->SetIndices(ib->_ib);
		_device->DrawIndexedPrimitive(prim_type,0,0,vcount,0,prim_count);
	}
	else
	{
		if (prim_type==D3DPT_TRIANGLELIST) prim_count=vcount/3;
		else if (prim_type==D3DPT_POINTLIST) prim_count=vcount;
		else if (prim_type==D3DPT_LINELIST) prim_count=vcount/2;
		_device->SetIndices(NULL);
		_device->DrawPrimitive(prim_type,0,prim_count);
	}
#endif
}

TextureID Gfx::load_texture(const std::string& filename,int& width,int& height)
{
#ifdef USE_D3D9
	LPDIRECT3DTEXTURE9	texture;
	D3DXIMAGE_INFO		info;
	HRESULT hRet=D3DXCreateTextureFromFileEx(_device,filename.c_str(),D3DX_DEFAULT,D3DX_DEFAULT,D3DX_DEFAULT,
											 0,D3DFMT_UNKNOWN, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT, 0, &info, NULL,
											 &texture);
	if (hRet!=D3D_OK)
	{
		return NULL;
	}

	D3DSURFACE_DESC	desc;
	texture->GetLevelDesc(0,&desc);
	width=desc.Width;
	height=desc.Height;

	return texture;
#else
	__asm int 3
	return TEXTURE_INVALID;
#endif
}

TextureID Gfx::new_texture(const int sx,const int sy,const D3DFORMAT format,const bool rt,ZBufferID* z_buffer)
{
#ifdef USE_D3D9
	LPDIRECT3DTEXTURE9	texture;
	DWORD				usage=0;
	D3DPOOL				pool=D3DPOOL_MANAGED;
	if (rt) { usage=D3DUSAGE_RENDERTARGET; pool=D3DPOOL_DEFAULT; }

	if (_device->CreateTexture((UINT)sx,sy,1,usage,format,pool,&texture,NULL)!=D3D_OK)
	{
		return TEXTURE_INVALID;
	}

	if ((z_buffer) && (rt))
	{
		// Create a Z-Buffer
		*z_buffer=create_z_buffer(sx,sy,format);
	}

	return texture;
#else
	__asm int 3
	return TEXTURE_INVALID;
#endif
}

ZBufferID Gfx::create_z_buffer(const int sx,const int sy,const D3DFORMAT format)
{
#ifdef USE_D3D9
	D3DFORMAT				zformat=D3DFMT_D32;
	LPDIRECT3DSURFACE9		z_buffer;

	HRESULT hRet=_device->CreateDepthStencilSurface(UINT(sx),UINT(sy),zformat,D3DMULTISAMPLE_NONE,0,TRUE,&z_buffer,NULL);
	switch (hRet)
	{
		case D3D_OK:
			break;
		case D3DERR_INVALIDCALL:
		case D3DERR_OUTOFVIDEOMEMORY:
		case E_OUTOFMEMORY:
		default:
			z_buffer=NULL;
			break;
	}

	return z_buffer;
#else
	__asm int 3
	return TEXTURE_INVALID;
#endif
}

void Gfx::refresh_clip_matrix()
{
	_clip_matrix=(_world_matrix*_view_matrix)*_projection_matrix;
	_refresh_clip_matrix=false;
}

void Gfx::set_render_target(const int index,SurfaceID target,ZBufferID z_buffer)
{
	_device->SetRenderTarget((DWORD)index,target);
	if (z_buffer) { _device->SetDepthStencilSurface(z_buffer); _set_depth_buffer=z_buffer; }
}

void Gfx::unset_render_target(const int index)
{
	_device->SetRenderTarget((DWORD)index,_default_render_target);
	if (_set_depth_buffer!=_default_depth_buffer)
	{
		_device->SetDepthStencilSurface(_default_depth_buffer);
		_set_depth_buffer=_default_depth_buffer;
	}
}

SurfaceID Gfx::get_surface(TextureID texture)
{
	SurfaceID tmp;
	texture->GetSurfaceLevel(0,&tmp);

	return tmp;
}

void Gfx::set_pass(const PassType pass) 
{ 
	_pass=pass; 
	_env._props._pass=(int)pass; 
	BLIT->flush();
	BLIT3D->flush(); 
}
