#pragma once

#include "Defs.h"
#include "Gfx.h"

class Material
{
protected:
	typedef std::map<std::string,Vec4f>	CustomParams;
public:
	Material(void);
	virtual ~Material(void);

	inline	void		set_name(const std::string& name) { _name=name; }
	inline	std::string	get_name() { return _name; }
			void		set(const unsigned long fvf);
			void		unset();
			void		set_alpha(const float a);
			float		get_alpha();
			void		set_material_color(float r,float g,float b,float a);
			Color		get_material_color();
			void		set_emissive_color(float r,float g,float b,float a);
			void		set_shader(const char* shader);
	inline	void		set_cull_mode(DWORD cull_mode) { _cull_mode=cull_mode; }
	inline	void		set_texture(long slot,long tex_id) { _textures[slot]=tex_id; }
	inline	long		get_texture(long slot) { return _textures[slot]; }
	inline	void		set_texture_sampling(long slot,const D3DTEXTUREFILTERTYPE filter) { _textures_filter[slot]=filter; }
	inline	void		set_alpha(bool active,D3DBLEND src_blend,D3DBLEND dest_blend)
	{
		_alpha_enable=active;
		_src_blend=src_blend;
		_dest_blend=dest_blend;
	}
	inline	void		set_color_vertex(const bool b,const bool e) { _color_vertex=b; _color_vertex_e=e; }
	inline	void		set_z_write(const bool b) { _z_write=b; }
	inline	void		set_point_sprite(const bool b) { _point_sprite=b; }
	inline	void		set_alpha_test(const bool b,const unsigned char r) { _alpha_test=b; _alpha_ref=r; }
	inline	bool		pass_enabled(const PassType pass) { return ((_passes>>((int)pass))&0x1)!=0; }
	inline	void		enable_all_passes() { _passes=0xFFFFFFFF; }
	inline	void		disable_all_passes() { _passes=0; }
	inline	void		enable_pass(const PassType pass) { _passes|=1<<((int)pass); }
	inline	void		disable_pass(const PassType pass) { _passes&=~(1<<((int)pass)); }

	inline	void		add_custom_param(const std::string& name,const Vec4f& val) { auto it=_custom_params.find(name); if (it==_custom_params.end()) _custom_params.insert(std::make_pair(name,val)); else it->second=val; }
	inline	void		set_custom_param(const std::string& name,const Vec4f& val) { auto it=_custom_params.find(name); it->second=val; }

			Material*	clone();

protected:
	// Material properties
	std::string				_name;
	D3DMATERIAL9			_material;
	DWORD					_cull_mode;
	long					_textures[MAX_TEX];
	D3DTEXTUREFILTERTYPE	_textures_filter[MAX_TEX];
	bool					_alpha_enable;
	bool					_color_vertex;
	bool					_color_vertex_e;
	bool					_z_write;
	bool					_point_sprite;
	bool					_alpha_test;
	unsigned char			_alpha_ref;
	D3DBLEND				_src_blend;
	D3DBLEND				_dest_blend;
	uint32_t				_passes;
	std::string				_shader;
	CustomParams			_custom_params;
protected:
	// Old state
	unsigned long			_old_z_write;
	unsigned long			_old_lighting;
	unsigned long			_old_alpha_test;
	unsigned long			_old_alpha_ref;

#ifdef USE_SHADERS
protected:
	typedef enum { CONST_CLIPMATRIX=0, 
				   CONST_WORLDMATRIX0=1,
				   CONST_WORLDMATRIX255=256,
				   CONST_MATERIAL_AMBIENT=257,CONST_MATERIAL_DIFFUSE=258,CONST_MATERIAL_SPECULAR=259,CONST_MATERIAL_SPECULARPARAMS=260,CONST_MATERIAL_EMISSIVE=261,
				   CONST_AMBIENT_LIGHT=300, CONST_LIGHTS=301,
				   CONST_FOG_CENTER=400, CONST_FOG_PARAMS=401, CONST_FOG_COLOR=402,
				   CONST_RTRES=500,CONST_GLOBALTIME=501,
				   CONST_CUSTOM=0xFFFF } ConstantType;
	struct Constant
	{
		ConstantType	_type;
		int				_pos;
		int				_size;
		std::string		_extra;
	};
	typedef std::vector<Constant>	Constants;
	struct ShaderDefine
	{
		std::string	_name;
		std::string	_value;
	};
	typedef std::vector<ShaderDefine>	ShaderDefines;
	struct ShaderVariant
	{
#ifdef USE_D3D9
		LPDIRECT3DVERTEXDECLARATION9	_decl;
		LPDIRECT3DVERTEXSHADER9			_vertex;
		LPDIRECT3DPIXELSHADER9			_pixel;
#endif
		Constants		_vertex_constant;
		Constants		_pixel_constant;
		ShaderDefines	_defines;
	};
	typedef std::map<RenderEnv,ShaderVariant*>	ShaderVariants;
	struct Shader
	{
		std::string		_name;
		ShaderVariants	_variants;
	};
	typedef std::map<std::string,Shader*>		Shaders;
	typedef std::map<int,VertexDeclaration>		Decls;

			void	set_shader(const std::string& shader_name,RenderEnv* env,const int fvf);
			void	unset_shader();
	static	void	create_shader_variant(Shader* shader,RenderEnv* env,const int fvf);
public:
	static	void				clear_shaders();
	static	VertexDeclaration	get_vertex_decl(const int fvf);
	static	ConstantType		get_binding_by_name(const std::string& name);
			void				set_constants(const bool vertex_constants,Constants* constants);
#ifdef USE_D3D9
	static	void				process_constants(ID3DXConstantTable* constants,Constants* dest);
#endif

protected:
	static	Shaders	g_shaders;
	static	Decls	g_decls;
#endif
};

Material*	get_material(const std::string& name);
void		load_material_3ds(FILE* file,const std::string& current_filename);