#ifndef __MATHLIB_H
#define __MATHLIB_H

#pragma once

#include <math.h>
#include <d3dx9.h>
#include <vector>

const float PI = 3.1415926535f;
const float PI_OVER_180 = PI / 180.0f;
const float PI_UNDER_180 = 180.0f / PI;
const float PI_OVER_2 = PI / 2.0f;
const float PI_OVER_TWO = PI_OVER_2;
const float PI_TIMES_2 = PI * 2.0f;
const float PI_TIMES_TWO = PI_TIMES_2;

const float VERY_SMALL_FLOAT = 1e-8f;
const float SMALLEST_FLOAT = -99999999.0f;
const float BIGGEST_FLOAT = 99999999.0f;
const float EPSILON_FLOAT = 1e-5f;
const float EPSILON_DOUBLE = 1e-15f;

const float SQRT2=1.414214f;

#define Sqrt(X) (float)(sqrtf(X))
#define Tan(X) (float)(tanf(X))
#define Floor(X) (float)(floorf(X))
#define Ceil(X) (float)(ceilf(X))
#define Atof(x) (float)(atof((x)))
#define Abs(x) (float)(fabs((x)))
#define Exp(x) (float)(exp((x)))
#define Sqr(x) ((x)*(x))
#define Sign(x) (((x)>0.0f)?(1.0f):((((x)<0.0f)?(-1.0f):(0.0f))))

inline float Gauss(float x,float mean,float variance)
{
	float c=(float)(1.0f/(variance*Sqrt(PI_TIMES_2)));
	float p=-(Sqr(x-mean)/(2*Sqr(variance)));
	float e=Exp(p);
	return 2.5f*c*e;
}

inline float deg_to_rad(const float ang) { return ang*PI_OVER_180; }
inline float rad_to_deg(const float ang) { return ang*PI_UNDER_180; }

#define Min(x,y) (((x)<(y))?(x):(y))
#define Max(x,y) (((x)>(y))?(x):(y))

#define Clamp(X,MIN,MAX) (((X)<(MIN))?(MIN):(((X)>(MAX))?(MAX):(X)))
/*
template<typename T>
inline	void swap(T& v1,T& v2)
{
	T tmp=v1;
	v1=v2;
	v2=tmp;
}
*/
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

class Vec2f
{
public:
	float x;
	float y;
public:
	Vec2f();									
	Vec2f(const float a, const float b);			

	float			normalize();							
	float			norm() const;							
	float			norm_squared() const;					
	void			set(const float a,const float b);

	Vec2f&			operator*=(const float s);

	float&			operator[](const int i);
	float&			operator()(const int i);
	float			operator[](const int i) const;
	float			operator()(const int i) const;

	float			dot(const Vec2f v) const;
};


inline bool operator<(const Vec2f& o1,const Vec2f& o2)
{
	if (o1.x<o2.x) return true;
	if (o1.x==o2.x)
	{
		return o1.y<o2.y;
	}
	return false;
}


class Vec3f
{
public:
	float x;
	float y;
	float z;
public:
	Vec3f();									
	Vec3f(const Vec3f& v);			
	Vec3f(const Vec3f* v);			
	Vec3f(const float v[3]);								
	Vec3f(const float a, const float b, const float c);	

	float			normalize();						
	float			norm() const;							
	float			norm_squared() const;						
	void			set(const float a,const float b,const float c);
	bool			is_null() const;

	float&			operator[](const int i);
	float&			operator()(const int i);
	float			operator[](const int i) const;
	float			operator()(const int i) const;

	Vec3f&	operator*=(float s);

	float			dot(const Vec3f v) const;
	Vec3f			cross(const Vec3f v) const;
};

inline bool operator<(const Vec3f& o1,const Vec3f& o2)
{
	if (o1.x<o2.x) return true;
	if (o1.x==o2.x)
	{
		if (o1.y<o2.y) return true;
		if (o1.y==o2.y)
		{
			return o1.z<o2.z;
		}
	}
	return false;
}

class Vec4f
{
public:
	union {
		float x;
		float r;
		float x1;
	};
	union {
		float y;
		float g;
		float y1;
	};
	union {
		float z;
		float b;
		float x2;
	};
	union {
		float w;
		float a;
		float y2;
	};
public:
	Vec4f();									
	Vec4f(const Vec2f& v2);
	Vec4f(const Vec3f& v3);
	Vec4f(const float a, const float b, const float c, const float d);	
	float			normalize();							
	float			norm() const;								
	float			norm_squared() const;						
	void			set(const float a,const float b,const float c,const float d);
	bool			inside(const Vec2f& v) const;
	void			harmonize_rect();

	// For Rect only
	void			offset(const float x,const float y);

	Vec4f&			operator*=(float s);
	float&			operator[](const int i);
	float			operator[](const int i) const;
};

inline bool operator<(const Vec4f& o1,const Vec4f& o2)
{
	if (o1.x<o2.x) return true;
	if (o1.x==o2.x)
	{
		if (o1.y<o2.y) return true;
		if (o1.y==o2.y)
		{
			if (o1.z<o2.z) return true;
			if (o1.z==o2.z)
			{
				return o1.w<o2.w;
			}
		}
	}
	return false;
}

/*********************************************************************************************
*                                        >>> Vec2 <<<                                       *
*********************************************************************************************/


inline Vec2f::Vec2f()
{
}


inline Vec2f::Vec2f(const float a, const float b)
{
	x = a;
	y = b;
}


inline void Vec2f::set(const float a,const float b)
{
	x=a; y=b;;
}


inline float Vec2f::normalize()
{
	register float norm = (float)sqrt(x*x + y*y);
	x = x/norm;
	y = y/norm;

	return norm;
}


inline float Vec2f::norm() const
{
	return (float)sqrt(x*x + y*y);
}


inline float Vec2f::norm_squared() const
{
	return x*x + y*y;
}

/**
*  Adds two vectors..
*/

inline Vec2f operator+(const Vec2f& vec1, const Vec2f& vec2)
{
	Vec2f retVec;

	retVec.x = vec1.x + vec2.x;
	retVec.y = vec1.y + vec2.y;

	return retVec;
}


inline Vec2f operator/(const Vec2f& vec1, const Vec2f& vec2)
{
	Vec2f retVec;

	retVec.x = vec1.x/vec2.x;
	retVec.y = vec1.y/vec2.y;

	return retVec;
}


inline Vec2f operator/(const Vec2f& vec, const float s)
{
	Vec2f retVec;

	retVec.x = vec.x / s;
	retVec.y = vec.y / s;

	return retVec;
}


inline Vec2f operator-(const Vec2f& vec1, const Vec2f& vec2)
{
	Vec2f retVec;

	retVec.x = vec1.x - vec2.x;
	retVec.y = vec1.y - vec2.y;

	return retVec;
}


// Produto com um escalar

inline Vec2f operator*(const Vec2f& vec, float s)
{
	Vec2f retVec;

	retVec.x = vec.x * s;
	retVec.y = vec.y * s;

	return retVec;
}

// Produto com um escalar

inline Vec2f operator*(const float s, const Vec2f& vec)
{
	Vec2f retVec;

	retVec.x = vec.x * s;
	retVec.y = vec.y * s;

	return retVec;
}

// Negar o vector (produto com -1)

inline Vec2f operator-(const Vec2f& vec)
{
	Vec2f retVec;

	retVec.x = -vec.x;
	retVec.y = -vec.y;

	return retVec;
}


inline Vec2f& Vec2f::operator*=(const float s)
{
	// Scale coefficients
	x*=s; y*=s;

	return *this;
}


inline bool operator==(const Vec2f& vec1, const Vec2f& vec2)
{
	return ((vec1.x == vec2.x) &&
			(vec1.y == vec2.y));
}


inline bool is_equal(const Vec2f& vec1, const Vec2f& vec2,float epsilon)
{
	return ((fabs(vec1.x-vec2.x)<epsilon) &&
			(fabs(vec1.y-vec2.y)<epsilon));
}


inline bool operator!=(const Vec2f& vec1, const Vec2f& vec2)
{
	return ((vec1.x != vec2.x) ||
		(vec1.y != vec2.y)    );
}


inline float dot_product(const Vec2f& vec1, const Vec2f& vec2)
{
	return vec1.x * vec2.x + vec1.y * vec2.y;
}


inline float& Vec2f::operator [](const int i)
{
	if (i==0) return x;
	return y;
}


inline float& Vec2f::operator ()(const int i)
{
	if (i==0) return x;
	return y;
}


inline float Vec2f::operator [](const int i) const
{
	if (i==0) return x;
	return y;
}


inline float Vec2f::operator ()(const int i) const 
{
	if (i==0) return x;
	return y;
}


inline	float Vec2f::dot(const Vec2f v) const 
{
	return x*v.x+y*v.y;
}

/*********************************************************************************************
*                                        >>> Vec3f <<<                                       *
*********************************************************************************************/

inline Vec3f::Vec3f()
{
}


inline Vec3f::Vec3f(const Vec3f& v)
{
	x = (float)v.x;
	y = (float)v.y;
	z = (float)v.z;
}


inline Vec3f::Vec3f(const Vec3f* v)
{
	x = (float)v->x;
	y = (float)v->y;
	z = (float)v->z;
}

inline Vec3f::Vec3f(const float v[3])
{
	x=v[0];
	y=v[1];
	z=v[2];
}


inline Vec3f::Vec3f(const float a, const float b, const float c)
{
	x = a;
	y = b;
	z = c;
}


inline void Vec3f::set(const float a,const float b,const float c)
{
	x=a; y=b; z=c;
}


inline float Vec3f::normalize()
{
	register float norm = (float)sqrt(x*x + y*y + z*z);
	x = x/norm;
	y = y/norm;
	z = z/norm;

	return norm;
}


inline float Vec3f::norm() const
{
	return (float)sqrt(x*x + y*y + z*z);
}


inline float Vec3f::norm_squared() const
{
	return x*x + y*y + z*z;
}


inline bool Vec3f::is_null() const
{
	return ((fabs(x)<EPSILON_FLOAT) &&
		(fabs(y)<EPSILON_FLOAT) &&
		(fabs(z)<EPSILON_FLOAT));
}


inline float& Vec3f::operator [](const int i)
{
	if (i==0) return x;
	else if (i==1) return y;
	else return z;
}


inline float& Vec3f::operator ()(const int i)
{
	if (i==0) return x;
	else if (i==1) return y;
	else return z;
}


inline float Vec3f::operator [](const int i) const
{
	if (i==0) return x;
	else if (i==1) return y;
	else return z;
}


inline float Vec3f::operator ()(const int i) const 
{
	if (i==0) return x;
	else if (i==1) return y;
	else return z;
}


inline Vec3f& Vec3f::operator*=(const float s)
{
	// Scale coefficients
	x*=s; y*=s; z*=s;

	return *this;
}


inline Vec3f operator+(const Vec3f& vec1, const Vec3f& vec2)
{
	Vec3f retVec;

	retVec.x = vec1.x + vec2.x;
	retVec.y = vec1.y + vec2.y;
	retVec.z = vec1.z + vec2.z;

	return retVec;
}


inline Vec3f operator-(const Vec3f& vec1, const Vec3f& vec2)
{
	Vec3f retVec;

	retVec.x = vec1.x - vec2.x;
	retVec.y = vec1.y - vec2.y;
	retVec.z = vec1.z - vec2.z;

	return retVec;
}

// Produto com um escalar

inline Vec3f operator*(const Vec3f& vec, float s)
{
	Vec3f retVec;

	retVec.x = vec.x * s;
	retVec.y = vec.y * s;
	retVec.z = vec.z * s;

	return retVec;
}

// Produto com um escalar

inline Vec3f operator*(const float s, const Vec3f& vec)
{
	Vec3f retVec;

	retVec.x = vec.x * s;
	retVec.y = vec.y * s;
	retVec.z = vec.z * s;

	return retVec;
}

// Divisiao com um escalar

inline Vec3f operator/(const Vec3f& vec, const float s)
{
	Vec3f retVec;

	retVec.x = vec.x / s;
	retVec.y = vec.y / s;
	retVec.z = vec.z / s;

	return retVec;
}

// Divisao com um escalar

inline Vec3f operator/(const float s, const Vec3f& vec)
{
	Vec3f retVec;

	retVec.x = vec.x / s;
	retVec.y = vec.y / s;
	retVec.z = vec.z / s;

	return retVec;
}

// Negar o vector (produto com -1)

inline Vec3f operator-(const Vec3f& vec)
{
	Vec3f retVec;

	retVec.x = -vec.x;
	retVec.y = -vec.y;
	retVec.z = -vec.z;

	return retVec;
}


inline bool operator==(const Vec3f& vec1, const Vec3f& vec2)
{
	return ((fabs(vec1.x-vec2.x)<EPSILON_FLOAT) &&
		(fabs(vec1.y-vec2.y)<EPSILON_FLOAT) &&
		(fabs(vec1.z-vec2.z)<EPSILON_FLOAT));
}


inline bool is_equal(const Vec3f& vec1, const Vec3f& vec2,float epsilon)
{
	return ((fabs(vec1.x-vec2.x)<epsilon) &&
		(fabs(vec1.y-vec2.y)<epsilon) &&
		(fabs(vec1.z-vec2.z)<epsilon));
}


inline bool operator!=(const Vec3f& vec1, const Vec3f& vec2)
{
	return !((fabs(vec1.x-vec2.x)<EPSILON_FLOAT) &&
		(fabs(vec1.y-vec2.y)<EPSILON_FLOAT) &&
		(fabs(vec1.z-vec2.z)<EPSILON_FLOAT));
}


inline float dot_product(const Vec3f& vec1, const Vec3f& vec2)
{
	return vec1.x * vec2.x + vec1.y * vec2.y + vec1.z * vec2.z;
}


inline float dot_product(const Vec3f *vec1,const Vec3f *vec2)
{
	return vec1->x * vec2->x + vec1->y * vec2->y + vec1->z * vec2->z;
}


inline Vec3f cross_product(const Vec3f& u, const Vec3f& v)
{
	Vec3f retVec;

	retVec.x = u.y * v.z - u.z * v.y;
	retVec.y = u.z * v.x - u.x * v.z;
	retVec.z = u.x * v.y - u.y * v.x;

	return retVec;
}


inline float Vec3f::dot(const Vec3f v) const 
{
	return x*v.x+y*v.y+z*v.z;
}


inline Vec3f Vec3f::cross(const Vec3f v) const 
{
	Vec3f retVec;

	retVec.x = y * v.z - z * v.y;
	retVec.y = z * v.x - x * v.z;
	retVec.z = x * v.y - y * v.x;

	return retVec;
}

/*********************************************************************************************
*                                        >>> Vec4 <<<                                       *
*********************************************************************************************/


inline Vec4f::Vec4f()
{
}


inline Vec4f::Vec4f(const float a, const float b, const float c, const float d)
{
	x = a;
	y = b;
	z = c;
	w = d;
}

inline Vec4f::Vec4f(const Vec2f& v2)
{
	x=v2.x;
	y=v2.y;
	z=w=0.0f;
}

inline Vec4f::Vec4f(const Vec3f& v3)
{
	x=v3.x;
	y=v3.y;
	z=v3.z;
	w=0.0f;
}

inline float Vec4f::normalize()
{
	register float norm = (float)sqrt(x*x + y*y + z*z + w*w);
	x = x/norm;
	y = y/norm;
	z = z/norm;
	w = w/norm;

	return norm;
}


inline float Vec4f::norm() const
{
	return (float)sqrt(x*x + y*y + z*z + w*w);
}


inline float Vec4f::norm_squared() const
{
	return x*x + y*y + z*z + w*w;
}


inline void Vec4f::set(const float a,const float b,const float c,const float d)
{
	x=a; y=b; z=c; w=d;
}


inline void Vec4f::offset(const float a,const float b)
{
	x=x+a;
	y=y+b;
	z=z+a;
	w=w+b;
}


inline Vec4f& Vec4f::operator*=(const float s)
{
	// Scale coefficients
	x*=s; y*=s; z*=s; w*=s;

	return *this;
}


inline Vec4f operator+(const Vec4f& vec1, const Vec4f& vec2)
{
	Vec4f retVec;

	retVec.x = vec1.x + vec2.x;
	retVec.y = vec1.y + vec2.y;
	retVec.z = vec1.z + vec2.z;
	retVec.w = vec1.w + vec2.w;

	return retVec;
}


inline Vec4f operator-(const Vec4f& vec1, const Vec4f& vec2)
{
	Vec4f retVec;

	retVec.x = vec1.x - vec2.x;
	retVec.y = vec1.y - vec2.y;
	retVec.z = vec1.z - vec2.z;
	retVec.w = vec1.w - vec2.w;

	return retVec;
}

// Produto com um escalar

inline Vec4f operator*(const Vec4f& vec, const float s)
{
	Vec4f retVec;

	retVec.x = vec.x * s;
	retVec.y = vec.y * s;
	retVec.z = vec.z * s;
	retVec.w = vec.w * s;

	return retVec;
}

// Produto com um escalar

inline Vec4f operator*(const float s, const Vec4f& vec)
{
	Vec4f retVec;

	retVec.x = vec.x * s;
	retVec.y = vec.y * s;
	retVec.z = vec.z * s;
	retVec.w = vec.w * s;

	return retVec;
}

// Divisiao com um escalar

inline Vec4f operator/(const Vec4f& vec, const float s)
{
	Vec4f retVec;

	retVec.x = vec.x / s;
	retVec.y = vec.y / s;
	retVec.z = vec.z / s;
	retVec.w = vec.w / s;

	return retVec;
}

// Divisao com um escalar

inline Vec4f operator/(const float s, const Vec4f& vec)
{
	Vec4f retVec;

	retVec.x = vec.x / s;
	retVec.y = vec.y / s;
	retVec.z = vec.z / s;
	retVec.w = vec.w / s;

	return retVec;
}

// Multiplicacao por membro

inline Vec4f operator/(const Vec4f& vec1, const Vec4f& vec2)
{
	Vec4f retVec;

	retVec.x = vec1.x * vec2.x;
	retVec.y = vec1.y * vec2.y;
	retVec.z = vec1.z * vec2.z;
	retVec.w = vec1.w * vec2.w;

	return retVec;
}

// Negar o vector (produto com -1)

inline Vec4f operator-(const Vec4f& vec)
{
	Vec4f retVec;

	retVec.x = -vec.x;
	retVec.y = -vec.y;
	retVec.z = -vec.z;
	retVec.w = -vec.w;

	return retVec;
}


inline bool operator==(const Vec4f& vec1, const Vec4f& vec2)
{
	return ((vec1.x == vec2.x) &&
		(vec1.y == vec2.y) &&
		(vec1.z == vec2.z) &&
		(vec1.w == vec2.w)    );
}


inline bool is_equal(const Vec4f& vec1, const Vec4f& vec2,float epsilon)
{
	return ((fabs(vec1.x-vec2.x)<epsilon) &&
		(fabs(vec1.y-vec2.y)<epsilon) &&
		(fabs(vec1.z-vec2.z)<epsilon) &&
		(fabs(vec1.w-vec2.w)<epsilon));
}


inline bool operator!=(const Vec4f& vec1, const Vec4f& vec2)
{
	return ((vec1.x != vec2.x) ||
		(vec1.y != vec2.y) ||
		(vec1.z != vec2.z) ||
		(vec1.w != vec2.w)    );
}


inline float dot_product(const Vec4f& vec1, const Vec4f& vec2)
{
	return vec1.x * vec2.x + vec1.y * vec2.y + vec1.z * vec2.z + vec1.w * vec2.w;
}

// Produto com um escalar

inline Vec4f operator*(const Vec4f& vec1, const Vec4f& vec2)
{
	Vec4f retVec;

	retVec.x = vec1.x * vec2.x;
	retVec.y = vec1.y * vec2.y;
	retVec.z = vec1.z * vec2.z;
	retVec.w = vec1.w * vec2.w;

	return retVec;
}


inline float& Vec4f::operator [](const int i)
{
	if (i==0) return x;
	else if (i==1) return y;
	else if (i==2) return z;
	else return w;
}


inline float Vec4f::operator [](const int i) const
{
	if (i==0) return x;
	else if (i==1) return y;
	else if (i==2) return z;
	else return w;
}


inline bool Vec4f::inside(const Vec2f& v) const
{
	if ((v.x>=x1) && (v.x<=x2) &&
		(v.y>=y1) && (v.y<=y2)) return true;
	return false;
}


inline	void Vec4f::harmonize_rect()
{
	float tmp;
	if (x1>x2) { tmp=x1; x1=x2; x2=tmp; }
	if (y1>y2) { tmp=y1; y1=y2; y2=tmp; }
}

inline	Vec2f	rotate_around_point(Vec2f source,Vec2f origin,float ang)
{
	Vec2f disp=source-origin;

	Vec2f res;
	res.x=disp.x*cos(ang)-disp.y*sin(ang);
	res.y=disp.x*sin(ang)+disp.y*cos(ang);

	res=res+origin;

	return res;
}

inline	float	interpolate(const float t,const float v1,const float v2)
{
	return v1+t*(v2-v1);
}

inline	Vec2f	interpolate_v2(const float t,const Vec2f& v1,const Vec2f& v2)
{
	Vec2f	tmp(v1.x,v1.y);
	tmp.x+=t*(v2.x-v1.x);
	tmp.y+=t*(v2.y-v1.y);
	return tmp;
}

inline	Vec2f	interpolate_v2(const float t,const Vec2f* v1,const Vec2f* v2)
{
	Vec2f	tmp(v1->x,v1->y);
	tmp.x+=t*(v2->x-v1->x);
	tmp.y+=t*(v2->y-v1->y);
	return tmp;
}

inline	Vec3f	interpolate_v3(const float t,const Vec3f& v1,const Vec3f& v2)
{
	Vec3f	tmp(v1.x,v1.y,v1.z);
	tmp.x+=t*(v2.x-v1.x);
	tmp.y+=t*(v2.y-v1.y);
	tmp.z+=t*(v2.z-v1.z);
	return tmp;
}

inline	Vec3f	interpolate_v3(const float t,const Vec3f* v1,const Vec3f* v2)
{
	Vec3f	tmp(v1->x,v1->y,v1->z);
	tmp.x+=t*(v2->x-v1->x);
	tmp.y+=t*(v2->y-v1->y);
	tmp.z+=t*(v2->z-v1->z);
	return tmp;
}

inline	Vec4f	interpolate_v4(const float t,const Vec4f& v1,const Vec4f& v2)
{
	Vec4f	tmp(v1.x,v1.y,v1.z,v1.w);
	tmp.x+=t*(v2.x-v1.x);
	tmp.y+=t*(v2.y-v1.y);
	tmp.z+=t*(v2.z-v1.z);
	tmp.w+=t*(v2.w-v1.w);
	return tmp;
}

inline	Vec4f	interpolate_v4(const float t,const Vec4f* v1,const Vec4f* v2)
{
	Vec4f	tmp(v1->x,v1->y,v1->z,v1->w);
	tmp.x+=t*(v2->x-v1->x);
	tmp.y+=t*(v2->y-v1->y);
	tmp.z+=t*(v2->z-v1->z);
	tmp.w+=t*(v2->w-v1->w);
	return tmp;
}

typedef Vec4f	Color;
typedef Vec4f	Rect;

// Converts 4 floats representing a color to a colorref (DWORD)
inline unsigned long float_to_colorref(const float r,const float g,const float b,const float a) 
{ 
	return (int(a*255.0f) << 24)+(int(r*255.0f) << 16)+(int(g*255.0f) << 8)+(int(b*255.0f)); 
}
// Converts a 4d vector representing a color to a colorref (DWORD)
inline unsigned long float_to_colorref(const Color *color) 
{ 
	return ((int(color->a*255.0f)&0xFF) << 24)+((int(color->r*255.0f)&0xFF) << 16)+((int(color->g*255.0f)&0xFF) << 8)+((int(color->b*255.0f)&0xFF)); 
}

inline unsigned long float_to_colorref(const Color &color) 
{ 
	return ((int(color.a*255.0f)&0xFF) << 24)+((int(color.r*255.0f)&0xFF) << 16)+((int(color.g*255.0f)&0xFF) << 8)+((int(color.b*255.0f)&0xFF)); 
}

inline unsigned long float_to_colorref(const Vec3f &color) 
{ 
	return ((255 << 24)+
		((int(color.x*255.0f)&0xFF) << 16)+
		((int(color.y*255.0f)&0xFF) << 8)+
		((int(color.z*255.0f)&0xFF))); 
}

// Converts a colorref to a vector representating the color
inline Color colorref_to_float(const unsigned long color) 
{ 
	Color tmp;

	tmp.a=float((color & 0x00FF000000) >> 24)/255.0f;
	tmp.r=float((color & 0x0000FF0000) >> 16)/255.0f;
	tmp.g=float((color & 0x000000FF00) >> 8)/255.0f;
	tmp.b=float((color & 0x00000000FF))/255.0f;

	return tmp;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


class Quaternion
{
public:
	Quaternion();											
	Quaternion(const float _x, const float _y, const float _z, const float _w);		
	Quaternion(const Quaternion& o);		
	Quaternion(const D3DXQUATERNION& o);		

	void operator+=(const Quaternion& q);					///< Adds Quaternion q to this Quaternion
	void operator-=(const Quaternion& q);					///< Subtracts Quaternion q from this Quaternion

	void add(const Quaternion& q, const Quaternion& r);		///< This Quaternion will be set to the sum q + r
	void sub(const Quaternion& q, const Quaternion& r);		///< This Quaternion will be set to the subtraction q - r
	void mult(const Quaternion& q, const Quaternion& r);	///< This Quaternion will be set to the product qr

	void conj();											///< This Quaternion will be set to its conjugate
	void conj(const Quaternion& q);							///< This Quaternion will be set to the conjugate of q

	void invert();											///< This Quaternion will be inverted
	void invert(const Quaternion& q);						///< This Quaternion will be set to the inverse of q
	void invert(const Quaternion* q);
	void unit_invert();										///< This is assumed to be a unit Quaternion and it will be inverted (deprecated - same as invert)
	void unit_invert(const Quaternion& q);					///< This Quaternion will be set to the inverse of unit Quaternion q (deprecated - same as invert)

	float norm() const;											///< Returns the norm (there is an bug in this function)
	float normalize();										///< Normalizes the Quaternion

	void identity_quaternion();								///< This Quaternion will be set to the identity Quaternion
	bool is_identity() const;

	void rotation_quaternion(const float _x, const float _y, const float _z, const float ang);	///< This Quaternion will be set to a Quaternion representing a rotation of ang degrees (*not radians*) aroung vector (x,y,z)
	void rotation_quaternion(const Vec3f& vec, const float ang);				///< This Quaternion will be set to a Quaternion representing a rotation of ang degrees (*not radians*) aroung vector vec

	//	void euler_quaternion(const float yaw, const float pitch, const float roll);
	void to_euler(float& rotX,float& rotY,float& rotZ);

	void rotate_point(Vec3f& dst, const Vec3f& org) const;				///< The rotation encoded in the Quaternion will be applied to org and written to dst
	void slerp(const Quaternion& to, float t, Quaternion& res);		///< Interpolates betwwen this Quaternion and Quaternion to. t shoud be beeetween 0 and 1.

	void set(const float _x,const float _y,const float _z,const float _w);

public:
	float x, y, z, w;
};


inline Quaternion::Quaternion()
{
}


inline Quaternion::Quaternion(const float _x, const float _y, const float _z, const float _w)
{
	x = _x;
	y = _y;
	z = _z;
	w = _w;
}

inline Quaternion::Quaternion(const Quaternion& o)
{
	x=o.x;
	y=o.y;
	z=o.z;
	w=o.w;
}

inline Quaternion::Quaternion(const D3DXQUATERNION& o)
{
	x=o.x;
	y=o.y;
	z=o.z;
	w=o.w;
}

inline void Quaternion::identity_quaternion()
{
	x = y = z = 0.0f;
	w = 1.0f;
}


inline bool Quaternion::is_identity() const
{
	if ((x==0.0f) && (y==0.0f) && (z==0.0f) && (w==1.0f)) return true;
	return false;
}



inline void Quaternion::operator+=(const Quaternion& q)
{
	x += q.x;
	y += q.y;
	z += q.z;
	w += q.w;
}


inline void Quaternion::operator-=(const Quaternion& q)
{
	x -= q.x;
	y -= q.y;
	z -= q.z;
	w -= q.w;
}


inline bool operator!=(const Quaternion& q1, const Quaternion& q2)
{
	return ((fabs(q1.x-q2.x) > EPSILON_FLOAT) ||
		(fabs(q1.y-q2.y) > EPSILON_FLOAT) ||
		(fabs(q1.z-q2.z) > EPSILON_FLOAT) ||
		(fabs(q1.w-q2.w) > EPSILON_FLOAT));
}


inline void Quaternion::add(const Quaternion& q, const Quaternion& r)
{
	x = q.x + r.x;
	y = q.y + r.y;
	z = q.z + r.z;
	w = q.w + r.w;
}


inline void Quaternion::sub(const Quaternion& q, const Quaternion& r)
{
	x = q.x - r.x;
	y = q.y - r.y;
	z = q.z - r.z;
	w = q.w - r.w;
}


inline void Quaternion::mult(const Quaternion& q, const Quaternion& r)
{
	float A, B, C, D, E, F, G, H;

	A = (q.w + q.x)*(r.w + r.x);
	B = (q.z - q.y)*(r.y - r.z);
	C = (q.w - q.x)*(r.y + r.z); 
	D = (q.y + q.z)*(r.w - r.x);
	E = (q.x + q.z)*(r.x + r.y);
	F = (q.x - q.z)*(r.x - r.y);
	G = (q.w + q.y)*(r.w - r.z);
	H = (q.w - q.y)*(r.w + r.z);

	w = B + (-E - F + G + H) /2;
	x = A - (E + F + G + H)/2; 
	y = C + (E - F + G - H)/2; 
	z = D + (E - F - G + H)/2;
}



inline void Quaternion::conj()
{
	x = -x;
	y = -y;
	z = -z;
}


inline void Quaternion::conj(const Quaternion& q)
{
	x = -q.x;
	y = -q.y;
	z = -q.z;
	w = q.w;
}


inline float Quaternion::norm() const
{
	return x * x + y * y + z * z + w * w;	// Norma mal calculada
}


inline float Quaternion::normalize()
{
	float n=sqrt(x * x + y * y + z * z + w * w);

	x/=n;
	y/=n;
	z/=n;
	w/=n;

	return n;
}


inline void Quaternion::invert(const Quaternion& q)
{
	float inv_norm = 1.0f / q.norm();

	x = -q.x * inv_norm;
	y = -q.y * inv_norm;
	z = -q.z * inv_norm;

	w = q.w * inv_norm;
}


inline void Quaternion::invert(const Quaternion* q)
{
	float inv_norm = 1.0f / q->norm();

	x = -q->x * inv_norm;
	y = -q->y * inv_norm;
	z = -q->z * inv_norm;
	w =  q->w * inv_norm;
}


inline void Quaternion::invert()
{
	float inv_norm = 1.0f / norm();

	x *= -inv_norm;
	y *= -inv_norm;
	z *= -inv_norm;

	w *= inv_norm;
}



inline void Quaternion::unit_invert()
{
	conj();
}


inline void Quaternion::unit_invert(const Quaternion& q)
{
	conj(q);
	w = q.w;
}


inline void Quaternion::rotation_quaternion(const float _x, const float _y, const float _z, const float ang)
{
	float ang_2 = ang * 0.5f * PI_OVER_180;
	float sin_ang_2 = sin(ang_2);

	x = _x * sin_ang_2;
	y = _y * sin_ang_2;
	z = _z * sin_ang_2;
	w = cos(ang_2);
}


inline void Quaternion::rotation_quaternion(const Vec3f& vec, const float a)
{
	float ang = -a;		
	float ang_2 = ang * 0.5f * PI_OVER_180;
	float sin_ang_2 = sin(ang_2);
	Vec3f v = vec;
	v.normalize();

	x = v.x * sin_ang_2;
	y = v.y * sin_ang_2;
	z = v.z * sin_ang_2;
	w = cos(ang_2);
}

// Assumes a unit Quaternion

inline void Quaternion::rotate_point(Vec3f& dst, const Vec3f& org) const
{
	Quaternion p(org.x, org.y, org.z,0);		// NOTA: Optimizar isto 'a mao ...
	Quaternion r,r1,c;

	c.conj(*this);
	r1.mult(*this,p);
	r.mult(r1,c);

	dst.x = r.x;
	dst.y = r.y;
	dst.z = r.z;	
}


inline void Quaternion::set(const float _x,const float _y,const float _z,const float _w)
{
	x=_x;
	y=_y;
	z=_z;
	w=_w;
}


inline void Quaternion::slerp(const Quaternion& to, float t, Quaternion& res)
{
	float	to1[4];
	float	omega, cosom, sinom, scale0, scale1;
	const float DELTA = 0.0001f;


	// calc cosine
	cosom = x * to.x + y * to.y + z * to.z + w * to.w;

	// adjust signs (if necessary)

	if ( cosom <0.0 ){ cosom = -cosom; to1[0] = - to.x;
	to1[1] = - to.y;
	to1[2] = - to.z;
	to1[3] = - to.w;
	} else  {
		to1[0] = to.x;
		to1[1] = to.y;
		to1[2] = to.z;
		to1[3] = to.w;
	}

	// calculate coefficients
	if ( (1.0f - cosom) > DELTA ) {
		// standard case (slerp)
		omega = acos(cosom);
		sinom = sin(omega);
		scale0 = sin((1.0f - t) * omega) / sinom;
		scale1 = sin(t * omega) / sinom;

	} else {        
		// "from" and "to" Quaternions are very close 
		//  ... so we can do a linear unsigned charerpolation
		scale0 = 1.0f - t;
		scale1 = t;
	}

	// calculate final values
	res.x = float(scale0 * x + scale1 * to1[0]);
	res.y = float(scale0 * y + scale1 * to1[1]);
	res.z = float(scale0 * z + scale1 * to1[2]);
	res.w = float(scale0 * w + scale1 * to1[3]);
}


inline Quaternion vector_rotation(const Vec3f& f,const Vec3f& t) 
{
	Vec3f from=f;
	from.normalize();
	Vec3f to=t;
	to.normalize();

	if (from==to)
	{
		return Quaternion(0,0,0,1);
	}

	Vec3f rot;
	if (from==-to)
	{
		// Existem infinitas escolhas para o vector de rota�o, ERRO ...
		rot=Vec3f(0.0f,1.0f,0.0f);
	}
	else
	{
		rot=cross_product(from,to);
	}

	float cos_2ang=dot_product(from,to);
	rot.normalize();
	float sin_ang=sqrt((1 - cos_2ang) / 2);
	float cos_ang=sqrt(1 - sin_ang * sin_ang);
	return Quaternion(rot.x*sin_ang,rot.y*sin_ang,rot.z*sin_ang,cos_ang);
}


inline Quaternion vector_to_quaternion(Vec3f& vec) 
{
	Vec3f tmp(0.0f,0.0f,1.0f);

	return vector_rotation(tmp,vec);
}


inline	Quaternion	helper_get_ori2d(const Vec3f& vec)
{
	// Look at
	// Rot around y axis
	// Get "base" view vector
	Vec3f	delta=vec; delta.y=0.0f; delta.normalize();
	// Get angle around Y axis
	float	ang_y=delta.z;
	if (ang_y<-1.0f) ang_y=-1.0f;
	else if (ang_y>1.0f) ang_y=1.0f;
	ang_y=(float)rad_to_deg((float)acos(ang_y));
	// Rotate around Y axis
	Quaternion ori; ori.identity_quaternion(); 
	if (delta.x<0.0f) ori.rotation_quaternion(0.0f,1.0f,0.0f,-ang_y);
	else ori.rotation_quaternion(0.0f,1.0f,0.0f,ang_y);
	return ori;
}


inline	float	helper_get_angle_from_xz(const Vec3f& vec)
{
	// Look at
	// Rot around y axis
	// Get "base" view vector
	Vec3f	delta=vec; delta.y=0.0f; delta.normalize();
	// Get angle around Y axis
	float	ang_y=delta.x;
	if (ang_y<-1.0f) ang_y=-1.0f;
	else if (ang_y>1.0f) ang_y=1.0f;
	ang_y=(float)rad_to_deg((float)acos(ang_y));
	// Rotate around Y axis
	if (delta.z<0.0f) return -ang_y;
	
	return ang_y;
}
// ********************************************************************/
// Menos eficientes - use with care !!!


inline Quaternion operator+(const Quaternion& q, const Quaternion& r)
{
	Quaternion ret;

	ret.x = q.x + r.x;
	ret.y = q.y + r.y;
	ret.z = q.z + r.z;
	ret.w = q.w + r.w;

	return ret;
}


inline Quaternion operator-(const Quaternion& q, const Quaternion& r)
{
	Quaternion ret;

	ret.x = q.x - r.x;
	ret.y = q.y - r.y;
	ret.z = q.z - r.z;
	ret.w = q.w - r.w;

	return ret;
}


inline Quaternion operator*(const Quaternion& q, const Quaternion& r)
{
	Quaternion ret;

	float A, B, C, D, E, F, G, H;

	A = (q.w + q.x)*(r.w + r.x);
	B = (q.z - q.y)*(r.y - r.z);
	C = (q.w - q.x)*(r.y + r.z); 
	D = (q.y + q.z)*(r.w - r.x);
	E = (q.x + q.z)*(r.x + r.y);
	F = (q.x - q.z)*(r.x - r.y);
	G = (q.w + q.y)*(r.w - r.z);
	H = (q.w - q.y)*(r.w + r.z);

	ret.w = B + (-E - F + G + H) /2;
	ret.x = A - (E + F + G + H)/2; 
	ret.y = C + (E - F + G - H)/2; 
	ret.z = D + (E - F - G + H)/2;

	return ret;
}


inline void Quaternion::to_euler(float& rotX,float& rotY,float& rotZ)
{
	rotX=(float)rad_to_deg((float)(atan((2.0f*(x*w+y*z))/(1.0f-2.0f*(Sqr(z)+Sqr(w))))));
	rotY=(float)rad_to_deg((float)(asin((2.0f*(x*z-y*w)))));
	rotZ=-(float)rad_to_deg((float)(atan((2.0f*(x*y+z*w))/(1.0f-2.0f*(Sqr(y)+Sqr(z))))));
}


inline Quaternion operator-(const Quaternion& q)
{
	Quaternion ret;

	float inv_norm=1.0f /q.norm();

	ret.x=-inv_norm*q.x;
	ret.y=-inv_norm*q.y;
	ret.z=-inv_norm*q.z;

	ret.w=inv_norm*q.w;

	return ret;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
class Interpolator
{
protected:
	struct Elem
	{
		float	t;
		T		v;
	};
	typedef std::vector<Elem> Elems;
public:
	inline	~Interpolator() { _elems.clear(); }

	inline	void	add(float t,const T& v) { Elem elem; elem.t=t; elem.v=v; _elems.push_back(elem); }
	inline	T		get(float t) 
	{
		size_t	lower_bound=0xFFFF;
		T		ret;

		Elems::reverse_iterator r=_elems.rbegin();
		while (r!=_elems.rend())
		{
			if (r->t<t) break;
			++r;
		}
		if (r==_elems.rend()) return T();
		
		// R has the start t (t0)
		float t0=(*r).t;

		Elems::iterator e=r.base();
		if (e==_elems.end()) { return r->v; }

		float t1=(*e).t;

		float tt=(t-t0)/(t1-t0);

		ret=r->v*(1-tt)+e->v*tt;

		return ret;
	} 

protected:
	Elems	_elems;
};
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class Matrix  
{
public:
	inline Matrix() { ; }
	inline Matrix(const Vec3f& pos,const Quaternion& ori,const Vec3f& scale) { set(pos,ori,scale); }
	inline ~Matrix() { ; }

	inline	void	set(const Vec4f& v1,const Vec4f& v2,const Vec4f& v3,const Vec4f& v4) 
	{ 
		_m[0][0]=v1.x; _m[1][0]=v1.y; _m[2][0]=v1.z; _m[3][0]=v1.w; 
		_m[0][1]=v2.x; _m[1][1]=v2.y; _m[2][1]=v2.z; _m[3][1]=v2.w; 
		_m[0][2]=v3.x; _m[1][2]=v3.y; _m[2][2]=v3.z; _m[3][2]=v3.w; 
		_m[0][3]=v4.x; _m[1][3]=v4.y; _m[2][3]=v4.z; _m[3][3]=v4.w; 
	}

	inline	void	set(const float f11,const float f12,const float f13,const float f14,
						const float f21,const float f22,const float f23,const float f24,
						const float f31,const float f32,const float f33,const float f34,
						const float f41,const float f42,const float f43,const float f44)
	{ 
		_m[0][0]=f11; _m[1][0]=f21; _m[2][0]=f31; _m[3][0]=f41; 
		_m[0][1]=f12; _m[1][1]=f22; _m[2][1]=f32; _m[3][1]=f42; 
		_m[0][2]=f13; _m[1][2]=f23; _m[2][2]=f33; _m[3][2]=f43; 
		_m[0][3]=f14; _m[1][3]=f24; _m[2][3]=f34; _m[3][3]=f44; 
	}
	inline 	void	set(const Vec3f& pos,const Quaternion& ori);
	inline 	void	set(const Vec3f& pos,const Quaternion& ori,const Vec3f& scale);
	inline 	void	set(const Quaternion& ori);
	
	inline	void		identity();
	inline	void		zero();
	inline	void		transpose() { Matrix tmp; tmp.set(_m[0][0],_m[1][0],_m[2][0],_m[3][0],_m[0][1],_m[1][1],_m[2][1],_m[3][1],_m[0][2],_m[1][2],_m[2][2],_m[3][2],_m[0][3],_m[1][3],_m[2][3],_m[3][3]); memcpy(&_m,&(tmp._m),16*sizeof(float)); }
	inline	void		invert();
	inline	void		pre_multiply_3x3(Vec3f* vec) const;
	inline	void		pre_multiply_4x4(Vec3f* vec) const;
	inline	void		pre_multiply_4x4(Vec4f* vec) const;
	inline	void		pre_multiply_3x3(Vec3f& dest,const Vec3f& vec) const;
	inline	void		pre_multiply_4x4(Vec3f& dest,const Vec3f& vec) const;
	inline	void		pre_multiply_4x4(Vec4f& dest,const Vec4f& vec) const;
			Matrix&		operator*=(float v);

			void		log();

			float		det();

			void		look_at(const Vec3f& from,const Vec3f& to,const Vec3f& up_axis);

public:
	float	_m[4][4];

public:
	static	Matrix	IDENTITY;
};

inline void Matrix::identity() 
{
	zero();
	_m[0][0]=_m[1][1]=_m[2][2]=_m[3][3]=1.0f;
}

inline void Matrix::zero() 
{
	memset(&_m,0,sizeof(float)*16);
}

inline Matrix operator+(const Matrix& mat1, const Matrix& mat2)
{
	Matrix ret;

	for (unsigned char i=0; i<4; i++)
		for (unsigned char j=0; j<4; j++) ret._m[i][j]=mat1._m[i][j]+mat2._m[i][j];

	return ret;
}

inline Matrix operator*(const Matrix& mat1, const Matrix& mat2)
{
	Matrix	ret;
	for (unsigned char iRow=0; iRow<4; iRow++)
	{
		for (unsigned char iCol=0; iCol<4; iCol++)
		{
			ret._m[iCol][iRow] =
				mat1._m[iCol][0]*mat2._m[0][iRow] +
				mat1._m[iCol][1]*mat2._m[1][iRow] +
				mat1._m[iCol][2]*mat2._m[2][iRow] +
				mat1._m[iCol][3]*mat2._m[3][iRow];
		}
	}
	return ret;
}

inline void Matrix::pre_multiply_3x3(Vec3f& dest,const Vec3f& vec) const
{
	Vec3f ret;

	ret.x=vec.x*_m[0][0]+vec.y*_m[1][0]+vec.z*_m[2][0];
	ret.y=vec.x*_m[0][1]+vec.y*_m[1][1]+vec.z*_m[2][1];
	ret.z=vec.x*_m[0][2]+vec.y*_m[1][2]+vec.z*_m[2][2];

	dest.x=ret.x;
	dest.y=ret.y;
	dest.z=ret.z;
}

inline void Matrix::pre_multiply_4x4(Vec3f& dest,const Vec3f& vec) const
{
	Vec3f ret;

	ret.x=vec.x*_m[0][0]+vec.y*_m[1][0]+vec.z*_m[2][0]+_m[3][0];
	ret.y=vec.x*_m[0][1]+vec.y*_m[1][1]+vec.z*_m[2][1]+_m[3][1];
	ret.z=vec.x*_m[0][2]+vec.y*_m[1][2]+vec.z*_m[2][2]+_m[3][2];

	dest.x=ret.x;
	dest.y=ret.y;
	dest.z=ret.z;
}

inline void Matrix::pre_multiply_4x4(Vec4f& dest,const Vec4f& vec) const
{
	Vec4f ret;

	ret.x=vec.x*_m[0][0]+vec.y*_m[1][0]+vec.z*_m[2][0]+vec.w*_m[3][0];
	ret.y=vec.x*_m[0][1]+vec.y*_m[1][1]+vec.z*_m[2][1]+vec.w*_m[3][1];
	ret.z=vec.x*_m[0][2]+vec.y*_m[1][2]+vec.z*_m[2][2]+vec.w*_m[3][2];
	ret.w=vec.x*_m[0][3]+vec.y*_m[1][3]+vec.z*_m[2][3]+vec.w*_m[3][3];

	dest.x=ret.x;
	dest.y=ret.y;
	dest.z=ret.z;
	dest.w=ret.w;
}

void Matrix::pre_multiply_3x3(Vec3f* vec) const
{
	Vec3f ret;

	ret.x=vec->x*_m[0][0]+vec->y*_m[1][0]+vec->z*_m[2][0];
	ret.y=vec->x*_m[0][1]+vec->y*_m[1][1]+vec->z*_m[2][1];
	ret.z=vec->x*_m[0][2]+vec->y*_m[1][2]+vec->z*_m[2][2];

	vec->x=ret.x;
	vec->y=ret.y;
	vec->z=ret.z;
}

void Matrix::pre_multiply_4x4(Vec3f* vec) const
{
	Vec3f ret;

	ret.x=vec->x*_m[0][0]+vec->y*_m[1][0]+vec->z*_m[2][0]+_m[3][0];
	ret.y=vec->x*_m[0][1]+vec->y*_m[1][1]+vec->z*_m[2][1]+_m[3][1];
	ret.z=vec->x*_m[0][2]+vec->y*_m[1][2]+vec->z*_m[2][2]+_m[3][2];

	vec->x=ret.x;
	vec->y=ret.y;
	vec->z=ret.z;
}

void Matrix::pre_multiply_4x4(Vec4f* vec) const
{
	Vec4f ret;

	ret.x=vec->x*_m[0][0]+vec->y*_m[1][0]+vec->z*_m[2][0]+vec->w*_m[3][0];
	ret.y=vec->x*_m[0][1]+vec->y*_m[1][1]+vec->z*_m[2][1]+vec->w*_m[3][1];
	ret.z=vec->x*_m[0][2]+vec->y*_m[1][2]+vec->z*_m[2][2]+vec->w*_m[3][2];
	ret.w=vec->x*_m[0][3]+vec->y*_m[1][3]+vec->z*_m[2][3]+vec->w*_m[3][3];

	vec->x=ret.x;
	vec->y=ret.y;
	vec->z=ret.z;
	vec->w=ret.w;
}

inline Matrix& Matrix::operator*=(const float s)
{
    // Scale coefficients
	for (unsigned char i=0; i<4; i++)
		for (unsigned char j=0; j<4; j++) _m[i][j]*=s;

    return *this;
}

inline void Matrix::set(const Vec3f& pos,const Quaternion& ori)
{
	float xx=ori.x*ori.x;
	float xy=ori.x*ori.y;
	float xz=ori.x*ori.z;
	float xw=ori.x*ori.w;

	float yy=ori.y*ori.y;
	float yz=ori.y*ori.z;
	float yw=ori.y*ori.w;

	float zz=ori.z*ori.z;
	float zw=ori.z*ori.w;

//	float ww=ori.w*ori.w;

	_m[0][0]=1.0f-2.0f*(yy+zz);
	_m[1][0]=     2.0f*(xy-zw);
	_m[2][0]=     2.0f*(xz+yw);
	_m[0][1]=     2.0f*(xy+zw);
	_m[1][1]=1.0f-2.0f*(xx+zz);
	_m[2][1]=     2.0f*(yz-xw);
	_m[0][2]=     2.0f*(xz-yw);
	_m[1][2]=     2.0f*(yz+xw);
	_m[2][2]=1.0f-2.0f*(xx+yy);
	_m[0][3]=0.0f;
	_m[1][3]=0.0f;
	_m[2][3]=0.0f;
	_m[3][0]=pos.x;
	_m[3][1]=pos.y;
	_m[3][2]=pos.z;
	_m[3][3]=1.0f;
}

inline void Matrix::set(const Vec3f& pos,const Quaternion& ori,const Vec3f& scale)
{
	float xx=ori.x*ori.x;
	float xy=ori.x*ori.y;
	float xz=ori.x*ori.z;
	float xw=ori.x*ori.w;

	float yy=ori.y*ori.y;
	float yz=ori.y*ori.z;
	float yw=ori.y*ori.w;

	float zz=ori.z*ori.z;
	float zw=ori.z*ori.w;

	float ww=ori.w*ori.w;

	_m[0][0]=(1.0f-2.0f*(yy+zz))*scale.x; 
	_m[1][0]=(     2.0f*(xy-zw))*scale.y; 
	_m[2][0]=(     2.0f*(xz+yw))*scale.z; 
	_m[0][1]=(     2.0f*(xy+zw))*scale.x; 
	_m[1][1]=(1.0f-2.0f*(xx+zz))*scale.y;  
	_m[2][1]=(     2.0f*(yz-xw))*scale.z; 
	_m[0][2]=(     2.0f*(xz-yw))*scale.x; 
	_m[1][2]=(     2.0f*(yz+xw))*scale.y; 
	_m[2][2]=(1.0f-2.0f*(xx+yy))*scale.z; 
	_m[0][3]=0.0f;
	_m[1][3]=0.0f;
	_m[2][3]=0.0f;
	_m[3][0]=pos.x;
	_m[3][1]=pos.y;
	_m[3][2]=pos.z;
	_m[3][3]=1.0f;
}

inline void Matrix::set(const Quaternion& ori)
{
	set(Vec3f(0.0f,0.0f,0.0f),ori,Vec3f(1.0f,1.0f,1.0f));
}

inline void Matrix::invert() 
{
	float tmp[12]; /* temp array for pairs */
	float src[16]; /* array of transpose source matrix */
	float dst[16];
	float det; /* determinant */
	/* transpose matrix */
	for (unsigned char i=0;i<4; i++) 
	{
		for (unsigned char j=0; j<4; j++)
		{
			src[i*4+j]=_m[j][i];
		}
	}
	/* calculate pairs for first 8 elements (cofactors) */
	tmp[0]=src[10]*src[15];
	tmp[1]=src[11]*src[14];
	tmp[2]=src[9]*src[15];
	tmp[3]=src[11]*src[13];
	tmp[4]=src[9]*src[14];
	tmp[5]=src[10]*src[13];
	tmp[6]=src[8]*src[15];
	tmp[7]=src[11]*src[12];
	tmp[8]=src[8]*src[14];
	tmp[9]=src[10]*src[12];
	tmp[10]=src[8]*src[13];
	tmp[11]=src[9]*src[12];
	/* calculate first 8 elements (cofactors) */
	dst[0]=tmp[0]*src[5]+tmp[3]*src[6]+tmp[4]*src[7];
	dst[0]-=tmp[1]*src[5]+tmp[2]*src[6]+tmp[5]*src[7];
	dst[1]=tmp[1]*src[4]+tmp[6]*src[6]+tmp[9]*src[7];
	dst[1]-=tmp[0]*src[4]+tmp[7]*src[6]+tmp[8]*src[7];
	dst[2]=tmp[2]*src[4]+tmp[7]*src[5]+tmp[10]*src[7];
	dst[2]-=tmp[3]*src[4]+tmp[6]*src[5]+tmp[11]*src[7];
	dst[3]=tmp[5]*src[4]+tmp[8]*src[5]+tmp[11]*src[6];
	dst[3]-=tmp[4]*src[4]+tmp[9]*src[5]+tmp[10]*src[6];
	dst[4]=tmp[1]*src[1]+tmp[2]*src[2]+tmp[5]*src[3];
	dst[4]-=tmp[0]*src[1]+tmp[3]*src[2]+tmp[4]*src[3];
	dst[5]=tmp[0]*src[0]+tmp[7]*src[2]+tmp[8]*src[3];
	dst[5]-=tmp[1]*src[0]+tmp[6]*src[2]+tmp[9]*src[3];
	dst[6]=tmp[3]*src[0]+tmp[6]*src[1]+tmp[11]*src[3];
	dst[6]-=tmp[2]*src[0]+tmp[7]*src[1]+tmp[10]*src[3];
	dst[7]=tmp[4]*src[0]+tmp[9]*src[1]+tmp[10]*src[2];
	dst[7]-=tmp[5]*src[0]+tmp[8]*src[1]+tmp[11]*src[2];
	/* calculate pairs for second 8 elements (cofactors) */
	tmp[0]=src[2]*src[7];
	tmp[1]=src[3]*src[6];
	tmp[2]=src[1]*src[7];
	tmp[3]=src[3]*src[5];
	tmp[4]=src[1]*src[6];
	tmp[5]=src[2]*src[5];
	tmp[6]=src[0]*src[7];
	tmp[7]=src[3]*src[4];
	tmp[8]=src[0]*src[6];
	tmp[9]=src[2]*src[4];
	tmp[10]=src[0]*src[5];
	tmp[11]=src[1]*src[4];
	/* calculate second 8 elements (cofactors) */
	dst[8]=tmp[0]*src[13]+tmp[3]*src[14]+tmp[4]*src[15];
	dst[8]-=tmp[1]*src[13]+tmp[2]*src[14]+tmp[5]*src[15];
	dst[9]=tmp[1]*src[12]+tmp[6]*src[14]+tmp[9]*src[15];
	dst[9]-=tmp[0]*src[12]+tmp[7]*src[14]+tmp[8]*src[15];
	dst[10]=tmp[2]*src[12]+tmp[7]*src[13]+tmp[10]*src[15];
	dst[10]-=tmp[3]*src[12]+tmp[6]*src[13]+tmp[11]*src[15];
	dst[11]=tmp[5]*src[12]+tmp[8]*src[13]+tmp[11]*src[14];
	dst[11]-=tmp[4]*src[12]+tmp[9]*src[13]+tmp[10]*src[14];
	dst[12]=tmp[2]*src[10]+tmp[5]*src[11]+tmp[1]*src[9];
	dst[12]-=tmp[4]*src[11]+tmp[0]*src[9]+tmp[3]*src[10];
	dst[13]=tmp[8]*src[11]+tmp[0]*src[8]+tmp[7]*src[10];
	dst[13]-=tmp[6]*src[10]+tmp[9]*src[11]+tmp[1]*src[8];
	dst[14]=tmp[6]*src[9]+tmp[11]*src[11]+tmp[3]*src[8];
	dst[14]-=tmp[10]*src[11]+tmp[2]*src[8]+tmp[7]*src[9];
	dst[15]=tmp[10]*src[10]+tmp[4]*src[8]+tmp[9]*src[9];
	dst[15]-=tmp[8]*src[9]+tmp[11]*src[10]+tmp[5]*src[8];
	/* calculate determinant */
	det=src[0]*dst[0]+src[1]*dst[1]+src[2]*dst[2]+src[3]*dst[3];
	/* calculate matrix inverse */
	det=1/det;
	for (unsigned char j=0; j<16; j++) dst[j]*=det;

	memcpy(&(_m[0][0]),&(dst[0]),sizeof(float)*16);
}

inline bool operator==(const Matrix& mat1,const Matrix& mat2)
{
	if (mat1._m[0][0]!=mat2._m[0][0]) return false;
	if (mat1._m[0][1]!=mat2._m[0][1]) return false;
	if (mat1._m[0][2]!=mat2._m[0][2]) return false;
	if (mat1._m[0][3]!=mat2._m[0][3]) return false;
	if (mat1._m[1][0]!=mat2._m[1][0]) return false;
	if (mat1._m[1][1]!=mat2._m[1][1]) return false;
	if (mat1._m[1][2]!=mat2._m[1][2]) return false;
	if (mat1._m[1][3]!=mat2._m[1][3]) return false;
	if (mat1._m[2][0]!=mat2._m[2][0]) return false;
	if (mat1._m[2][1]!=mat2._m[2][1]) return false;
	if (mat1._m[2][2]!=mat2._m[2][2]) return false;
	if (mat1._m[2][3]!=mat2._m[2][3]) return false;
	if (mat1._m[3][0]!=mat2._m[3][0]) return false;
	if (mat1._m[3][1]!=mat2._m[3][1]) return false;
	if (mat1._m[3][2]!=mat2._m[3][2]) return false;
	if (mat1._m[3][3]!=mat2._m[3][3]) return false;

	return true;
}

inline bool operator!=(const Matrix& mat1,const Matrix& mat2)
{
	if (mat1._m[0][0]!=mat2._m[0][0]) return true;
	if (mat1._m[0][1]!=mat2._m[0][1]) return true;
	if (mat1._m[0][2]!=mat2._m[0][2]) return true;
	if (mat1._m[0][3]!=mat2._m[0][3]) return true;
	if (mat1._m[1][0]!=mat2._m[1][0]) return true;
	if (mat1._m[1][1]!=mat2._m[1][1]) return true;
	if (mat1._m[1][2]!=mat2._m[1][2]) return true;
	if (mat1._m[1][3]!=mat2._m[1][3]) return true;
	if (mat1._m[2][0]!=mat2._m[2][0]) return true;
	if (mat1._m[2][1]!=mat2._m[2][1]) return true;
	if (mat1._m[2][2]!=mat2._m[2][2]) return true;
	if (mat1._m[2][3]!=mat2._m[2][3]) return true;
	if (mat1._m[3][0]!=mat2._m[3][0]) return true;
	if (mat1._m[3][1]!=mat2._m[3][1]) return true;
	if (mat1._m[3][2]!=mat2._m[3][2]) return true;
	if (mat1._m[3][3]!=mat2._m[3][3]) return true;

	return false;
}



#endif