// ModelManager.cpp: implementation of the ModelManager class.
//
//////////////////////////////////////////////////////////////////////

#include "ModelManager.h"
#include "Utils.h"
#include "Image_A8R8G8B8.h"
#include "VoxelWorld.h"
#include "Camera.h"
#include "Light.h"

ModelManager* MODEL_MANAGER=NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ModelManager::ModelManager()
{
	if (!MODEL_MANAGER) MODEL_MANAGER=this;
}

ModelManager::~ModelManager()
{
	clear_all();
	if (MODEL_MANAGER==this) MODEL_MANAGER=NULL;
}

long ModelManager::new_voxel_model(const char *name)
{
	long	model_id=find_by_name(name);
	Model*	model=NULL;
	if (model_id!=-1)
	{
		clear(model_id);
	}

	model=new Model;
	for (size_t i=0; i<_models.size(); i++)
	{
		if (!_models[i])
		{
			model_id=i;
			_models[i]=model;
		}
	}
	if (model_id==-1)
	{
		_models.push_back(model);
		model_id=_models.size()-1;
	}
	
	model->_name=_strdup(name);
	model->_voxel_model=NULL;
	model->_voxel_build=new VoxelBuild;
	model->_set_aabb=false;

	return model_id;
}

long ModelManager::new_build_geom_model(const char *name)
{
	long	model_id=find_by_name(name);
	Model*	model=NULL;
	if (model_id!=-1)
	{
		clear(model_id);
	}

	model=new Model;
	for (size_t i=0; i<_models.size(); i++)
	{
		if (!_models[i])
		{
			model_id=i;
			_models[i]=model;
		}
	}
	if (model_id==-1)
	{
		_models.push_back(model);
		model_id=_models.size()-1;
	}

	model->_name=_strdup(name);
	model->_geom_model=NULL;
	model->_geom_build=new GeomBuild;
	model->_set_aabb=false;

	return model_id;
}

long ModelManager::load_3ds(const char *name,const char* filename)
{
	std::vector<BaseMesh*>	meshes;
	std::vector<Camera*>	cameras;
	std::vector<Light*>		lights;
	if (!(::load_3ds(filename,&meshes,&cameras,&lights)))
	{
		char buffer[512];
		sprintf_s((char*)&buffer,512,"Failed to load 3DS \"%s\"!",filename);
		MessageBox(NULL,buffer,"Error",MB_OK|MB_ICONERROR);
		return -1;
	}

	for (size_t i=0; i<cameras.size(); i++) delete cameras[i];
	for (size_t i=0; i<lights.size(); i++) delete lights[i];
	cameras.clear();
	lights.clear();

	if (meshes.size()==0)
	{
		return -1;
	}
	
	long	model_id=find_by_name(name);
	Model*	model=NULL;
	if (model_id!=-1)
	{
		clear(model_id);
	}

	model=new Model;
	model->_name=_strdup(name);
	model->_mesh_model=meshes[0];
	model->_set_aabb=false;

	if (model_id==-1)
	{
		_models.push_back(model);
		model_id=_models.size()-1;
	}

	for (size_t i=1; i<meshes.size(); i++) delete meshes[i];
	meshes.clear();

	return model_id;
}

long ModelManager::add_model(const char* name,BaseMesh* mesh)
{
	long	model_id=find_by_name(name);
	Model*	model=NULL;
	if (model_id!=-1)
	{
		clear(model_id);
	}

	model=new Model;
	model->_name=_strdup(name);
	model->_mesh_model=mesh;
	model->_set_aabb=false;

	if (model_id==-1)
	{
		_models.push_back(model);
		model_id=_models.size()-1;
	}

	return model_id;
}

long ModelManager::find_by_name(const char *name)
{
	for (unsigned long i=0; i<_models.size(); i++)
	{
		if (_models[i]) 
		{
			if (_models[i]->_name)
			{
				if (!_stricmp(_models[i]->_name,name)) return i;
			}
		}
	}

	return -1;
}

void ModelManager::clear(const char* name)
{
	long model_id=find_by_name(name);
	if (model_id!=-1)
	{
		clear(model_id);
	}
}

void ModelManager::clear(long id)
{
	if (id==-1) return;
	if ((long)_models.size()<=id) return;

	if (_models[id]) 
	{
		free(_models[id]->_name);
		if (_models[id]->_voxel_model)
		{
			delete _models[id]->_voxel_model;
			_models[id]->_voxel_model=NULL;
		}
		if (_models[id]->_voxel_build)
		{
			delete _models[id]->_voxel_build;
			_models[id]->_voxel_build=NULL;
		}
		if (_models[id]->_geom_model)
		{
			delete _models[id]->_geom_model;
			_models[id]->_geom_model=NULL;
		}
		if (_models[id]->_geom_build)
		{
			delete _models[id]->_geom_build;
			_models[id]->_geom_build=NULL;
		}
		delete _models[id];
		_models[id]=NULL;
	}
}

void ModelManager::clear_all()
{
	for (unsigned long i=0; i<_models.size(); i++)
	{
		clear(i);
	}
	_models.clear();
}

void ModelManager::draw(const long model_id,const Vec3f& pos,const Quaternion& ori,const Vec3f& scale,Camera* camera,Material* override_material)
{
	Model* model=get_model(model_id);
	if (!model) return;

	if (model->_voxel_model)
	{
		model->_voxel_model->_pos=pos;
		model->_voxel_model->_ori=ori;
//		model->_voxel_model->_scale=scale;
		model->_voxel_model->render(camera);
	}
	else if (model->_geom_model)
	{
		GFX->set_world_transformation(0,pos.x,pos.y,pos.z,ori,scale.x,scale.y,scale.z);
		model->_geom_model->draw();
		GFX->unset_world_transformation(0);
	}
	else if (model->_mesh_model)
	{
		GFX->set_world_transformation(0,pos.x,pos.y,pos.z,ori,scale.x,scale.y,scale.z);
		model->_mesh_model->draw(override_material);
		GFX->unset_world_transformation(0);
	}
}

void ModelManager::model_build(const long model_id)
{
	Model* model=get_model(model_id);
	if (!model) return;

	if (model->_voxel_build)
	{
		Vec3f	vr,vu,vd;
		bool	inv_x=false,inv_y=false,inv_z=false;
		switch (model->_voxel_build->_dir) 
		{
			case EXTRUSION_DIR_PX: vr.set(0.0f,1.0f,0.0f); vu.set(0.0f,0.0f,1.0f); vd.set(1.0f,0.0f,0.0f); break;
			case EXTRUSION_DIR_PY: vr.set(1.0f,0.0f,0.0f); vu.set(0.0f,1.0f,0.0f); vd.set(0.0f,0.0f,1.0f); inv_z=true; break;
			case EXTRUSION_DIR_PZ: vr.set(1.0f,0.0f,0.0f); vu.set(0.0f,0.0f,1.0f); vd.set(0.0f,1.0f,0.0f); inv_y=true; break;
			case EXTRUSION_DIR_NX: vr.set(0.0f,1.0f,0.0f); vu.set(0.0f,0.0f,1.0f); vd.set(1.0f,0.0f,0.0f); inv_x=true; break;
			case EXTRUSION_DIR_NY: vr.set(1.0f,0.0f,0.0f); vu.set(0.0f,1.0f,0.0f); vd.set(0.0f,0.0f,1.0f); inv_z=true; inv_y=true; break;
			case EXTRUSION_DIR_NZ: vr.set(1.0f,0.0f,0.0f); vu.set(0.0f,0.0f,1.0f); vd.set(0.0f,1.0f,0.0f); inv_z=true; break;
		}

		// Load images and get pixel size
		long pixel_size_x=0;
		long pixel_size_y=0;
		long min_height=0xFF;
		long max_height=0;

		std::vector<Image_A8R8G8B8*>	images;
		std::vector<Image_A8R8G8B8*>	himages;
		VoxelBuild*						build=model->_voxel_build;
		for (size_t i=0; i<build->_slices.size(); i++)
		{
			if (build->_slices[i].h1<min_height) min_height=build->_slices[i].h1;
			if (build->_slices[i].h2>max_height) max_height=build->_slices[i].h2;

			Image_A8R8G8B8* img=new Image_A8R8G8B8;
			if (!img->load_dds(build->_slices[i].color_image.c_str()))
			{
				char buffer[512];
				sprintf_s((char*)&buffer,512,"Failed to load DDS \"%s\"!",build->_slices[i].color_image.c_str());
				MessageBox(NULL,buffer,"Error",MB_OK|MB_ICONERROR);

				for (size_t i=0; i<images.size(); i++) delete images[i];
				for (size_t i=0; i<himages.size(); i++) if (himages[i]) delete himages[i];
				images.clear(); himages.clear();
				delete model->_voxel_build;
				return;
			}
			if ((int)img->get_size_x()>pixel_size_x) pixel_size_x=img->get_size_x();
			if ((int)img->get_size_y()>pixel_size_y) pixel_size_y=img->get_size_y();
			images.push_back(img);

			if (build->_slices[i].height_image!="")
			{
				img=new Image_A8R8G8B8;
				if (!img->load_dds(build->_slices[i].height_image.c_str()))
				{
					char buffer[512];
					sprintf_s((char*)&buffer,512,"Failed to load DDS \"%s\"!",build->_slices[i].height_image.c_str());
					MessageBox(NULL,buffer,"Error",MB_OK|MB_ICONERROR);

					for (size_t i=0; i<images.size(); i++) delete images[i];
					for (size_t i=0; i<himages.size(); i++) if (himages[i]) delete himages[i];
					images.clear(); himages.clear(); 
					delete model->_voxel_build;
					return;
				}
				himages.push_back(img);
			}
			else himages.push_back(NULL);
		}

		Vec3f	v((float)pixel_size_x,(float)(max_height-min_height+1),(float)pixel_size_y);
		long	vx=(long)dot_product(vr,v);
		long	vy=(long)dot_product(vu,v);
		long	vz=(long)dot_product(vd,v);
		float	svx=build->_size.x/(float)vx;
		float	svy=build->_size.y/(float)vy;
		float	svz=build->_size.z/(float)vz;
	
		model->_voxel_model=new VoxelWorld;
		model->_voxel_model->create(vx,vy,vz,vx,vy,vz,svx,svy,svz);
		if (build->_texture_id!=-1) 
		{
			model->_voxel_model->set_material_texture(build->_texture_id,build->_ix,build->_iy);
			for (int i=0; i<256; i++) model->_voxel_model->set_uv_scale(i,build->_material_scale[i]);
		}
		model->_voxel_model->set_offset(Vec3f(-build->_size.x*0.5f,0.0f,-build->_size.z*0.5f));
		model->_voxel_model->set_color_mode(true);

		for (size_t i=0; i<build->_slices.size(); i++)
		{
			Image_A8R8G8B8* img=images[i];
			Image_A8R8G8B8* himg=himages[i];
			unsigned char*	src=img->get_data();
			unsigned char*	hsrc=(himg)?(himg->get_data()):(NULL);
			Color			c;
			int				min_y,max_y,tmp1,tmp2;
			float			r,g;
			int				sx=(int)img->get_size_x();
			int				sz=(int)img->get_size_y();
			Vec3f			tmp;
			unsigned char	mat=1;
			int				ox=0,oy=0;
			for (int zz=0; zz<sz; zz++)
			{
				for (int xx=0; xx<sx; xx++)
				{
					if (src[3]!=0)
					{
						min_y=build->_slices[i].h1;
						max_y=build->_slices[i].h2;
						ox=build->_slices[i].ox;
						oy=build->_slices[i].oy;
						if (himg)
						{
							r=((float)hsrc[2])/255.0f;
							g=((float)hsrc[1])/255.0f;
							tmp1=(int)((1.0f-r)*min_y+r*max_y);
							tmp2=(int)((1.0f-g)*min_y+g*max_y);
							min_y=(tmp1<tmp2)?(tmp1):(tmp2);
							max_y=(tmp1<tmp2)?(tmp2):(tmp1);
						}
						min_y=min_y-min_height;
						max_y=max_y-min_height;
						for (int yy=min_y; yy<=max_y; yy++)
						{
							tmp.set((float)xx+ox,(float)yy,(float)zz+oy);
							tmp.set(dot_product(vr,tmp),dot_product(vu,tmp),dot_product(vd,tmp));
							if (inv_x) tmp.x=model->_voxel_model->_sx-tmp.x-1.0f;
							if (inv_y) tmp.y=model->_voxel_model->_sy-tmp.y-1.0f;
							if (inv_z) tmp.z=model->_voxel_model->_sz-tmp.z-1.0f;
							if ((build->_texture_id!=-1) && (hsrc)) mat=hsrc[0];
							if ((tmp.x>=0) && (tmp.x<model->_voxel_model->_sx) &&
								(tmp.y>=0) && (tmp.y<model->_voxel_model->_sy) &&
								(tmp.z>=0) && (tmp.z<model->_voxel_model->_sz))
							{
								model->_voxel_model->set_c((int)tmp.x,(int)tmp.y,(int)tmp.z,mat,*(unsigned long*)src);
							}
						}
					}

					src+=4;
					if (hsrc) hsrc+=4;
				}
			}
		}

		model->_voxel_model->rebuild();

		for (size_t i=0; i<images.size(); i++) delete images[i];
		for (size_t i=0; i<himages.size(); i++) if (himages[i]) delete himages[i];
		images.clear(); himages.clear();

		delete model->_voxel_build;
		model->_voxel_build=NULL;
	}
	else if (model->_geom_build)
	{
		model->_geom_model=NULL;
		auto it=model->_geom_build->_ops.begin();
		while (it!=model->_geom_build->_ops.end())
		{
			switch ((*it)._type)
			{
				case GEOMOP_CUBE:
					{
						Vec3f			center=Vec3f(0.0f,0.5f*(*it)._size.y,0.0f);
						Vec3f			dir_x((*it)._size.x,0.0f,0.0f);
						Vec3f			dir_y(0.0f,(*it)._size.y,0.0f);
						Vec3f			dir_z(0.0f,0.0f,(*it)._size.z);
						unsigned long	c=0xFFFFFFFF;
						Rect			uv[6] = { Rect(0.0f,0.0f,0.25f,0.25f), Rect(0.25f,0.0f,0.5f,0.25f), Rect(0.5f,0.0f,0.75f,0.25f), Rect(0.75f,0.0f,1.0f,0.25f), Rect(0.0f,0.25f,0.25f,0.5f), Rect(0.25f,0.25f,0.5f,0.5f) };
						model->_geom_model=geom_add_cube(model->_geom_model,center,dir_x,dir_y,dir_z,c,0xFF,&(uv[0]));
					}
			}
			++it;
		}

		model->_geom_model->close(true);
		
		Material* material=new Material;
		if (model->_geom_build->_texture_id!=-1)
		{
			material->set_texture_sampling(0,D3DTEXF_POINT);
			material->set_shader("basic");
			material->set_texture(0,model->_geom_build->_texture_id);
		}
		else
		{
			material->set_shader("basic");
		}
		model->_geom_model->set_material(material,true);

		delete model->_geom_build;
		model->_geom_build=NULL;
	}
}

void ModelManager::voxel_size(const long voxel_model_id,const float sx,const float sy,const float sz)
{
	Model* model=get_model(voxel_model_id);
	if (!model) return;
	if (!model->_voxel_build) return;
	
	model->_voxel_build->_size.set(sx,sy,sz);
}

void ModelManager::voxel_slice(const long voxel_model_id,const int offset_x,const int offset_y,const long h1,const long h2,const char* color_image,const char* heightmap)
{
	Model* model=get_model(voxel_model_id);
	if (!model) return;
	if (!model->_voxel_build) return;

	VoxelSlice vs;
	vs.h1=h1;
	vs.h2=h2;
	vs.ox=offset_x;
	vs.oy=offset_y;
	vs.color_image=color_image;
	vs.height_image=(heightmap)?(heightmap):("");

	model->_voxel_build->_slices.push_back(vs);
}

void ModelManager::voxel_extrusion_dir(const long voxel_model_id,const ExtrusionDir dir)
{
	Model* model=get_model(voxel_model_id);
	if (!model) return;
	if (!model->_voxel_build) return;

	model->_voxel_build->_dir=dir;
}

void ModelManager::voxel_materials(const long voxel_model_id,const long texture_id,const long ix,const long iy)
{
	Model* model=get_model(voxel_model_id);
	if (!model) return;
	if (!model->_voxel_build) return;

	model->_voxel_build->_texture_id=texture_id;
	model->_voxel_build->_ix=ix;
	model->_voxel_build->_iy=iy;
}

void ModelManager::voxel_materials_scale(const long voxel_model_id,const long material_id,const float sx,const float sy,const float sz)
{
	Model* model=get_model(voxel_model_id);
	if (!model) return;
	if (!model->_voxel_build) return;

	model->_voxel_build->_material_scale[material_id]=Vec3f(1.0f/sx,1.0f/sy,1.0f/sz);
}

ModelManager::Model* ModelManager::get_model(const long model_id)
{
	if (model_id==-1) return NULL;
	if (_models.size()<=(size_t)model_id) return NULL;

	return _models[model_id];
}

BaseMesh* ModelManager::get_mesh(const long model_id)
{
	if (model_id==-1) return NULL;
	if (_models.size()<=(size_t)model_id) return NULL;

	return _models[model_id]->_mesh_model;
}

void ModelManager::geom_set_texture(const long geom_model_id,const long texture_id)
{
	Model* model=get_model(geom_model_id);
	if (!model) return;
	if (!model->_geom_build) return;

	model->_geom_build->_texture_id=texture_id;
}

long ModelManager::geom_push_op(const long geom_model_id,const GeomOpType op_type)
{
	Model* model=get_model(geom_model_id);
	if (!model) return -1;
	if (!model->_geom_build) return -1;

	GeomOp op(op_type);

	model->_geom_build->_ops.push_back(op);

	return model->_geom_build->_ops.size()-1;
}

void ModelManager::geom_set_size(const long geom_model_id,const long op_id,const float sx,const float sy,const float sz)
{
	Model* model=get_model(geom_model_id);
	if (!model) return;
	if (!model->_geom_build) return;
	
	model->_geom_build->_ops[op_id]._size.set(sx,sy,sz);
}

bool ModelManager::get_aabb(const long model_id,Vec3f& pmin,Vec3f& pmax)
{
	Model* model=get_model(model_id);
	if (!model) return false;

	if (model->_set_aabb)
	{
		pmin=model->_pmin;
		pmax=model->_pmax;
		return true;
	}

	if (model->_voxel_model)
	{
		__asm int 3
	}
	else if (model->_geom_model)
	{
		__asm int 3
	}
	else if (model->_mesh_model)
	{
		model->_mesh_model->get_aabb(pmin,pmax);
		return true;
	}	
	return false;
}

void ModelManager::set_aabb(const long model_id,const Vec3f& pmin,const Vec3f& pmax)
{
	Model* model=get_model(model_id);
	if (!model) return;

	model->_pmin=pmin;
	model->_pmax=pmax;
	model->_set_aabb=true;
}
