#include "NavMesh.h"
#include "BlitManager3d.h"
#include "Tilemap3d.h"
#include "ModelObj3d.h"
#include "NavPath.h"

class Recast_QueryFilter : public dtQueryFilter
{
public:
	Recast_QueryFilter() { ; }
	virtual ~Recast_QueryFilter() { ; }

	virtual bool passFilter(const dtPolyRef ref,const dtMeshTile* tile,const dtPoly* poly) const
	{ 
		return true; 
	}
	virtual float getCost(const float* pa, const float* pb,const dtPolyRef prevRef, const dtMeshTile* prevTile, const dtPoly* prevPoly,
						  const dtPolyRef curRef, const dtMeshTile* curTile, const dtPoly* curPoly,		
						  const dtPolyRef nextRef, const dtMeshTile* nextTile, const dtPoly* nextPoly) const 
	{ 
		float cost=sqrtf(Sqr(pa[0]-pb[0])+Sqr(pa[1]-pb[1])+Sqr(pa[2]-pb[2]));

		return cost; 
	}
};

NavMesh::NavMesh()
{
	_max_vertex_per_poly=6;
	_agent_radius=1.0f;
	_agent_height=2.0f;
	_agent_step=0.25f;
	_cellsize=0.1f;
	_cellheight=0.1f;
	_pmin.set(FLT_MAX,FLT_MAX,FLT_MAX);
	_pmax.set(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	
	_pmesh=NULL;
	_dmesh=NULL;
	_nav_mesh=NULL;
	_default_query_filter=NULL;
	_query_object=NULL;
};

NavMesh::~NavMesh(void)
{
	if (_pmesh)
	{
		rcFreePolyMesh(_pmesh);
		_pmesh=NULL;
	}
	if (_dmesh)
	{
		rcFreePolyMeshDetail(_dmesh);
		_dmesh=NULL;
	}
	if (_nav_mesh)
	{
		dtFreeNavMesh(_nav_mesh);
		_nav_mesh=NULL;
	}
	if (_query_object)
	{
		dtFreeNavMeshQuery(_query_object);
		_query_object=NULL;
	}
	if (_default_query_filter)
	{
		delete _default_query_filter;
		_default_query_filter=NULL;
	}
}

void NavMesh::add_tilemap3d(Tilemap3d* tilemap)
{
	tilemap->load_quads_to_array(&_verts,&_tris);
}

void NavMesh::add_modelobj3d(ModelObj3d* model_obj)
{
	BaseMesh* mesh=model_obj->get_mesh();

	Vec3f	p[3];
	size_t	base_index;
	for (size_t i=0; i<mesh->get_triangle_count(); i++)
	{
		mesh->get_triangle(i,&p[0]);
		for (int j=0; j<3; j++)
		{
			p[j].set(p[j].x*model_obj->_scale.x,p[j].y*model_obj->_scale.y,p[j].z*model_obj->_scale.z);
			model_obj->_ori.rotate_point(p[j],p[j]);
			p[j]=p[j]+model_obj->_pos;
		}
		
		base_index=_verts.size();
		_verts.push_back(p[0]);
		_verts.push_back(p[1]);
		_verts.push_back(p[2]);

		_tris.push_back(base_index);
		_tris.push_back(base_index+1);
		_tris.push_back(base_index+2);
	}
}

void NavMesh::init()
{
	build_navmesh();
}

void NavMesh::draw()
{
	if (GFX->get_pass()!=PASS_ALPHA) return;

	if ((_verts.size()>0) && (_tris.size()>0))
	{
		for (size_t i=0; i<_tris.size(); i+=3)
		{
			Vec3f p0=_verts[_tris[i]];
			Vec3f p1=_verts[_tris[i+1]];
			Vec3f p2=_verts[_tris[i+2]];

			BLIT3D->add_triangle(p0,p1,p2,Color(1.0f,1.0f,0.0f,0.8f),BLIT_NORMAL);
		}
	}

	if (_pmesh)
	{
		unsigned short	vindex[6];
		Vec3f			v[6];

		for (int i=0; i<_pmesh->npolys; i++)
		{
			int		n_vert=_pmesh->nvp; 
			Vec3f	center(0.0f,0.0f,0.0f);
			for (int j=0; j<_pmesh->nvp; j++)
			{
				vindex[j]=_pmesh->polys[i*_pmesh->nvp*2+j];
				if (vindex[j]==RC_MESH_NULL_IDX)
				{
					n_vert=j;
					break;
				}
				v[j].set(_pmesh->verts[vindex[j]*3+0]*_pmesh->cs+_pmesh->bmin[0],
						(_pmesh->verts[vindex[j]*3+1]-1)*_pmesh->ch+_pmesh->bmin[1],
						_pmesh->verts[vindex[j]*3+2]*_pmesh->cs+_pmesh->bmin[2]);
				center=center+v[j];
			}
			center=center/(float)n_vert;

			Vec3f	normal=cross_product(v[1]-v[0],v[2]-v[0]); normal.normalize();

			Color c(0.5f,1.0f,0.5f,0.75f);

			for (int j=0; j<n_vert; j++)
			{
				//c.set((float)(rand()%255)/255.0f,(float)(rand()%255)/255.0f,(float)(rand()%255)/255.0f,0.75f);
				BLIT3D->add_triangle(center+normal*0.1f,v[j]+normal*0.1f,v[(j+1)%n_vert]+normal*0.1f,c);
			}
		}
	}
}

void NavMesh::build_navmesh()
{
	if ((_verts.size()==0) || (_tris.size()==0)) return;
	//
	// Step 1. Initialize build config.
	//
	Vec3f p;
	for (size_t i=0; i<_verts.size(); i++)
	{
		p=_verts[i];
		_pmin.x=Min(_pmin.x,p.x);
		_pmin.y=Min(_pmin.y,p.y);
		_pmin.z=Min(_pmin.z,p.z);
		_pmax.x=Max(_pmax.x,p.x);
		_pmax.y=Max(_pmax.y,p.y);
		_pmax.z=Max(_pmax.z,p.z);
	}

	// Init build configuration from GUI
	// See data from http://digestingduck.blogspot.com/2009/08/recast-settings-uncovered.html
	// and http://digestingduck.blogspot.com/2009/08/recast-and-detour-14-released.html
	//
	rcConfig	cfg;

	memset(&cfg, 0, sizeof(cfg));
	cfg.cs = _cellsize;
	cfg.ch = _cellheight;
	cfg.walkableSlopeAngle = 45.0f;
	cfg.walkableHeight = (int)ceilf(_agent_height/ cfg.ch);
	cfg.walkableClimb = (int)floorf(_agent_step/ cfg.ch);
	cfg.walkableRadius = (int)ceilf(_agent_radius/ cfg.cs);
	cfg.maxEdgeLen = (int)(1.0f / cfg.cs);
	cfg.maxSimplificationError = 0.0f;
	cfg.minRegionArea = 1;
	cfg.mergeRegionArea = 1;
	cfg.maxVertsPerPoly = (int)_max_vertex_per_poly;
	cfg.detailSampleDist = 2.0f;
	cfg.detailSampleMaxError = 1.0f;

	// Set the area where the navigation will be build.
	// Here the bounds of the input mesh are used, but the
	// area could be specified by an user defined box, etc.
	Vec3f pmin=_pmin-Vec3f(_cellsize,_cellheight,_cellsize);
	Vec3f pmax=_pmax+Vec3f(_cellsize,_cellheight,_cellsize);
	cfg.bmin[0]=pmin.x; cfg.bmin[1]=pmin.y; cfg.bmin[2]=pmin.z;
	cfg.bmax[0]=pmax.x; cfg.bmax[1]=pmax.y; cfg.bmax[2]=pmax.z;

	rcCalcGridSize(cfg.bmin, cfg.bmax, cfg.cs, &cfg.width, &cfg.height);

	//
	// Step 2. Rasterize input polygon soup.
	//

	// Allocate voxel heightfield where we rasterize our input data to.
	rcContext		ctx(false);
	rcHeightfield*	solid=rcAllocHeightfield();
	if (!solid)
	{
		__asm int 3
		return;
	}
	if (!rcCreateHeightfield(&ctx, *solid, cfg.width, cfg.height, cfg.bmin, cfg.bmax, cfg.cs, cfg.ch))
	{
		__asm int 3
		return;
	}

	// Allocate array that can hold triangle area types.
	// If you have multiple meshes you need to process, allocate
	// and array which can hold the max number of triangles you need to process.
	long			n_tri=_tris.size()/3;
	unsigned char*	triareas=new unsigned char[n_tri];
	if (!triareas)
	{
		__asm int 3
		return;
	}

	// Find triangles which are walkable based on their slope and rasterize them.
	// If your input data is multiple meshes, you can transform them here, calculate
	// the are type for each of the meshes and rasterize them.
	float*	verts_ptr=&(_verts[0].x);
	int*	tris_ptr=&(_tris[0]);

	memset(triareas,0,n_tri*sizeof(unsigned char));
	rcMarkWalkableTriangles(&ctx,cfg.walkableSlopeAngle,verts_ptr,_verts.size(),tris_ptr,n_tri,triareas);
	rcRasterizeTriangles(&ctx,verts_ptr,_verts.size(),tris_ptr,triareas,n_tri,*solid,cfg.walkableClimb);

	delete [] triareas;
	triareas=NULL;

	//
	// Step 3. Filter walkables surfaces.
	//

	// Once all geoemtry is rasterized, we do initial pass of filtering to
	// remove unwanted overhangs caused by the conservative rasterization
	// as well as filter spans where the character cannot possibly stand.
	rcFilterLowHangingWalkableObstacles(&ctx,cfg.walkableClimb,*solid);
	rcFilterLedgeSpans(&ctx,cfg.walkableHeight,cfg.walkableClimb,*solid);
	rcFilterWalkableLowHeightSpans(&ctx,cfg.walkableHeight,*solid);

	//
	// Step 4. Partition walkable surface to simple regions.
	//

	// Compact the heightfield so that it is faster to handle from now on.
	// This will result more cache coherent data as well as the neighbours
	// between walkable cells will be calculated.
	rcCompactHeightfield* chf=rcAllocCompactHeightfield();
	if (!chf)
	{
		__asm int 3
		return;
	}
	if (!rcBuildCompactHeightfield(&ctx,cfg.walkableHeight,cfg.walkableClimb,*solid,*chf))
	{
		__asm int 3
		return;
	}

	rcFreeHeightField(solid);
	solid = NULL;

	// Erode the walkable area by agent radius.
	if (!rcErodeWalkableArea(&ctx,cfg.walkableRadius,*chf))
	{
		__asm int 3
		return;
	}

	// Prepare for region partitioning, by calculating distance field along the walkable surface.
	if (!rcBuildDistanceField(&ctx, *chf))
	{
		__asm int 3
		return;
	}

	// Partition the walkable surface into simple regions without holes.
	if (!rcBuildRegions(&ctx, *chf, 0, cfg.minRegionArea, cfg.mergeRegionArea))
	{
		__asm int 3
		return;
	}//*/

	//
	// Step 5. Trace and simplify region contours.
	//

	// Create contours.
	rcContourSet* cset=rcAllocContourSet();
	if (!cset)
	{
		__asm int 3
		return;
	}
	if (!rcBuildContours(&ctx,*chf,cfg.maxSimplificationError,cfg.maxEdgeLen,*cset))
	{
		__asm int 3
		return;
	}

	//
	// Step 6. Build polygons mesh from contours.
	//
	// Build polygon navmesh from the contours.
	if (_pmesh)
	{
		rcFreePolyMesh(_pmesh);
		_pmesh=NULL;
	}
	_pmesh=rcAllocPolyMesh();
	if (!_pmesh)
	{
		__asm int 3
		return;
	}
	if (!rcBuildPolyMesh(&ctx,*cset,cfg.maxVertsPerPoly,*_pmesh))
	{
		__asm int 3
		return;
	}

	//
	// Step 7. Create detail mesh which allows to access approximate height on each polygon.
	//
	if (_dmesh)
	{
		rcFreePolyMeshDetail(_dmesh);
		_dmesh=NULL;
	}
	_dmesh=rcAllocPolyMeshDetail();
	if (!_dmesh)
	{
		__asm int 3
		return;
	}

	if (!rcBuildPolyMeshDetail(&ctx, *_pmesh, *chf, cfg.detailSampleDist, cfg.detailSampleMaxError, *_dmesh))
	{
		__asm int 3
		return;
	}

	rcFreeCompactHeightfield(chf);
	chf = NULL;
	rcFreeContourSet(cset);
	cset = NULL;

	_verts.clear();
	_tris.clear();

	build_detour_data();
}

void NavMesh::build_detour_data()
{
	if (_nav_mesh)
	{
		dtFreeNavMesh(_nav_mesh);
		_nav_mesh=NULL;
	}

	if ((!_pmesh) || (!_dmesh))
	{
		return;
	}

	dtNavMeshCreateParams params;
	memset(&params, 0, sizeof(params));
	params.verts = _pmesh->verts;
	params.vertCount = _pmesh->nverts;
	params.polys = _pmesh->polys;
	params.polyAreas = _pmesh->areas;
	params.polyFlags = _pmesh->flags;
	params.polyCount = _pmesh->npolys;
	params.nvp = _pmesh->nvp;
	params.detailMeshes = _dmesh->meshes;
	params.detailVerts = _dmesh->verts;
	params.detailVertsCount = _dmesh->nverts;
	params.detailTris = _dmesh->tris;
	params.detailTriCount = _dmesh->ntris;
	params.offMeshConVerts = NULL;
	params.offMeshConRad = NULL;
	params.offMeshConDir = NULL;
	params.offMeshConAreas = NULL;
	params.offMeshConFlags = NULL;
	params.offMeshConUserID = NULL;
	params.offMeshConCount = NULL;
	params.walkableHeight = _agent_height;
	params.walkableRadius = _agent_radius;
	params.walkableClimb = _agent_step;

	Vec3f pmin=_pmin-Vec3f(_cellsize,_cellheight,_cellsize);
	Vec3f pmax=_pmax+Vec3f(_cellsize,_cellheight,_cellsize);

	params.bmin[0]=pmin.x; params.bmin[1]=pmin.y; params.bmin[2]=pmin.z;
	params.bmax[0]=pmax.x; params.bmax[1]=pmax.y; params.bmax[2]=pmax.z;
	params.cs = _cellsize;
	params.ch = _cellheight;
	params.buildBvTree = true;

	unsigned char*	navData=0;
	int				navDataSize=0;

	if (!dtCreateNavMeshData(&params,&navData,&navDataSize))
	{
		return;
	}

	_nav_mesh=dtAllocNavMesh();
	if (!_nav_mesh)
	{
		dtFree(navData);
		return;
	}

	dtStatus status;

	status=_nav_mesh->init(navData,navDataSize,DT_TILE_FREE_DATA);
	if (dtStatusFailed(status))
	{
		dtFree(navData);
		dtFreeNavMesh(_nav_mesh);
		_nav_mesh=NULL;
		return;
	}

	_query_object=dtAllocNavMeshQuery();
	if (dtStatusFailed(_query_object->init(_nav_mesh,1000)))
	{
		_query_object=NULL;
		return;
	}
	_default_query_filter=new Recast_QueryFilter;
}

NavPath* NavMesh::get_path(const Vec3f& pos1,const Vec3f& pos2,NavPath* input_path,dtQueryFilter* input_query_filter)
{
	dtStatus		ret;

	NavPath* dest=input_path;
	if (dest) dest->clear();
	else dest=new NavPath;

	dtQueryFilter*	query_filter=input_query_filter;
	if (!query_filter) 
	{
		query_filter=_default_query_filter;
	}

	Vec3f			query_extents(20.0f,20.0f,20.0f);
	dtPolyRef		start_ref,end_ref;
	Vec3f			pt;

	dest->_source=pos1;
	dest->_target=pos2;

	ret=_query_object->findNearestPoly(&pos1.x,&query_extents.x,query_filter,&start_ref,&pt.x);
	if ((dtStatusFailed(ret)) || (start_ref==0))
	{
		if (!input_query_filter) delete dest;
		else dest->_target=pos1;
		return NULL;
	}

	query_extents.set(20.0f,20.0f,20.0f);
	
	Vec3f	tp;
	ret=_query_object->findNearestPoly(&pos2.x,&query_extents.x,query_filter,&end_ref,&tp.x);
	if ((dtStatusFailed(ret)) || (end_ref==0))
	{
		if (!input_query_filter) delete dest;
		else dest->_target=pos1;
		return NULL;
	}

	dtPolyRef		path[512];
	int				path_len;
	memset(&path[0],0,sizeof(dtPolyRef)*512);
	ret=_query_object->findPath(start_ref,end_ref,&pos1.x,&pos2.x,query_filter,&path[0],&path_len,512);
	if (dtStatusFailed(ret))
	{
		if (!input_query_filter) delete dest;
		else dest->_target=pos1;
		return NULL;
	}

	dtPolyRef			path_refs[2048];
	unsigned char		path_flags[2048];
	ret=_query_object->findStraightPath(&(pos1.x),&(pos2.x),&path[0],path_len,&(dest->_path_pos[0].x),&path_flags[0],&path_refs[0],&(dest->_path_pos_len),2048,0);
	if (dtStatusFailed(ret))
	{
		if (!input_query_filter) delete dest;
		else dest->_target=pos1;
		return NULL;
	}

	// Check if the last point is the target point
	Vec3f last_point=(dest->_path_pos[dest->_path_pos_len-1]);
	float dist=(last_point-tp).norm_squared();
	if (dist>1.0f)
	{
		dest->_path_pos[dest->_path_pos_len]=tp;
		dest->_path_pos_len++;
	}

	return dest;
}
