// ParticleSystem.cpp: implementation of the ParticleSystem class.
//
//////////////////////////////////////////////////////////////////////

#include "ParticleSystem.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "Material.h"
#include "Mesh.h"
#include "Gfx.h"
#include "Utils.h"
#include "Screen.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ParticleSystem::ParticleSystem(const char* name) : Drawable("ps2d",name)
{
	_rate_start=_rate_end=0.0f;
	_life_base=2000.0f;
	_life_variance=0.0f;
	_speed_base=0.0f;
	_speed_variance=0.0f;
	_vb=NULL;
	_allocated=0;
	set_particle_size(1.0f,1.0f);
	set_particle_color_start(1.0f,1.0f,1.0f,1.0f);
	set_particle_color_end(1.0f,1.0f,1.0f,1.0f);
	_priority=-100.0f;
	_life=999999.0f;
	_material=new Material;
}

ParticleSystem::~ParticleSystem()
{
	_particles.clear();
	if (_vb) { delete _vb; _vb=NULL; }
	delete _material;
}

void ParticleSystem::draw()
{
	if (!_draw) return;
	if (_particles.size()==0) return;

	BLIT->flush();

	if (!_vb)
	{
		// Create vertex buffer for the particle
		_vb=create_vertex_buffer((long)(_particles.size()+PSBUFFER));
		_allocated=(long)_particles.size()+PSBUFFER;
	}
	else
	{
		// Check if they have space
		if ((size_t)_allocated<_particles.size())
		{
			// Resize buffer
			delete _vb;
			_vb=create_vertex_buffer(_allocated*2);
			_allocated=_allocated*2;
		}
	}

	// Fill buffer
	VBPart* ptr=(VBPart*)_vb->get_write_pointer();
	if (!ptr) return;

	Particles::iterator par;
	for (par=_particles.begin(); par!=_particles.end(); ++par)
	{
		Particle* tmp=*par;
		ptr->x=(*par)->pos.x;
		ptr->y=(*par)->pos.y;
		ptr->z=(*par)->pos.z;
		float life=((*par)->current_life)*((*par)->total_life);
		ptr->color=get_color(life);
		ptr->size=get_size(life);
		ptr++;
	}
	_vb->release_write_pointer();

	// Render buffer
	_material->set_color_vertex(true,false);
	_material->set_cull_mode(D3DCULL_NONE);
	_material->set_point_sprite(true);
	_material->set_z_write(false);
	_material->set_texture(0,_current_image);
	if (_current_image==-1) _material->set_shader("basic");
	else _material->set_shader("basic");

	if (_blend_mode==BLEND_NONE) _material->set_alpha(false,D3DBLEND_ONE,D3DBLEND_ZERO);
	else if (_blend_mode==BLEND_ALPHA) _material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
	else if (_blend_mode==BLEND_ADDALPHA ) _material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_ONE);
	else if (_blend_mode==BLEND_ADD) _material->set_alpha(true,D3DBLEND_ONE,D3DBLEND_ONE);

	_material->set(D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_PSIZE);
	// Draw!
	GFX->render(D3DPT_POINTLIST,_vb,(long)_particles.size());

	_material->unset();
}

void ParticleSystem::anim(const float ttt)
{
	// Spawn particles
	float	t=timeGetTime()-_last_frame_time;
	
	_life=_life-t;

	long	n_part=long(get_rate(_life*_total_life)*t/1000.0f);
	float	px=_pos.x;
	float	py=_pos.y;
	float	pz=_pos.z;

	if (_life>0.0f)
	{
		for (long i=0; i<n_part; i++)
		{
			Particle* tmp=new Particle;
			
			tmp->pos.set(px,py,pz);
			tmp->speed.set((float)(rand()%100-50),(float)(rand()%100-50),0.0f);
			tmp->speed.normalize();
			tmp->speed*=_speed_base+get_random_number(-_speed_variance,_speed_variance);
			tmp->current_life=_life_base+get_random_number(-_life_variance,_life_variance);
			tmp->total_life=1.0f/tmp->current_life;
			_particles.push_back(tmp);
		}
	}
	else
	{
		if (_particles.size()==0)
		{
			delete this;
			SCREEN->remove_drawable(this);
			return;
		}
	}
	// Animate particles
	Particles::iterator par=_particles.begin();
	while (par!=_particles.end())
	{
		Particle *tmp=*par;
		tmp->pos=tmp->pos+tmp->speed*t;
		tmp->current_life-=t;
		if (tmp->current_life<0.0f) 
		{
			par=_particles.erase(par);
		}
		else
		{
			++par;
		}
	}
	_last_frame_time=(float)timeGetTime();
}

VertexBuffer* ParticleSystem::create_vertex_buffer(long size)
{
	VertexBuffer*	tmp=new VertexBuffer;
	
	if (tmp->create(size,sizeof(VBPart),D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_PSIZE,true)) return tmp;

	delete tmp;
	ERROR_MSG("Can't create vertex buffer!");

	return NULL;
}

float ParticleSystem::get_size(float life)
{
	return _size_end*(1.0f-life)+_size_start*life;
}

D3DCOLOR ParticleSystem::get_color(float life)
{
	Color final_color;
	fetch_color(1.0f-life,&final_color);

	Vec3f	tmp=GFX->get_color_modulation();

	return D3DCOLOR_COLORVALUE(tmp.x*final_color.x,tmp.y*final_color.y,tmp.z*final_color.z,final_color.w);
}

float ParticleSystem::get_rate(float life)
{
	return _rate_end*(1.0f-life)+_rate_end*life;
}

void ParticleSystem::fetch_color(float t,Color* c)
{
	size_t lower_bound=0xFFFF;

	ColorElems::reverse_iterator r=_color.rbegin();
	while (r!=_color.rend())
	{
		if ((*r).t<t) break;
		++r;
	}
	if (r==_color.rend()) return;
	// R has the start t (t0)
	float t0=(*r).t;
	
	ColorElems::iterator e=r.base();
	if (e==_color.end()) { *c=(*r).c; return; }

	float t1=(*e).t;

	float tt=(t-t0)/(t1-t0);

	*c=(*r).c*(1-tt)+(*e).c*tt;
}
