// ParticleSystem3d.cpp: implementation of the ParticleSystem3d class.
//
//////////////////////////////////////////////////////////////////////

#include "ParticleSystem3d.h"
#include "TextureManager.h"
#include "BlitManager3d.h"
#include "Material.h"
#include "Mesh.h"
#include "Camera.h"
#include "Gfx.h"
#include "Utils.h"
#include "Screen.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ParticleSystem3d::ParticleSystem3d(const char* name) : Object3d("ps3d",name)
{
	_rate_start=_rate_end=0.0f;
	_life_base=2000.0f;
	_life_variance=0.0f;
	_speed_base=0.0f;
	_speed_variance=0.0f;
	_vb=NULL;
	_ib=NULL;
	_allocated=0;
	set_particle_size(1.0f,1.0f);
	set_particle_color_start(1.0f,1.0f,1.0f,1.0f);
	set_particle_color_end(1.0f,1.0f,1.0f,1.0f);
	_life=999999.0f;
	_last_frame_time=(float)timeGetTime();
	_current_image=0;
	_blend_mode=BLEND_ALPHA;
	_emitter=EMITTER3D_OMNI;
	_last_frac_part_count=0.0f;
	_local_reference=false;
	_normals=false;
	_material=new Material;
	_material->set_color_vertex(true,false);
	_material->set_cull_mode(D3DCULL_NONE);
	_material->set_z_write(false);
	_material->disable_all_passes();
	_material->enable_pass(PASS_ALPHA);
}

ParticleSystem3d::~ParticleSystem3d()
{
	_particles.clear();
	if (_vb) { delete _vb; _vb=NULL; }
	if (_ib) { delete _ib; _ib=NULL; }
	delete _material;
	_material=NULL;
}

void ParticleSystem3d::draw()
{
	if (_particles.size()==0) return;
	if (!_material->pass_enabled(GFX->get_pass())) return;

	if (_link)
	{
		_pos=_link->_pos;
		_ori=_link->_ori;
	}

	BLIT3D->flush();

	if (!_vb)
	{
		_vb=create_vertex_buffer((long)(_particles.size()+PSBUFFER)*4);
		_ib=create_index_buffer((long)(_particles.size()+PSBUFFER)*6);
		_allocated=(long)_particles.size()+PSBUFFER;
	}
	else
	{
		// Check if they have space
		if ((size_t)_allocated<_particles.size())
		{
			// Resize buffer
			delete _vb;
			delete _ib;
			_vb=create_vertex_buffer(_allocated*2*4);
			_ib=create_index_buffer(_allocated*2*6);
			_allocated=_allocated*2;
		}
	}

	// Fill buffer
	Vec3f		right=CAMERA3D->_right_axis;
	Vec3f		up=CAMERA3D->_up_axis;
	Quaternion	inv_ori; inv_ori.invert(_ori);
	inv_ori.rotate_point(right,right);
	inv_ori.rotate_point(up,up);

	if (_normals)
	{
		Vec3f	normal=CAMERA3D->_normal_axis;

		VBPart_N*	ptr=(VBPart_N*)_vb->get_write_pointer();
		if (!ptr) return;

		WORD*		idx=_ib->get_write_pointer();
		if (!idx) return;

		Particles::iterator par;
		Particle*			tmp;
		float				s;
		float				life;
		DWORD				color;
		int					index=0;
		for (par=_particles.begin(); par!=_particles.end(); ++par)
		{
			tmp=*par;
			life=((*par)->current_life)*((*par)->total_life);
			color=get_color(life);
			s=get_size(life);

			ptr->p=(-right+up)*s+tmp->pos;
			ptr->n=normal;
			ptr->color=color;
			ptr->uv.set(0.0f,0.0f);
			ptr++;

			ptr->p=( right+up)*s+tmp->pos;
			ptr->n=normal;
			ptr->color=color;
			ptr->uv.set(1.0f,0.0f);
			ptr++;

			ptr->p=( right-up)*s+tmp->pos;
			ptr->n=normal;
			ptr->color=color;
			ptr->uv.set(1.0f,1.0f);
			ptr++;

			ptr->p=(-right-up)*s+tmp->pos;
			ptr->n=normal;
			ptr->color=color;
			ptr->uv.set(0.0f,1.0f);
			ptr++;

			*idx=index; idx++;
			*idx=index+1; idx++;
			*idx=index+2; idx++;
			*idx=index; idx++;
			*idx=index+2; idx++;
			*idx=index+3; idx++;

			index+=4;
		}
		_vb->release_write_pointer();
		_ib->release_write_pointer();
	}
	else
	{
		VBPart*	ptr=(VBPart*)_vb->get_write_pointer();
		if (!ptr) return;

		WORD*		idx=_ib->get_write_pointer();
		if (!idx) return;

		Particles::iterator par;
		Particle*			tmp;
		float				s;
		float				life;
		DWORD				color;
		int					index=0;
		for (par=_particles.begin(); par!=_particles.end(); ++par)
		{
			tmp=*par;
			life=((*par)->current_life)*((*par)->total_life);
			color=get_color(life);
			s=get_size(life)*0.5f;

			ptr->p=(-right+up)*s+tmp->pos;
			ptr->color=color;
			ptr->uv.set(0.0f,0.0f);
			ptr++;

			ptr->p=( right+up)*s+tmp->pos;
			ptr->color=color;
			ptr->uv.set(1.0f,0.0f);
			ptr++;

			ptr->p=( right-up)*s+tmp->pos;
			ptr->color=color;
			ptr->uv.set(1.0f,1.0f);
			ptr++;

			ptr->p=(-right-up)*s+tmp->pos;
			ptr->color=color;
			ptr->uv.set(0.0f,1.0f);
			ptr++;

			*idx=index; idx++;
			*idx=index+1; idx++;
			*idx=index+2; idx++;
			*idx=index; idx++;
			*idx=index+2; idx++;
			*idx=index+3; idx++;

			index+=4;
		}
		_vb->release_write_pointer();
		_ib->release_write_pointer();
	}

	// Render buffer
	unsigned long fvf;
	if (_normals) fvf=D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX1;
	else fvf=D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1; 

	if (_local_reference) GFX->set_world_transformation(0,_pos.x,_pos.y,_pos.z,_ori);
	else GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));

	_material->set(fvf);

	GFX->render(D3DPT_TRIANGLELIST,_vb,(long)_particles.size()*4,_ib,(long)_particles.size()*6);

	if (_local_reference) GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));

	_material->unset();
}

void ParticleSystem3d::anim(const float t)
{
	// Spawn particles
	if (_life!=FLT_MAX) _life=_life-t;

	long	n_part;

	float tmp_part;
	if (_life!=FLT_MAX)
	{
		tmp_part=get_rate(_life*_total_life)*t/1000.0f;
	}
	else
	{
		tmp_part=get_rate(1.0f)*t/1000.0f;
	}
	n_part=(long)(tmp_part);
	if (_last_frac_part_count>1)
	{
		n_part+=(long)floorf(_last_frac_part_count);
		_last_frac_part_count=_last_frac_part_count-(long)floorf(_last_frac_part_count);
	}
	else _last_frac_part_count+=(tmp_part-n_part);

	if (_life>0.0f)
	{
		new_particles(n_part);
	}
	// Animate particles
	Particles::iterator par=_particles.begin();
	while (par!=_particles.end())
	{
		Particle *tmp=*par;
		tmp->pos=tmp->pos+tmp->speed*t;
		if (_emitter==EMITTER3D_CONE_GRAV)
		{
			tmp->speed.y-=t*0.00001f;
		}
		tmp->current_life-=t;
		if (tmp->current_life<0.0f) 
		{
			par=_particles.erase(par);
		}
		else
		{
			++par;
		}
	}
	_last_frame_time=(float)timeGetTime();
}

VertexBuffer* ParticleSystem3d::create_vertex_buffer(long size)
{
	VertexBuffer*	tmp=new VertexBuffer;
	bool			ret;

	if (_normals) ret=tmp->create(size,sizeof(VBPart_N),D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX1,true);
	else ret=tmp->create(size,sizeof(VBPart),D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1,true);

	if (ret) return tmp;

	delete tmp;

	ERROR_MSG("Can't create vertex buffer!");

	return NULL;
}

IndexBuffer* ParticleSystem3d::create_index_buffer(long size)
{
	IndexBuffer*	tmp=new IndexBuffer;

	if (tmp->create(size,true)) return tmp;

	delete tmp;

	ERROR_MSG("Can't create vertex buffer!");

	return NULL;
}

float ParticleSystem3d::get_size(float life)
{
	return _size_end*(1.0f-life)+_size_start*life;
}

D3DCOLOR ParticleSystem3d::get_color(float life)
{
	Color final_color;
	fetch_color(1.0f-life,&final_color);

	Vec3f	tmp=GFX->get_color_modulation();

	return D3DCOLOR_COLORVALUE(tmp.x*final_color.x,tmp.y*final_color.y,tmp.z*final_color.z,final_color.w);
}

float ParticleSystem3d::get_rate(float life)
{
	return _rate_end*(1.0f-life)+_rate_start*life;
}

void ParticleSystem3d::fetch_color(float t,Color* c)
{
	size_t lower_bound=0xFFFF;

	ColorElems::reverse_iterator r=_color.rbegin();
	while (r!=_color.rend())
	{
		if ((*r).t<t) break;
		++r;
	}
	if (r==_color.rend()) return;
	// R has the start t (t0)
	float t0=(*r).t;

	ColorElems::iterator e=r.base();
	if (e==_color.end()) { *c=(*r).c; return; }

	float t1=(*e).t;

	float tt=(t-t0)/(t1-t0);

	*c=(*r).c*(1-tt)+(*e).c*tt;
}

void ParticleSystem3d::set_blend_mode(const BlendMode mode)
{ 
	_blend_mode=mode; 	
	if (_blend_mode==BLEND_NONE) _material->set_alpha(false,D3DBLEND_ONE,D3DBLEND_ZERO);
	else if (_blend_mode==BLEND_ALPHA) _material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
	else if (_blend_mode==BLEND_ADDALPHA ) _material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_ONE);
	else if (_blend_mode==BLEND_ADD) _material->set_alpha(true,D3DBLEND_ONE,D3DBLEND_ONE);
}

void ParticleSystem3d::set_image(const long img) 
{ 
	_current_image=img; _material->set_texture(0,_current_image);  
	_material->set_shader("basic"); 
}

void ParticleSystem3d::set_normals(const bool b) 
{ 
	_normals=b; 
	if (b) _material->set_emissive_color(0.0f,0.0f,0.0f,0.0f); 
	else _material->set_emissive_color(1.0f,1.0f,1.0f,1.0f); 
}

void ParticleSystem3d::new_particles(const long n_part)
{
	float	px,py,pz;

	if (_local_reference)
	{
		px=py=pz=0.0f;
	}
	else
	{
		px=_pos.x;
		py=_pos.y;
		pz=_pos.z;
	}

	if (_emitter==EMITTER3D_OMNI)
	{
		for (long i=0; i<n_part; i++)
		{
			Particle* tmp=new Particle;

			tmp->pos.set(px,py,pz);
			tmp->speed.set((float)(rand()%100-50),(float)(rand()%100-50),0.0f);
			tmp->speed.normalize();
			tmp->speed*=_speed_base+get_random_number(-_speed_variance,_speed_variance);
			tmp->current_life=_life_base+get_random_number(-_life_variance,_life_variance);
			tmp->total_life=1.0f/tmp->current_life;
			_particles.push_back(tmp);
		}
	}
	else if ((_emitter==EMITTER3D_CONE_GRAV) || (_emitter==EMITTER3D_CONE))
	{
		for (long i=0; i<n_part; i++)
		{
			Particle* tmp=new Particle;

			tmp->pos.set(px,py,pz);
			tmp->speed.set((float)(rand()%100-50),100.0f,(float)(rand()%100-50));
			tmp->speed.normalize();
			_ori.rotate_point(tmp->speed,tmp->speed);
			tmp->speed*=_speed_base+get_random_number(-_speed_variance,_speed_variance);
			tmp->current_life=_life_base+get_random_number(-_life_variance,_life_variance);
			tmp->total_life=1.0f/tmp->current_life;
			_particles.push_back(tmp);
		}
	}
	else if (_emitter==EMITTER3D_CIRCUMFERENCE_XZ)
	{
		for (long i=0; i<n_part; i++)
		{
			Particle* tmp=new Particle;

			Vec3f delta=Vec3f(rand()%512-256.0f,0,rand()%512-256.0f); 
			delta.normalize(); delta=delta*_emitter_param[0];

			tmp->pos.set(px+delta.x,py+delta.y,pz+delta.z);
			tmp->speed.set(0.0f,1.0f,0.0f);
			tmp->speed.normalize();
			tmp->speed*=_speed_base+get_random_number(-_speed_variance,_speed_variance);
			tmp->current_life=_life_base+get_random_number(-_life_variance,_life_variance);
			tmp->total_life=1.0f/tmp->current_life;
			_particles.push_back(tmp);
		}
	}
	else if (_emitter==EMITTER3D_CIRCLE_XZ)
	{
		for (long i=0; i<n_part; i++)
		{
			Particle* tmp=new Particle;

			Vec3f delta=Vec3f(rand()%512-256.0f,0,rand()%512-256.0f); 
			delta.normalize(); delta=delta*get_random_number(0.0f,_emitter_param[0]);

			tmp->pos.set(px+delta.x,py+delta.y,pz+delta.z);
			tmp->speed.set(0.0f,1.0f,0.0f);
			tmp->speed.normalize();
			tmp->speed*=_speed_base+get_random_number(-_speed_variance,_speed_variance);
			tmp->current_life=_life_base+get_random_number(-_life_variance,_life_variance);
			tmp->total_life=1.0f/tmp->current_life;
			_particles.push_back(tmp);
		}
	}
}