// Screen.cpp: implementation of the Screen class.
//
//////////////////////////////////////////////////////////////////////

#include "Screen.h"
#include "Gfx.h"
#include "Sound.h"
#include "Font.h"
#include "Mouse.h"
#include "Keyboard.h"
#include "Timers.h"
#include "Defs.h"
#include "TextureManager.h"
#include "ModelManager.h"
#include "Tokenizer.h"
#include "Utils.h"
#include "BlitManager.h"
#include <algorithm>

void ScreenMouseCallback(unsigned long msg,unsigned long wparam,unsigned long lparam)
{
	if (SCREEN) SCREEN->mouse_callback(msg,wparam,lparam);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Screen*	SCREEN=NULL;

Screen::Screen()
{
	_exit=false;
	_font_size=26;
	_font=NULL;
	_prev_screen=NULL;
	_anim_enable=true;
	set_mouse_event_callback(ScreenMouseCallback);

	HWND desktop=GetDesktopWindow();
	RECT desktop_rect;
	GetWindowRect(desktop,&desktop_rect);

	_desktop_res.x=(float)(desktop_rect.right-desktop_rect.left);
	_desktop_res.y=(float)(desktop_rect.bottom-desktop_rect.right);
}

Screen::~Screen()
{
	if (SCREEN==this) SCREEN=NULL;
	if (_font) 
	{
		delete _font; 
		_font=NULL; 
	}
	remove_drawables();
	for (unsigned long i=0; i<_drawables.size(); i++)
	{
		if (_drawables[i])
		{
			delete _drawables[i];
			_drawables[i]=NULL;
		}
	}
	_drawables.clear();
}

long Screen::run()
{
	// Wait for keyboard to be idle
	while (1)
	{
		KEY->poll();
		if (!KEY->any_pressed()) break;
	}

	MSG				msg;
	unsigned long	t;

	unsigned long	lastTime;
	float			ms;
	
	_prev_screen=SCREEN;
	SCREEN=this;

	_exit=false;

	lastTime=timeGetTime();

	while (!_exit)
	{
		ms=1000.0f/50.0f;

		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		// Stop app going more than a given frames per second
		do
		{
			t=timeGetTime();
		} while (t-lastTime<ms);
		float elapsed=float(t)-float(lastTime);
		if (elapsed>(1000.0f/30.0f)) elapsed=1000.0f/30.0f;
		run(elapsed);
		lastTime=t;
	}

	SCREEN=_prev_screen;
	_prev_screen=NULL;

	return _exit_code;
}

class DrawableSortOrder
{
public:
	bool operator()(Drawable* o1,Drawable* o2)
	{
		return o1->_pos.z>o2->_pos.z;
	}
};

void Screen::run(float t)
{
	if (KEY) KEY->poll();
	if (MOUSE) MOUSE->poll();
	if (TIMERS) TIMERS->poll();

	if (_anim_enable) anim(t);

	remove_drawables();

	// Sort drawables
	std::sort(_drawables.begin(),_drawables.end(),DrawableSortOrder());

	if (!_exit)
	{
		render();
		GFX->swap_buffers();
	}
}

void Screen::remove_all_drawables(const bool should_delete)
{
	if (should_delete)
	{
		auto it=_drawables.begin();
		while (it!=_drawables.end())
		{
			delete (*it);
			++it;
		}
	}
	_drawables.clear();
}

void Screen::remove_drawables()
{
	Removes::iterator i=_removes.begin(); 
	while (i!=_removes.end())
	{
		Drawables::iterator j=_drawables.begin();
		while (j!=_drawables.end())
		{
			if ((*j)==(*i).drawable) 
			{
				j=_drawables.erase(j);
				break;
			}
			else ++j;
		}
		if ((*i).should_delete)
		{
			delete (*i).drawable;
		}
		i++;
	}
	_removes.clear();
}

void Screen::load_assets(std::string filename)
{
	FILE* file=NULL;
	fopen_s(&file,filename.c_str(),"rt");
	if (!file) 
	{
		ERROR_MSG("Can't find assets file!");
		return;
	}

	Tokenizer		tokens; tokens.add_delimiter(','); tokens.add_delimiter('=');
	char			buffer[1024];
	unsigned long	n;
	long			last_texture_id=-1;
	long			last_model_id=-1;

	while (!feof(file))
	{
		memset(buffer,0,1024);
		fgets((char*)&buffer,1024,file);
		if (buffer[0]=='#') continue;

		tokens.clear_tokens();
		n=tokens.tokenize((char*)&buffer);
		if (n>0)
		{
			char* type=tokens.get_token(0);
			if (!_stricmp(type,"texture"))
			{
				char* internal_name=tokens.get_token(1);
				char* filename=tokens.get_token(2);

				last_texture_id=TEXTURES->load(internal_name,filename);
			}
			else if (!_stricmp(type,"texture_wrap"))
			{
				if (last_texture_id!=-1)
				{
					if (!_stricmp(tokens[1],"true")) TEXTURES->set_wrap(last_texture_id,true);
					else TEXTURES->set_wrap(last_texture_id,false);
				}
			}
			else if (!_stricmp(type,"sound"))
			{
				char* internal_name=tokens.get_token(1);
				char* filename=tokens.get_token(2);
				char* loops=tokens.get_token(3);
				char* streams=tokens.get_token(4);
				bool loop;
				if (!_stricmp(loops,"true")) loop=true;
				else loop=false;
				bool stream;
				if (!_stricmp(streams,"true")) stream=true;
				else stream=false;

				if (SOUND) SOUND->load(internal_name,filename,stream,loop);
			}
			else if (!_stricmp(type,"frame"))
			{
				if ((last_texture_id!=-1) && (tokens.get_token_number()>=6))
				{
					char*	name=tokens.get_token(1);
					float	sx=(float)TEXTURES->get_size_x(last_texture_id);
					float	sy=(float)TEXTURES->get_size_y(last_texture_id);
					float	x1=(float)atof(tokens.get_token(2))/sx;
					float	y1=(float)atof(tokens.get_token(3))/sy;
					float	x2=(float)atof(tokens.get_token(4))/sx;
					float	y2=(float)atof(tokens.get_token(5))/sy;

					TEXTURES->add_frame(last_texture_id,name,x1,y1,x2,y2);
				}
			}
			else if (!_stricmp(type,"anim"))
			{
				if (last_texture_id!=-1)
				{
					char*	name=tokens[1];
					long	anim_id=TEXTURES->add_animation(last_texture_id,name);
					float time=200.0f;
					bool  loop=false;
					for (size_t i=2; i<tokens.get_token_number(); i++)
					{
						char* tmp=tokens.get_token(i);
						if (!_strnicmp(tmp,"time",4))
						{
							time=(float)atof(tokens[i+1]); i++;
							TEXTURES->set_animation_time(last_texture_id,anim_id,time);
						}
						else if (!_strnicmp(tmp,"loop",4))
						{
							if (!_stricmp(tokens[i+1],"true")) loop=true;
							i++;
							TEXTURES->set_animation_loop(last_texture_id,anim_id,loop);
						}
						else if (is_a_number(tmp))
						{
							TEXTURES->add_animation_frame(last_texture_id,anim_id,atol(tmp));
						}
						else 
						{
							TEXTURES->add_animation_frame(last_texture_id,anim_id,tmp);
						}
					}
				}
			}
			else if (!_stricmp(type,"model"))
			{
				char* name=tokens[1];
				if (!_stricmp(tokens[2],"voxel"))
				{
					if (last_model_id!=-1) MODEL_MANAGER->model_build(last_model_id);
					last_model_id=MODEL_MANAGER->new_voxel_model(name);
				}
				else if (!_stricmp(tokens[2],"build_geom"))
				{
					if (last_model_id!=-1) MODEL_MANAGER->model_build(last_model_id);
					last_model_id=MODEL_MANAGER->new_build_geom_model(name);
				}
				else
				{
					std::string tmp=to_lower(tokens[2]);
					size_t		idx=tmp.find(".");
					if (idx!=std::string::npos)
					{
						tmp=tmp.substr(idx+1,std::string::npos);
					}
					if (tmp=="3ds")
					{
						last_model_id=MODEL_MANAGER->load_3ds(name,tokens[2]);
					}
					else
					{
						char buffer[512];
						sprintf_s((char*)&buffer,512,"Unknown model type %s!",tokens[2]);
						MessageBox(NULL,buffer,"Error",MB_OK|MB_ICONERROR);
					}
				}
			}
			else if (!_stricmp(type,"model_aabb"))
			{
				Vec3f pmin((float)atof(tokens[1]),(float)atof(tokens[2]),(float)atof(tokens[3]));
				Vec3f pmax((float)atof(tokens[4]),(float)atof(tokens[5]),(float)atof(tokens[6]));
				MODEL_MANAGER->set_aabb(last_model_id,pmin,pmax);
			}
			else if (!_stricmp(type,"voxel_size"))
			{
				if (last_model_id!=-1)
				{
					MODEL_MANAGER->voxel_size(last_model_id,(float)atof(tokens[1]),(float)atof(tokens[2]),(float)atof(tokens[3]));
				}
			}
			else if (!_stricmp(type,"voxel_slice"))
			{
				if (last_model_id!=-1)
				{
					int ox=0,oy=0;
					int idx=3;
					if (!_stricmp(tokens[3],"offset"))
					{
						ox=atol(tokens[4]);
						oy=atol(tokens[5]);
						idx=6;
					}
					MODEL_MANAGER->voxel_slice(last_model_id,ox,oy,atol(tokens[1]),atol(tokens[2]),tokens[idx+0],tokens[idx+1]);
				}
			}
			else if (!_stricmp(type,"voxel_materials"))
			{
				if (last_model_id!=-1)
				{
					MODEL_MANAGER->voxel_materials(last_model_id,TEXTURES->find_by_name(tokens[1]),atol(tokens[2]),atol(tokens[3]));
				}
			}
			else if (!_stricmp(type,"voxel_extrusion_dir"))
			{
				if (last_model_id!=-1)
				{
					if (!_stricmp(tokens[1],"+x")) MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_PX);
					else if (!_stricmp(tokens[1],"+y")) MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_PY);
					else if (!_stricmp(tokens[1],"+z")) MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_PZ);
					else if (!_stricmp(tokens[1],"-x")) MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_NX);
					else if (!_stricmp(tokens[1],"-y")) MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_NY);
					else if (!_stricmp(tokens[1],"-z")) MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_NZ);
					else if (tokens[1][0]=='x') MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_PX);
					else if (tokens[1][0]=='y') MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_PY);
					else if (tokens[1][0]=='z') MODEL_MANAGER->voxel_extrusion_dir(last_model_id,ModelManager::EXTRUSION_DIR_PZ);
				}
			}
			else if (!_stricmp(type,"voxel_materials_scale"))
			{
				if (last_model_id!=-1)
				{
					MODEL_MANAGER->voxel_materials_scale(last_model_id,atol(tokens[1]),(float)atof(tokens[2]),(float)atof(tokens[3]),(float)atof(tokens[4]));
				}
			}
			else if (!_stricmp(type,"geom_cube"))
			{
				if (last_model_id!=-1)
				{
					long op_id=MODEL_MANAGER->geom_push_op(last_model_id,ModelManager::GEOMOP_CUBE);
					size_t index=1;
					while (index<tokens.get_token_number())
					{
						if (!_stricmp(tokens[index],"size")) { MODEL_MANAGER->geom_set_size(last_model_id,op_id,(float)atof(tokens[index+1]),(float)atof(tokens[index+2]),(float)atof(tokens[index+3])); index+=4; }
						else index++;
					}
				}
			}
			else if (!_stricmp(type,"geom_texture"))
			{
				if (last_model_id!=-1)
				{
					MODEL_MANAGER->geom_set_texture(last_model_id,TEXTURES->find_by_name(tokens[1]));
				}
			}
			else if (type[0]==0)
			{

			}
			else
			{
				char buffer[512];
				sprintf_s((char*)&buffer,512,"Unknown keyword %s!",type);
				MessageBox(NULL,buffer,"Error",MB_OK|MB_ICONERROR);
			}
		}
	}

	if (last_model_id!=-1) MODEL_MANAGER->model_build(last_model_id);

	fclose(file);
}

bool Screen::create_font()
{
	_font=new Font;
	if (!_font->init(_font_size,"MV Boli"))
	{
		delete _font;
		_font=NULL;
		ERROR_MSG("Can't create font!");
		return false;
	}

	return true;
}

void Screen::delete_font()
{
	if (_font)
	{
		delete _font;
		_font=NULL;
	}
}

void Screen::mouse_callback(unsigned long msg,unsigned long wparam,unsigned long lparam)
{
	if (msg==WM_ACTIVATE)
	{
		if ((LOWORD(wparam)==WA_ACTIVE) || (LOWORD(wparam)==WA_CLICKACTIVE))
		{
			_anim_enable=true;
			RECT rect;
			GetWindowRect(GFX->get_window(),&rect);
#ifndef DEV
			ClipCursor(&rect);
#endif
			if (SOUND) SOUND->resume_all();
		}
		else
		{
			_anim_enable=false;
			ClipCursor(NULL);

			if (SOUND) SOUND->pause_all();
		}
	}
}
