// TextureManager.cpp: implementation of the TextureManager class.
//
//////////////////////////////////////////////////////////////////////

#include "TextureManager.h"
#include "Utils.h"
#include "Image_A8R8G8B8.h"
#include "Gfx.h"

TextureManager* TEXTURES=NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TextureManager::TextureManager()
{
	if (!TEXTURES) TEXTURES=this;
}

TextureManager::~TextureManager()
{
	clear_all();
	if (TEXTURES==this) TEXTURES=NULL;
}

long TextureManager::load(const char *name,const char *filename)
{
	TextureID	texture;

	FILE *file=NULL;
	fopen_s(&file,filename,"rb");
	if (!file)
	{
		char buffer[512];
		char working_dir[512];
		GetCurrentDirectory(512,(char*)&working_dir);
		sprintf_s((char*)&buffer,512,"File not found (%s/%s)!",(char*)&working_dir,filename);
		ERROR_MSG((char*)&buffer);
		return -1;
	}
	fclose(file);

	int		width,height;

	texture=GFX->load_texture(filename,width,height);
	if (texture==TEXTURE_INVALID)
	{
		ERROR_MSG("Can't read texture!");
		return -1;
	}

	Texture* tex=new Texture;
	tex->filename=_strdup(filename);
	if (name) tex->name=_strdup(name); else tex->name=NULL;
	tex->texture=texture;

	tex->tx=tex->sx=width;
	tex->ty=tex->sy=height;
	tex->wrap=false;

	_textures.push_back(tex);

	return (long)_textures.size()-1;
}

long TextureManager::load(const char *name,Image_A8R8G8B8* img)
{
	// Create texture
	TextureID	texture=GFX->new_texture(img->get_size_x(),(UINT)img->get_size_y(),D3DFMT_A8R8G8B8);
	if (texture==TEXTURE_INVALID)
	{
		return -1;
	}

	// Fill texture
	int				pitch;
	unsigned char*	src=img->get_data();
	unsigned char*	dest=(unsigned char*)lock_texture(texture,pitch);

	for (size_t y=0; y<img->get_size_y(); y++)
	{
		memcpy(dest,src,img->get_size_x()*4);
		src+=img->get_size_x()*4;
		dest+=pitch;
	}

	unlock_texture(texture);

	// Add texture
	Texture* tex=new Texture;
	tex->filename=NULL;
	if (name) tex->name=_strdup(name); else tex->name=NULL;
	tex->texture=texture;

	tex->tx=(int)img->get_size_x();
	tex->ty=(int)img->get_size_y();
	tex->sx=(int)img->get_size_x();
	tex->sy=(int)img->get_size_y();
	tex->wrap=false;

	_textures.push_back(tex);

	return (long)_textures.size()-1;
}

long TextureManager::add(const char *name,LPDIRECT3DTEXTURE9 texture)
{
	Texture* tex=new Texture;
	tex->filename=NULL;
	if (name) tex->name=_strdup(name); else tex->name=NULL;
	tex->texture=texture;

	D3DSURFACE_DESC	desc;
	texture->GetLevelDesc(0,&desc);

	tex->sx=desc.Width;
	tex->sy=desc.Height;
	tex->wrap=false;

	for (size_t i=0; i<_textures.size(); i++)
	{
		if (_textures[i]==NULL)
		{
			_textures[i]=tex;
			return i;
		}
	}

	_textures.push_back(tex);

	return (long)_textures.size()-1;
}

long TextureManager::find_by_name(const char *name)
{
	for (unsigned long i=0; i<_textures.size(); i++)
	{
		if (_textures[i]) 
		{
			if (_textures[i]->name)
			{
				if (!_stricmp(_textures[i]->name,name)) return i;
			}
		}
	}

	return -1;
}

void TextureManager::clear(const char* name)
{
	for (unsigned long i=0; i<_textures.size(); i++)
	{
		if (_textures[i]) 
		{
			if (!_stricmp(_textures[i]->name,name))
			{
				free(_textures[i]->filename);
				_textures[i]->texture->Release();
				for (size_t j=0; j<_textures[i]->frames.size(); j++)
				{
					free(_textures[i]->frames[j].name);
				}
				for (size_t j=0; j<_textures[i]->anims.size(); j++)
				{
					free(_textures[i]->anims[j].name);
				}
				delete _textures[i];
				_textures[i]=NULL;
				return;
			}
		}
	}
}

void TextureManager::clear(long id)
{
	if (id==-1) return;
	if ((long)_textures.size()<=id) return;

	if (_textures[id]) 
	{
		free(_textures[id]->filename);
		_textures[id]->texture->Release();
		for (size_t j=0; j<_textures[id]->frames.size(); j++)
		{
			free(_textures[id]->frames[j].name);
		}
		for (size_t j=0; j<_textures[id]->anims.size(); j++)
		{
			free(_textures[id]->anims[j].name);
		}
		delete _textures[id];
		_textures[id]=NULL;
	}
}

void TextureManager::remove(long id)
{
	if (id==-1) return;
	if ((long)_textures.size()<=id) return;

	if (_textures[id]) 
	{
		free(_textures[id]->filename);
		for (size_t j=0; j<_textures[id]->frames.size(); j++)
		{
			free(_textures[id]->frames[j].name);
		}
		for (size_t j=0; j<_textures[id]->anims.size(); j++)
		{
			free(_textures[id]->anims[j].name);
		}
		delete _textures[id];
		_textures[id]=NULL;
	}
}

void TextureManager::clear_all()
{
	for (unsigned long i=0; i<_textures.size(); i++)
	{
		if (_textures[i]) 
		{
			free(_textures[i]->filename);
			_textures[i]->texture->Release();
			for (size_t j=0; j<_textures[i]->frames.size(); j++)
			{
				free(_textures[i]->frames[j].name);
			}
			for (size_t j=0; j<_textures[i]->anims.size(); j++)
			{
				free(_textures[i]->anims[j].name);
			}
			delete _textures[i];
		}
	}
	_textures.clear();
}

TextureID TextureManager::get_texture(long id)
{
	if (id==-1) return NULL;
	if (_textures[id])
	{
		return _textures[id]->texture;
	}

	return NULL;
}

int	TextureManager::get_size_x(long id)
{
	if (id==-1) return 0;
	if (_textures[id]) return _textures[id]->sx;

	return 0;
}

int	TextureManager::get_size_y(long id)
{
	if (id==-1) return 0;
	if (_textures[id]) return _textures[id]->sy;

	return 0;
}

int	TextureManager::get_texture_size_x(long id)
{
	if (id==-1) return 0;
	if (_textures[id]) return _textures[id]->tx;

	return 0;
}

int	TextureManager::get_texture_size_y(long id)
{
	if (id==-1) return 0;
	if (_textures[id]) return _textures[id]->ty;

	return 0;
}

std::string	TextureManager::get_texture_name(long id)
{
	if (id==-1) return "";
	if (_textures[id]) return _textures[id]->name;

	return "";
}

void TextureManager::set_wrap(const long id,const bool b)
{
	if (id==-1) return;
	_textures[id]->wrap=b;
}

bool TextureManager::get_wrap(const long id)
{
	if (id==-1) return false;
	return _textures[id]->wrap;
}

TextureManager::Texture* TextureManager::fetch_texture(const long texture_id)
{
	if (texture_id==-1) return NULL;
	if (texture_id>=(long)_textures.size()) return NULL;

	return _textures[texture_id];
}

void TextureManager::add_frame(const long texture_id,const char* name,const float u1,const float v1,const float u2,const float v2)
{
	Texture* tex=fetch_texture(texture_id);
	if (!tex) return;

	Frame f;
	f.name=_strdup(name);
	f.uv.set(u1,v1,u2,v2);

	tex->frames.push_back(f);
}

Rect TextureManager::get_uv(const long texture_id,const char* name)
{
	Texture* tex=fetch_texture(texture_id);
	if (!tex) return Rect(0.0f,0.0f,1.0f,1.0f);

	for (size_t i=0; i<tex->frames.size(); i++)
	{
		if (!_stricmp(name,tex->frames[i].name)) return tex->frames[i].uv;
	}

	return Rect(0.0f,0.0f,1.0f,1.0f);
}

long TextureManager::get_frame_id(const long texture_id,const char* name)
{
	Texture* tex=fetch_texture(texture_id);
	if (!tex) return -1;

	for (size_t i=0; i<tex->frames.size(); i++)
	{
		if (!_stricmp(name,tex->frames[i].name)) return i;
	}

	return -1;
}

Rect TextureManager::get_uv(const long texture_id,const long frame_id)
{
	if (frame_id==-1) return Rect(0.0f,0.0f,1.0f,1.0f);

	Texture* tex=fetch_texture(texture_id);
	if (!tex) return Rect(0.0f,0.0f,1.0f,1.0f);

	if (frame_id>=(long)tex->frames.size()) return Rect(0.0f,0.0f,1.0f,1.0f);

	return tex->frames[frame_id].uv;
}

long TextureManager::add_animation(const long texture_id,const char* name)
{
	Texture* tex=fetch_texture(texture_id);
	if (!tex) return -1;

	for (size_t i=0; i<tex->anims.size(); i++)
	{
		if (!_stricmp(name,tex->anims[i].name))
		{
			return i;
		}
	}

	Anim a;
	a.loop=false;
	a.time=200.0f;
	a.name=_strdup(name);

	tex->anims.push_back(a);

	return (long)tex->anims.size()-1;
}

long TextureManager::get_animation_id(const long texture_id,const char* name)
{
	Texture* tex=fetch_texture(texture_id);
	if (!tex) return -1;

	for (size_t i=0; i<tex->anims.size(); i++)
	{
		if (!_stricmp(name,tex->anims[i].name))
		{
			return i;
		}
	}

	return -1;
}

void TextureManager::set_animation_time(const long texture_id,const long anim_id,const float time)
{
	Anim* anm=fetch_anim(texture_id,anim_id);
	if (!anm) return;

	anm->time=time;
}

void TextureManager::set_animation_loop(const long texture_id,const long anim_id,const bool loop)
{
	Anim* anm=fetch_anim(texture_id,anim_id);
	if (!anm) return;

	anm->loop=loop;
}

void TextureManager::add_animation_frame(const long texture_id,const long anim_id,const long frame_id)
{
	Anim* anm=fetch_anim(texture_id,anim_id);
	if (!anm) return;

	anm->frames.push_back(frame_id);
}

void TextureManager::add_animation_frame(const long texture_id,const long anim_id,const char* frame_name)
{
	Anim* anm=fetch_anim(texture_id,anim_id);
	if (!anm) return;
	Texture*	tex=fetch_texture(texture_id);

	long frame_id=-1;
	for (size_t i=0; i<tex->frames.size(); i++)
	{
		if (!_stricmp(tex->frames[i].name,frame_name)) { frame_id=i; break; }
	}

	if (frame_id!=-1) anm->frames.push_back(frame_id);
}

TextureManager::Anim* TextureManager::fetch_anim(const long texture_id,const long anim_id)
{
	if (anim_id==-1) return NULL;

	Texture* tex=fetch_texture(texture_id);
	if (!tex) return NULL;

	if (anim_id>=(long)tex->anims.size()) return NULL;

	return &(tex->anims[anim_id]);
}

void* TextureManager::lock_texture(TextureID texture,int& pitch)
{
#ifdef USE_D3D9
	D3DLOCKED_RECT info;
	texture->LockRect(0,&info,NULL,0);
	
	pitch=info.Pitch;
	return info.pBits;
#else
	__asm int 3
	return NULL;
#endif
}

void TextureManager::unlock_texture(TextureID texture)
{
#ifdef USE_D3D9
	texture->UnlockRect(0);
#else
	__asm int 3
#endif
}
