#include "VoxelWorld.h"
#include "TextureManager.h"
#include "Camera.h"

#define VOXEL(X,Y,Z) _voxels[((Z)*_sx*_sy)+((Y)*_sx)+(X)]
#define VOXEL_COLOR(X,Y,Z) _voxels_color[((Z)*_sx*_sy)+((Y)*_sx)+(X)]
#define VOXEL_PTR(X,Y,Z) (_voxels+((Z)*_sx*_sy)+((Y)*_sx)+(X))
#define VOXEL_COLOR_PTR(X,Y,Z) (_voxels_color+((Z)*_sx*_sy)+((Y)*_sx)+(X))

VoxelWorld::VoxelWorld(void)
{
	_voxels=NULL;
	_clusters=NULL;
	_sx=_sy=_sz=0;
	_cluster_size_x=_cluster_size_y=_cluster_size_z=0;
	_cx=_cy=_cz=0;
	_voxel_size_x=_voxel_size_y=_voxel_size_z=0.0f;
	_color_mode=false;
	_voxels_color=NULL;
	_offset.set(0.0f,0.0f,0.0f); 

	float u,v;
	for (int i=0; i<256; i++) 
	{
		_uv_scale[i].set(1.0f,1.0f,1.0f);

		u=(((float)(i%8))*64.0f+0.5f)/512.0f;
		v=(((float)(i/8))*64.0f+0.5f)/512.0f;
		_uv_rect[i].set(u,v,u+63.5f/512.0f,v+63.5f/512.0f);

		_emissive_color[i]=0xFF000000;
		_color[i].set(1.0f,1.0f,1.0f,1.0f);
	}
	_pos.set(0.0f,0.0f,0.0f);
	_ori.set(0.0f,0.0f,0.0f,1.0f);
	_material_color.set(1.0f,1.0f,1.0f,1.0f);
	_face_cache=NULL;

	_material=new Material;
	_material->set_cull_mode(D3DCULL_NONE);
	_material->set_material_color(1.0f,1.0f,1.0f,1.0f);
	_material->set_emissive_color(1.0f,1.0f,1.0f,1.0f);
	_material->set_shader("basic");
	_material->set_texture_sampling(0,D3DTEXF_POINT);
	_material->set_alpha(false,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
	_material->set_color_vertex(true,true);
}

VoxelWorld::~VoxelWorld(void)
{
	clear();
	delete _material; _material=NULL;
}

void VoxelWorld::create(const int sx,const int sy,const int sz,const int cluster_size_x,const int cluster_size_y,const int cluster_size_z,const float voxel_size_x,const float voxel_size_y,const float voxel_size_z)
{
	clear();

	_voxels=new Voxel[sx*sy*sz];
	_sx=sx; _sy=sy; _sz=sz;
	memset(_voxels,0,sizeof(Voxel)*_sx*_sy*_sz);

	set_color_mode(_color_mode);

	_cluster_size_x=cluster_size_x;
	_cluster_size_y=cluster_size_y;
	_cluster_size_z=cluster_size_z;

	_cx=_sx/_cluster_size_x;
	_cy=_sy/_cluster_size_y;
	_cz=_sz/_cluster_size_z;

	_clusters=new VoxelCluster[_cx*_cy*_cz];
	memset(_clusters,0,sizeof(VoxelCluster)*_cx*_cy*_cz);
	int idx=0;
	for (int zz=0; zz<_cz; zz++)
	{
		for (int yy=0; yy<_cy; yy++)
		{
			for (int xx=0; xx<_cx; xx++)
			{
				_clusters[idx]._x=xx;
				_clusters[idx]._y=yy;
				_clusters[idx]._z=zz;

				++idx;
			}
		}
	}

	_voxel_size_x=voxel_size_x;
	_voxel_size_y=voxel_size_y;
	_voxel_size_z=voxel_size_z;
}

void VoxelWorld::set_color_mode(const bool color_mode)
{
	if (color_mode)
	{
		if (!_voxels_color)
		{
			_voxels_color=new unsigned long[_sx*_sy*_sz];

			unsigned long* tmp=_voxels_color;
			for (int i=0; i<_sx*_sy*_sz; i++)
			{
				*tmp=0xFF000000;
				tmp++;
			}
		}
		if (_material->get_texture(0)==-1) _material->set_shader("basic");
		else
		{
			_material->set_shader("basic");
			_material->set_texture_sampling(0,D3DTEXF_POINT);
		}
	}
	else
	{
		if (_voxels_color) delete[] _voxels_color;
		_voxels_color=NULL;
		_material->set_shader("basic");
		_material->set_texture_sampling(0,D3DTEXF_POINT);
	}

	_color_mode=color_mode;
}

void VoxelWorld::rebuild()
{
	int idx=0;

//	_stat_max_time=0;
//	unsigned long t0=timeGetTime();

	for (int zz=0; zz<_cz; zz++)
	{
		for (int yy=0; yy<_cy; yy++)
		{
			for (int xx=0; xx<_cx; xx++)
			{
				VoxelCluster* cluster=&(_clusters[idx]);

				rebuild_cluster(cluster);

				++idx;
			}
		}
	}

	if (_face_cache)
	{
		delete[] _face_cache;
		_face_cache=NULL;
	}

//	unsigned long t1=timeGetTime();

//	log("Rebuild time=%i ms (avg=%.2f ms, max=%i)",t1-t0,((float)t1-t0)/(float)(_cx*_cy*_cz),_stat_max_time);
}

void VoxelWorld::rebuild_cluster(VoxelCluster* cluster)
{
//	unsigned long t0=timeGetTime();

	if (!_face_cache)
	{
		_face_cache=new FaceCache[65536];
	}
	int				cube_count=0;
	int				face_count=0;
	int				x=cluster->_x;
	int				y=cluster->_y;
	int				z=cluster->_z;
	int				xx,yy,zz;
	int				base_x=x*_cluster_size_x;
	int				base_y=y*_cluster_size_y;
	int				base_z=z*_cluster_size_z;
	int				vx,vy,vz;
	int				voxel_type;
	unsigned char*	v_ptr;
	int				z_inc=(_sx*_sy);
	Vec3f			center;
	Vec3f			dirx(1.0f,0.0f,0.0f);
	Vec3f			diry(0.0f,1.0f,0.0f);
	Vec3f			dirz(0.0f,0.0f,1.0f);
	Vec3f			cube_vector_x(_voxel_size_x,0.0f,0.0f);
	Vec3f			cube_vector_y(0.0f,_voxel_size_y,0.0f);
	Vec3f			cube_vector_z(0.0f,0.0f,_voxel_size_z);
	Rect			rect;
	unsigned char	faces;
	Rect			face_uv[6];
	unsigned char	v;
	Color			c=_material_color;
	FaceCache*		tmp=_face_cache;

	for (zz=0; zz<_cluster_size_z; zz++)
	{
		vz=base_z+zz;
		for (yy=0; yy<_cluster_size_y; yy++)
		{
			vy=base_y+yy;
			for (xx=0; xx<_cluster_size_x; xx++)
			{
				vx=base_x+xx;

				v_ptr=VOXEL_PTR(vx,vy,vz);
				voxel_type=*v_ptr;

				if (voxel_type==0x00) continue;
				if (voxel_type==0xFF) continue;

				faces=0;

				if (vx==0) { faces|=FACE_MX; face_count++; }
				else 
				{
					v=*(v_ptr-1);
					if ((v==0) || (v==0xFF)) { faces|=FACE_MX; face_count++; }
				}

				if (vx==(_sx-1)) { faces|=FACE_PX; face_count++; }
				else 
				{
					v=*(v_ptr+1);
					if ((v==0) || (v==0xFF)) { faces|=FACE_PX; face_count++; }
				}

				if (vy==0) { faces|=FACE_MY; face_count++; }
				else 
				{
					v=*(v_ptr-_sx);
					if ((v==0) || (v==0xFF)) { faces|=FACE_MY; face_count++; }
				}

				if (vy==(_sy-1)) { faces|=FACE_PY; face_count++; }
				else 
				{
					v=*(v_ptr+_sx);
					if ((v==0) || (v==0xFF)) { faces|=FACE_PY; face_count++; }
				}

				if (vz==0) { faces|=FACE_MZ; face_count++; }
				else 
				{
					v=*(v_ptr-z_inc);
					if ((v==0) || (v==0xFF)) { faces|=FACE_MZ; face_count++; }
				}

				if (vz==(_sz-1)) { faces|=FACE_PZ; face_count++; }
				else 
				{
					v=*(v_ptr+z_inc);
					if ((v==0) || (v==0xFF)) { faces|=FACE_PZ; face_count++; }
				}

				if (faces==0) continue;

				tmp->_x=vx;
				tmp->_y=vy;
				tmp->_z=vz;
				tmp->_type=voxel_type;
				tmp->_faces=faces;
				tmp->_diffuse_color=(_voxels_color)?(*VOXEL_COLOR_PTR(vx,vy,vz)):(float_to_colorref(c.r*_color[voxel_type].r,c.g*_color[voxel_type].g,c.b*_color[voxel_type].b,c.a*_color[voxel_type].a));
				tmp++;
				cube_count++;
			}
		}
	}

	if (cube_count>0)
	{
		if (!cluster->_mesh)
		{
			cluster->_mesh=new GeomMeshE(true);
			cluster->_mesh->set_material(_material,false);
		}

		cluster->_mesh->alloc_vert(4*face_count);
		cluster->_mesh->alloc_index(6*face_count);

		WORD*	index_ptr=cluster->_mesh->get_write_pointer_i();
		auto	vertex_ptr=cluster->_mesh->get_write_pointer_v();

		FaceCache*	fc=_face_cache;
		int			vcount=0;
		int			icount=0;
		for (int i=0; i<cube_count; i++)
		{
			rect=_uv_rect[fc->_type];
			center.set(fc->_x*_voxel_size_x,fc->_y*_voxel_size_y,fc->_z*_voxel_size_z);

			if (fc->_faces&FACE_MZ) face_uv[0]=get_uv(fc->_type,0,rect,center,dirx,diry,dirz);
			if (fc->_faces&FACE_PZ) face_uv[1]=get_uv(fc->_type,1,rect,center,dirx,diry,dirz);
			if (fc->_faces&FACE_MX) face_uv[2]=get_uv(fc->_type,2,rect,center,dirx,diry,dirz);
			if (fc->_faces&FACE_PX) face_uv[3]=get_uv(fc->_type,3,rect,center,dirx,diry,dirz);
			if (fc->_faces&FACE_MY) face_uv[4]=get_uv(fc->_type,4,rect,center,dirx,diry,dirz);
			if (fc->_faces&FACE_PY) face_uv[5]=get_uv(fc->_type,5,rect,center,dirx,diry,dirz);

			geom_add_cube(vertex_ptr,index_ptr,vcount,icount,
						  center+_offset,cube_vector_x,cube_vector_y,cube_vector_z,
						  fc->_diffuse_color,_emissive_color[fc->_type],fc->_faces,&face_uv[0]);
		
			fc++;
		}
	}

//	unsigned long t2=timeGetTime();

//	if ((t2-t0)>_stat_max_time) _stat_max_time=t2-t0;
}

Rect VoxelWorld::get_uv(const unsigned char material,const int face,const Rect& uv_base,const Vec3f& center,const Vec3f& dirx,const Vec3f& diry,const Vec3f& dirz)
{
	const	Vec3f*	vu;
	const	Vec3f*	vv;
			float	su,sv;

	switch (face)
	{
		case 0: 
		case 1: vu=&dirx; vv=&diry; su=_voxel_size_x; sv=_voxel_size_y; break;
		case 2: 
		case 3: vu=&dirz; vv=&diry; su=_voxel_size_z; sv=_voxel_size_y; break;
		case 4: 
		case 5: vu=&dirx; vv=&dirz; su=_voxel_size_x; sv=_voxel_size_z; break;
	}

	Vec3f p1=center+*vv*sv;
	Vec3f p2=center+*vu*su;

	// Scaling world coord (tiling)
	p1.set(p1.x*_uv_scale[material].x,p1.y*_uv_scale[material].y,p1.z*_uv_scale[material].z);
	p2.set(p2.x*_uv_scale[material].x,p2.y*_uv_scale[material].y,p2.z*_uv_scale[material].z);

	Rect ret;

	ret.x1=dot_product(p1,*vu);
	ret.x2=dot_product(p2,*vu);
	ret.y1=dot_product(p1,*vv);
	ret.y2=dot_product(p2,*vv);

	ret.x1=ret.x1-(int)ret.x1;
	ret.y1=ret.y1-(int)(ret.y1-EPSILON_FLOAT);
	ret.x2=ret.x2-(int)(ret.x2-EPSILON_FLOAT);
	ret.y2=ret.y2-(int)ret.y2;

	ret.x1=uv_base.x2*(ret.x1)+uv_base.x1*(1.0f-ret.x1);
	ret.y1=uv_base.y1*(ret.y1)+uv_base.y2*(1.0f-ret.y1);
	ret.x2=uv_base.x2*(ret.x2)+uv_base.x1*(1.0f-ret.x2);
	ret.y2=uv_base.y1*(ret.y2)+uv_base.y2*(1.0f-ret.y2);

	return ret;
}

void VoxelWorld::render(Camera* camera)
{
	if (GFX->get_pass()==PASS_ALPHA) return;

	// Check if it is in the view frustum
	Vec3f			pmin,pmax;
	get_aabb(NULL,pmin,pmax);
	pmin=pmin+_pos;
	pmax=pmax+_pos;

	if (camera)
	{
		if (!camera->inside(pmin,pmax)) return;
	}

	GFX->set_world_transformation(0,_pos.x,_pos.y,_pos.z,_ori);
	
	UINT			max_verts=0;
	unsigned long	total_verts=0;
	UINT			drawn_clusters=0;

	for (int i=0; i<_cx*_cy*_cz; i++)
	{
		VoxelCluster cluster=_clusters[i];

		if (cluster._mesh)
		{
			UINT tmp=cluster._mesh->get_render_verts();
			if (tmp>0)
			{
				if (camera)
				{
					get_aabb(&cluster,pmin,pmax);
					pmin=pmin+_pos;
					pmax=pmax+_pos;

					if (!camera->inside(pmin,pmax)) continue;
				}
				
				UINT tmp=cluster._mesh->get_render_verts();
				if (tmp>0)
				{
					cluster._mesh->draw();
					drawn_clusters++;

					if (tmp>max_verts) max_verts=tmp;
					total_verts+=tmp;
				}

			}
		}
	}
}

void VoxelWorld::clear()
{
	if (_voxels) delete[] _voxels;
	if (_voxels_color) delete[] _voxels_color;
	if (_clusters) 
	{
		for (int i=0; i<(_cx*_cy*_cz); i++)
		{
			if (_clusters[i]._mesh)
			{
				delete _clusters[i]._mesh;
				_clusters[i]._mesh=NULL;
			}
		}
		delete[] _clusters;
	}
	_voxels=NULL;
	_voxels_color=NULL;
	_clusters=NULL;
	_sx=_sy=_sz=0;
	_cluster_size_x=_cluster_size_y=_cluster_size_z=0;
	_cx=_cy=_cz=0;
	_voxel_size_x=_voxel_size_y=_voxel_size_z=0.0f;
	if (_face_cache)
	{
		delete[] _face_cache;
		_face_cache=NULL;
	}
}

void VoxelWorld::set(const int x,const int y,const int z,const Voxel voxel)
{
#ifdef _DEBUG
	if ((x<0) || (x>=_sx) || (y<0) || (y>=_sy) || (z<0) || (z>=_sz)) __asm int 3
#endif
	VOXEL(x,y,z)=voxel;
}

void VoxelWorld::set_c(const int x,const int y,const int z,const Voxel voxel,const unsigned long color)
{
#ifdef _DEBUG
	if ((x<0) || (x>=_sx) || (y<0) || (y>=_sy) || (z<0) || (z>=_sz)) __asm int 3
#endif
	VOXEL(x,y,z)=voxel;
	if (_voxels_color) VOXEL_COLOR(x,y,z)=color;
}

void VoxelWorld::set(const Vec3f& pmin,const Vec3f& pmax,const Voxel voxel)
{
	int p1x=(int)((pmin.x-_pos.x)/_voxel_size_x);
	int p1y=(int)((pmin.y-_pos.y)/_voxel_size_y);
	int p1z=(int)((pmin.z-_pos.z)/_voxel_size_z);
	int p2x=(int)((pmax.x-_pos.x)/_voxel_size_x);
	int p2y=(int)((pmax.y-_pos.y)/_voxel_size_y);
	int p2z=(int)((pmax.z-_pos.z)/_voxel_size_z);

	for (int z=Max(p1z,0); z<Min(p2z,_sz); z++)
	{
		for (int y=Max(p1y,0); y<Min(p2y,_sy); y++)
		{
			for (int x=Max(p1x,0); x<Min(p2x,_sx); x++)
			{
				VOXEL(x,y,z)=voxel;
			}
		}
	}
}

unsigned char VoxelWorld::get(const int x,const int y,const int z)
{
	return VOXEL(x,y,z);
}

void VoxelWorld::set_material_texture(const long texture_id,const long ix,const long iy) 
{ 
	_material->set_texture(0,texture_id); 
	
	float sx=(float)TEXTURES->get_texture_size_x(texture_id);
	float sy=(float)TEXTURES->get_texture_size_y(texture_id);
	float ex=sx/(float)ix;
	float ey=sy/(float)iy;

	float u,v;
	for (int i=0; i<ix*iy; i++) 
	{
		u=(((float)(i%ix))*ex+0.5f)/sx;
		v=(((float)(i/ix))*ey+0.5f)/sy;
		_uv_rect[i].set(u,v,u+(ex-0.5f)/sx,v+(ey-0.5f)/sy);
	}
}

void VoxelWorld::set_material_alpha_enable(const bool b)
{
	_material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
}

bool VoxelWorld::raycast(const Vec3f& origin,const Vec3f& dir,int& ret_x,int& ret_y,int& ret_z,int& ret_face)
{
	// Naive raycasting: Check each cluster and see if there's an intersection there
	int		idx=0;
	float	t_min=FLT_MAX;
	float	current_t;
	bool	ret=false;

	for (int zz=0; zz<_cz; zz++)
	{
		for (int yy=0; yy<_cy; yy++)
		{
			for (int xx=0; xx<_cx; xx++)
			{
				VoxelCluster* cluster=&(_clusters[idx]);

				if (raycast(cluster,origin,dir,t_min,ret_x,ret_y,ret_z,ret_face,current_t))
				{
					t_min=current_t;
					ret=true;
				}

				++idx;
			}
		}
	}
	return ret;
}

bool VoxelWorld::raycast(const Vec3f& origin,const Vec3f& dir,float& t,const float t_max)
{
	// Optimized raycasting (raymarching)
	Vec3f	vdir[3];
	float	delta[3];
	Vec3f	s(fabs(dir.x)/_voxel_size_x,fabs(dir.y)/_voxel_size_y,fabs(dir.z)/_voxel_size_z);
	if (s.x<s.y)
	{
		if (s.y<s.z)
		{
			vdir[0].set(0.0f,0.0f,1.0f); delta[0]=dir.z/s.z;
			vdir[1].set(0.0f,1.0f,0.0f); delta[1]=dir.y/s.z;
			vdir[2].set(1.0f,0.0f,0.0f); delta[2]=dir.x/s.z;
		}
		else
		{
			vdir[0].set(0.0f,1.0f,0.0f); delta[0]=dir.y/s.y;
			vdir[1].set(0.0f,0.0f,1.0f); delta[1]=dir.z/s.y;
			vdir[2].set(1.0f,0.0f,0.0f); delta[2]=dir.x/s.y;
		}
	}
	else
	{
		if (s.x<s.z)
		{
			vdir[0].set(0.0f,0.0f,1.0f); delta[0]=dir.z/s.z;
			vdir[1].set(0.0f,1.0f,0.0f); delta[1]=dir.y/s.z;
			vdir[2].set(1.0f,0.0f,0.0f); delta[2]=dir.x/s.z;
		}
		else
		{
			vdir[0].set(1.0f,0.0f,0.0f); delta[0]=dir.x/s.x;
			vdir[1].set(0.0f,0.0f,1.0f); delta[1]=dir.z/s.x;
			vdir[2].set(0.0f,1.0f,0.0f); delta[2]=dir.y/s.x;
		}
	}
	
	bool	inside;
	Vec3f	current_pos=origin;
	int		vx=(int)floorf((0.5f+current_pos.x/_voxel_size_x));
	int		vy=(int)floorf((0.5f+current_pos.y/_voxel_size_y));
	int		vz=(int)floorf((0.5f+current_pos.z/_voxel_size_z));

	if ((vx<0) || (vy<0) || (vz<0) || (vx>=_sx) || (vy>=_sy) || (vz>=_sz)) inside=false;
	else inside=true;

	float	dist=0.0f;

	while (1)
	{
		vx=(int)floorf(((0.5f+current_pos.x)/_voxel_size_x));
		vy=(int)floorf(((0.5f+current_pos.y)/_voxel_size_y));
		vz=(int)floorf(((0.5f+current_pos.z)/_voxel_size_z));

		if ((vx<0) || (vy<0) || (vz<0) || (vx>=_sx) || (vy>=_sy) || (vz>=_sz)) 
		{
			if (inside) return false;
		}
		else
		{
			inside=true;
			Voxel v=get(vx,vy,vz);
			if ((v!=0) && (v!=0xFF))
			{
				t=dist;
				return true;
			}
		}

		current_pos=current_pos+vdir[0]*delta[0]+vdir[1]*delta[1]+vdir[2]*delta[2];
		dist+=sqrt(Sqr(delta[0])+Sqr(delta[1])+Sqr(delta[2]));
		
		if (dist>t_max) 
		{
			// Check last position
			vx=(int)floorf(((0.5f+origin.x+dir.x*t_max)/_voxel_size_x));
			vy=(int)floorf(((0.5f+origin.y+dir.y*t_max)/_voxel_size_y));
			vz=(int)floorf(((0.5f+origin.z+dir.z*t_max)/_voxel_size_z));

			if ((vx<0) || (vy<0) || (vz<0) || (vx>=_sx) || (vy>=_sy) || (vz>=_sz))  return false;

			Voxel v=get(vx,vy,vz);
			if ((v!=0) && (v!=0xFF))
			{
				t=dist;
				return true;
			}

			return false;
		}
	}

	return false;
}

void VoxelWorld::get_aabb(VoxelCluster* cluster,Vec3f& pmin,Vec3f& pmax)
{
	if (cluster)
	{
		int		x=cluster->_x;
		int		y=cluster->_y;
		int		z=cluster->_z;
		int		base_x=x*_cluster_size_x;
		int		base_y=y*_cluster_size_y;
		int		base_z=z*_cluster_size_z;

		pmin.set(base_x*_voxel_size_x,base_y*_voxel_size_y,base_z*_voxel_size_z); pmin=pmin-Vec3f(_voxel_size_x*0.5f,_voxel_size_y*0.5f,_voxel_size_z*0.5f);
		pmax=pmin+Vec3f(_voxel_size_x*_cluster_size_x,_voxel_size_y*_cluster_size_y,_voxel_size_z*_cluster_size_z); pmin=pmin-Vec3f(_voxel_size_x*0.5f,_voxel_size_y*0.5f,_voxel_size_z*0.5f);
	}
	else
	{
		pmin.set(0.0f,0.0f,0.0f);
		pmax.set(_voxel_size_x*_sx,_voxel_size_y*_sy,_voxel_size_z*_sz);
	}
}

bool VoxelWorld::raycast(VoxelCluster* cluster,const Vec3f& origin,const Vec3f& dir,const float t_min,int& ret_x,int& ret_y,int& ret_z,int& ret_face,float& ret_t)
{
	int		x=cluster->_x;
	int		y=cluster->_y;
	int		z=cluster->_z;
	int		xx,yy,zz;
	int		base_x=x*_cluster_size_x;
	int		base_y=y*_cluster_size_y;
	int		base_z=z*_cluster_size_z;
	int		vx,vy,vz;
	Voxel	voxel_type;

	Vec3f	pmin(base_x*_voxel_size_x,base_y*_voxel_size_y,base_z*_voxel_size_z); pmin=pmin-Vec3f(_voxel_size_x*0.5f,_voxel_size_y*0.5f,_voxel_size_z*0.5f);
	Vec3f	pmax=pmin+Vec3f(_voxel_size_x*_cluster_size_x,_voxel_size_y*_cluster_size_y,_voxel_size_z*_cluster_size_z); pmin=pmin-Vec3f(_voxel_size_x*0.5f,_voxel_size_y*0.5f,_voxel_size_z*0.5f);

	// AABB/Ray intersection
	float	t;
	int		plane;
	if (!aabb_ray_intersect(pmin,pmax,origin,dir,t,plane))
	{
		return false;
	}
	if (t>t_min) return false;

	float	tmp_min=t_min;
	bool	ret=false;
	
	for (zz=0; zz<_cluster_size_x; zz++)
	{
		vz=base_z+zz;
		for (yy=0; yy<_cluster_size_y; yy++)
		{
			vy=base_y+yy;
			for (xx=0; xx<_cluster_size_x; xx++)
			{
				vx=base_x+xx;

				voxel_type=VOXEL(vx,vy,vz);

				if ((voxel_type!=0) && (voxel_type!=0xFF))
				{
					Vec3f	pmin(vx*_voxel_size_x,vy*_voxel_size_y,vz*_voxel_size_z); pmin=pmin-Vec3f(_voxel_size_x*0.5f,_voxel_size_y*0.5f,_voxel_size_z*0.5f);
					Vec3f	pmax=pmin+Vec3f(_voxel_size_x,_voxel_size_y,_voxel_size_z); 

					if (aabb_ray_intersect(pmin,pmax,origin,dir,t,plane))
					{
						if (tmp_min>t)
						{
							tmp_min=t;
							ret_x=vx;
							ret_y=vy;
							ret_z=vz;
							if (plane==0)
							{
								if (dir.x>0.0f) ret_face=0;
								else ret_face=1;
							}
							else if (plane==1)
							{
								if (dir.y>0.0f) ret_face=3;
								else ret_face=2;
							}
							else if (plane==2)
							{
								if (dir.z>0.0f) ret_face=4;
								else ret_face=5;
							}
							ret_t=t;
							ret=true;
						}
					}
				}
			}
		}
	}

	return ret;
}

#define IR(x)	((unsigned long&)x)

#define FIND_CANDIDATE_PLANE(i)															\
	if (ray_origin[i]<pmin[i])															\
	{																					\
		Vec3f	raydir=ray_dir;															\
		/* Calculate T distances to candidate planes */									\
		if (IR(raydir[i])) max_t[i]=(pmin[i]-ray_origin[i])/raydir[i];					\
																						\
		inside=false;																	\
		coord[i]=pmin[i];																\
	}																					\
	else if (ray_origin[i]>pmax[i])														\
	{																					\
		Vec3f	raydir=ray_dir;															\
		/* Calculate T distances to candidate planes */									\
		if (IR(raydir[i])) max_t[i]=(pmax[i]-ray_origin[i])/raydir[i];					\
																						\
		inside=false;																	\
		coord[i]=pmax[i];																\
	}

#define COMPUTE_INTERSECT(i)															\
	if (i!=which_plane)																	\
	{																					\
		coord[i]=ray_origin[i]+max_t[which_plane]*ray_dir[i];							\
		if ((coord[i]<pmin[i]-EPSILON_FLOAT) ||											\
			(coord[i]>pmax[i]+EPSILON_FLOAT))	return false;							\
	}

bool VoxelWorld::aabb_ray_intersect(const Vec3f& pmin,const Vec3f& pmax,const Vec3f& ray_origin,const Vec3f& ray_dir,float& t,int& plane)
{
	bool	inside=true;
	Vec3f	min_b=pmin;
	Vec3f	max_b=pmax;
	Vec3f	max_t(-1.0f,-1.0f,-1.0f);
	Vec3f	coord;

	FIND_CANDIDATE_PLANE(0)
	FIND_CANDIDATE_PLANE(1)
	FIND_CANDIDATE_PLANE(2)

	// Ray origin inside bounding box
	if (inside)
	{
		t=0.0f;
		plane=0xff;
		return true;
	}

	unsigned char which_plane;
	if (IR(max_t.x)&0x80000000)
	{
		// T[0]<0
		if (IR(max_t.y)&0x80000000)
		{
			// T[0]<0
			// T[1]<0
			if (IR(max_t.z)&0x80000000)
			{
				// T[0]<0
				// T[1]<0
				// T[2]<0
				return false;
			}
			else
			{
				which_plane=2;
			}
		}
		else if (IR(max_t.z)&0x80000000)
		{
			// T[0]<0
			// T[1]>0
			// T[2]<0
			which_plane=1;
		}
		else
		{
			// T[0]<0
			// T[1]>0
			// T[2]>0
			if (IR(max_t.z)>IR(max_t.y))
			{
				which_plane=2;
			}
			else
			{
				which_plane=1;
			}
		}
	}
	else
	{
		// T[0]>0
		if (IR(max_t.y)&0x80000000)
		{
			// T[0]>0
			// T[1]<0
			if(IR(max_t.z)&0x80000000)
			{
				// T[0]>0
				// T[1]<0
				// T[2]<0
				which_plane=0;
			}
			else
			{
				// T[0]>0
				// T[1]<0
				// T[2]>0
				if (IR(max_t.z)>IR(max_t.x))
				{
					which_plane=2;
				}
				else
				{
					which_plane=0;
				}
			}
		}
		else if (IR(max_t.z)&0x80000000)
		{
			// T[0]>0
			// T[1]>0
			// T[2]<0
			if (IR(max_t.y)>IR(max_t.x))
			{
				which_plane=1;
			}
			else
			{
				which_plane=0;
			}
		}
		else
		{
			// T[0]>0
			// T[1]>0
			// T[2]>0
			which_plane=0;
			if (IR(max_t.y)>IR(max_t[which_plane]))	which_plane=1;
			if (IR(max_t.z)>IR(max_t[which_plane]))	which_plane=2;
		}
	}

	COMPUTE_INTERSECT(0)
	COMPUTE_INTERSECT(1)
	COMPUTE_INTERSECT(2)

	t=max_t[which_plane];
	plane=which_plane;

	return true;	
}

#undef VOXEL
#undef VOXEL_PTR
#undef VOXEL_COLOR
#undef VOXEL_COLOR_PTR
