
#include "Defs.h"

#include "Gfx.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "Keyboard.h"
#include "Timers.h"
#include "Sound.h"
#include "Utils.h"
#include "Screen.h"
#include "ParticleSystem.h"
#include "Mesh.h"
#include "Camera.h"
#include "Light.h"
#include "Sprite3d.h"
#include "Sprite.h"
#include "Font.h"
#include "lua/LUAMain.h"
#include "XML.h"

#include "lua/LUAEnv.h"

bool init_gfx(HINSTANCE hInst)
{
	new Gfx;

	GFX->set_window_mode(true);
	if (!GFX->init(hInst)) return false;
	
	new TextureManager();
	new BlitManager();
	new BlitManager3d();

	return true;
}

bool init_input(HINSTANCE hInst)
{
	new Keyboard;
	
	if (!KEY->init(hInst)) return false;

	return true;
}

bool init_sound()
{
	new Sound;

	if (!SOUND->init()) 
	{
		ERROR_MSG("Couldn't start sound... Swarm will be silent...");
	}

	return true;
}

bool init(HINSTANCE hInst)
{
	if (!init_gfx(hInst)) return false;
	if (!init_input(hInst)) return false;
	if (!init_sound()) return false;

	new Timers;

	return true;
}

void shutdown()
{
	delete SOUND;
	delete TIMERS;
	delete KEY;
	delete BLIT;
	delete TEXTURES;
	delete GFX;
}

class LUASprite : 
	public Sprite
{
public:
	LUASprite(int img,float pos_x,float pos_y);
	virtual ~LUASprite();

	virtual void	anim(const float t);

protected:
	LUAEnv* _master_env;
	LUAEnv* _thread_env;
};

LuaGlue lua_get_pos(lua_State* L)
{
	LUASprite* obj=lua_get<LUASprite>(L,0);
	if (!obj) return 0;

	lua_push_vec3(L,obj->_pos);
	return 1;
}

LuaGlue lua_set_pos(lua_State* L)
{
	LUASprite* obj=lua_get<LUASprite>(L,0);
	if (!obj) return 0;
	obj->_pos=lua_get_vec3(L,1);

	return 0;
}

LuaGlue lua_get_time(lua_State* L)
{
	lua_push_number(L,(float)timeGetTime());
	return 1;
}

LUASprite::LUASprite(int img,float pos_x,float pos_y) : Sprite(NULL,pos_x,pos_y,0.5f)
{
	_pos.set(pos_x,pos_y,0.5f);
	add_animation_frame(img);

	_master_env=new LUAEnv;
	_master_env->add_script("test.lua");
	_master_env->add_function("get_pos",lua_get_pos);
	_master_env->add_function("set_pos",lua_set_pos);
	_master_env->add_function("get_time",lua_get_time);

	_thread_env=_master_env->build_thread();

	LUAArray param;
	param.push_reference(this);

	_thread_env->resume("run",&param);
}

LUASprite::~LUASprite()
{

}

void LUASprite::anim(float t)
{
	_thread_env->resume();
}

class TestScreen : public Screen
{
public:
	TestScreen()
	{
		_tex_id=TEXTURES->load("ant","ant01.dds");

		_ps=new ParticleSystem(NULL);
		_ps->set_life(10000000);
		_ps->set_rate(100,100);
		_ps->set_image(_tex_id);
		_ps->set_particle_life(1000,100);
		_ps->set_particle_speed(0.25f,0.1f);
		_ps->set_particle_size(10.0f,50.0f);
		_ps->set_particle_color_start(1.0f,1.0f,1.0f,1.0f);
		_ps->set_particle_color(0.5f,0.0f,0.0f,0.0f,1.0f);
		_ps->set_particle_color_end(0.0f,0.0f,0.0f,0.0f);
		_ps->set_pos(512,384,0.5f);

		create_font();

		load_3ds("test.3ds",&_meshes,&_cameras,&_lights);

		_obj_rot.rotation_quaternion(Vec3f(0.0f,1.0f,0.0f),0.0f);

		if (_cameras.empty())
		{
			_camera=new Camera(NULL);
			_camera->_pos=Vec3f(0.0f,50.0f,-150.0f);
			_camera->_ori.rotation_quaternion(Vec3f(0.0f,1.0f,0.0f),0.0f);
		}
		else
		{
			_camera=_cameras[0];
		}

		_sprite3d=new Sprite3d("sprite3d",NULL);
		_sprite3d->set_image(_tex_id);
		_sprite3d->_pos.set(1.0f,0.0f,-1.0f);
		_sprite3d->set_size(Vec2f(0.5f,0.5f));

		_shadow=new Sprite3d("sprite3d",NULL);
		_shadow->set_image(_tex_id);
		_shadow->_pos.set(1.0f,0.05f,-1.0f);
		_shadow->set_color(0.0f,0.0f,0.0f,0.5f);
		_shadow->set_size(Vec2f(0.5f,0.5f));
		_shadow->set_ground(true);

		_ang=0.0f;

		_lua_sprite=new LUASprite(_tex_id,100,140);
		_lua_sprite->set_size(64.0f,64.0f);
	}
	virtual ~TestScreen()
	{
		delete _ps;
		for (size_t i=0; i<_meshes.size(); i++)
		{
			delete _meshes[i];
		}
		_meshes.clear();
		for (size_t i=0; i<_cameras.size(); i++)
		{
			delete _cameras[i];
		}
		_cameras.clear();
		for (size_t i=0; i<_lights.size(); i++)
		{
			delete _lights[i];
		}
		_lights.clear();
		TEXTURES->clear_all();
	}
	virtual void anim(const float t)
	{
		if (KEY->is_pressed(DIK_ESCAPE)) _exit=true;

		_ps->anim(t);
		_lua_sprite->anim(t);
	}

	virtual void render()
	{
		GFX->begin_scene();

		// 3d tests
		GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_ambient_light(Color(0.0f,0.0f,0.0f,1.0f));

		_ang+=0.05f; if (_ang>360.0f) _ang-=360.0f;
		
		_camera->_pos.x=8*sinf(_ang);
		_camera->_pos.z=8*cosf(_ang);
		_camera->set_lookat(_camera->_pos,Vec3f(0.0f,0.0f,0.0f));

		_camera->set();
		GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));
		GFX->clear_lights();
		for (size_t i=0; i<_lights.size(); i++)
		{
			GFX->add_light(_lights[i]);
		}

		GFX->set_pass(PASS_SOLID);
		for (size_t i=0; i<_meshes.size(); i++)	_meshes[i]->draw();
		_shadow->draw();
		_sprite3d->draw();
		GFX->set_pass(PASS_ALPHA);
		for (size_t i=0; i<_meshes.size(); i++)	_meshes[i]->draw();
		_shadow->draw();
		_sprite3d->draw();

		BLIT3D->flush();

		GFX->clear_lights();

		// 2d tests
		GFX->clear(D3DCLEAR_ZBUFFER,1.0f,1.0f,1.0f,1.0f,1.0f,0);

		GFX->set_2d_render();
		GFX->set_ambient_light(Color(1.0f,1.0f,1.0f,1.0f));

		GFX->set_camera2d(512.0f,384.0f,0.0f);

		_lua_sprite->draw();

		BLIT->add(_tex_id,50.0f,50.0f,0.5f,0.0f,0.0f,1.0f,1.0f,1.0f,1.0f,0.0f,0.0f,1.0f,1.0f);
		BLIT->flush();
		
		_ps->draw();

		_font->write_text(40.0f,100.0f,0.1f,0.0f,0.0f,0.0f,1.0f,"Testing, 1, 2...",TEXT_NORMAL);
		BLIT->flush();

		GFX->end_scene();
	}

public:
	long					_tex_id;
	ParticleSystem*			_ps;
	Quaternion				_cam_rot;
	Quaternion				_obj_rot;
	std::vector<BaseMesh*>	_meshes;
	std::vector<Camera*>	_cameras;
	std::vector<Light*>		_lights;
	Camera*					_camera;
	Sprite3d*				_sprite3d;
	Sprite3d*				_shadow;
	float					_ang;
	LUASprite*				_lua_sprite;
};


INT WINAPI WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,INT)
{
	if (!init(hInst)) return -1;

	Screen* tmp=new TestScreen;
	tmp->run();

	delete tmp;

	shutdown();

    return 0;
}
