
#include "Defs.h"

#include "Gfx.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "Keyboard.h"
#include "Timers.h"
#include "Sound.h"
#include "Utils.h"
#include "Screen.h"
#include "ParticleSystem.h"
#include "Mesh.h"
#include "Camera.h"
#include "Light.h"
#include "Sprite3d.h"
#include "Sprite.h"
#include "Tilemap3d.h"
#include "ModelObj3d.h"
#include "VoxelWorld.h"
#include "Launcher.h"
#include "MouseCursor.h"
#include "Mouse.h"
#include "UIButton.h"
#include "Font.h"
#include "Sparks.h"
#include "Explosion.h"
#include "ModelManager.h"
#include "lua/LUAMain.h"
#include "XML.h"

#include "lua/LUAEnv.h"

bool init_gfx(HINSTANCE hInst)
{
	new Gfx;

	GFX->set_window_mode(options_get_window_mode());
	GFX->set_resolution(options_get_res_x(),options_get_res_y(),options_get_bitdepth(),options_get_refresh_rate());
	if (!GFX->init(hInst)) return false;
	
	new TextureManager();
	new BlitManager();
	new BlitManager3d();
	new ModelManager;

	return true;
}

bool init_input(HINSTANCE hInst)
{
	new Keyboard;
	
	if (!KEY->init(hInst)) return false;

	new Mouse;
	
	if (!MOUSE->init()) return false;

	return true;
}

bool init_sound()
{
	new Sound;

	if (options_get_sound())
	{
		if (!SOUND->init()) 
		{
			ERROR_MSG("Couldn't start sound... Game will be silent...");
		}
	}

	return true;
}

bool init(HINSTANCE hInst)
{
	if (!init_gfx(hInst)) return false;
	if (!init_input(hInst)) return false;
	if (!init_sound()) return false;

	new Timers;

	return true;
}

void shutdown()
{
	delete SOUND;
	delete TIMERS;
	delete KEY;
	delete BLIT;
	delete TEXTURES;
	delete GFX;
}


class TestScreen : public Screen
{
public:
	TestScreen()
	{
		create_font();

		_camera=new Camera(NULL);
		_camera->_pos=Vec3f(0.0f,50.0f,-50.0f);
		_camera->_ori.rotation_quaternion(Vec3f(0.0f,1.0f,0.0f),0.0f);

		_camera_ang=0.0f;
		_ship_ang=0.0f;

		int voxels_texture=TEXTURES->find_by_name("voxel_materials");
		int mouse_cursor=TEXTURES->find_by_name("mouse_cursor");
		int exit_button=TEXTURES->find_by_name("exit_button");

		_mouse_cursor=new MouseCursor("mouse_cursor");
		_mouse_cursor->_pos.z=0.1f;

		_exit_button=new UIButton("exit_button",Vec3f(40.0f,GFX->get_virtual_resolution().y-80.0f,0.2f),"Exit the Application","exit_app");
		
		_voxels=new VoxelWorld;
		_voxels->create(128,128,128,16,16,16,0.5f,0.5f,0.5f);
		_voxels->set_material_texture(voxels_texture,8,8);
		_voxels->set_material_alpha_enable(false);

		for (int i=0; i<2; i++)
		{
			_voxels->set_uv_scale(i,Vec3f(1.0f,1.0f,1.0f));
			_voxels->set_color(i,Color(1.0f,1.0f,1.0f,1.0f));
			_voxels->set_emissive_color(i,Color(0.0f,0.0f,0.0f,1.0f));
		}

		// Ground plane
		for (int z=0; z<128; z++)
		{
			for (int x=0; x<128; x++)
			{
				_voxels->set(x,0,z,1);
			}
		}

		// Cube
		for (int z=0; z<16; z++)
		{
			for (int y=0; y<16; y++)
			{
				for (int x=0; x<16; x++)
				{
					_voxels->set(x+48,y+16,z+48,2);
				}
			}
		}

		// Broken column
		for (int z=0; z<8; z++)
		{
			for (int y=0; y<64; y++)
			{
				for (int x=0; x<8; x++)
				{
					if ((rand()%100)<70) _voxels->set(x+16,y,z+16,3);
				}
			}
		}

		_voxels->rebuild();

		_sparks=new Sparks;

		_light=new Light(NULL);
		_light->set_type(LIGHT_POINT);
		_light->_pos.set(0.0f,20.0f,20.0f);
		_light->set_range(0.0f,200.0f);
		_light->_ori.rotation_quaternion(Vec3f(1.0f,0.0f,0.0f),30.0f);
		_light->set_color(1.0f,1.0f,1.0f,1.0f);
	}
	virtual ~TestScreen()
	{
		delete _mouse_cursor;
		delete _voxels;
		delete _light;
		delete _camera;
		for (size_t i=0; i<_explosions.size(); i++) delete _explosions[i];
		_explosions.clear();
		TEXTURES->clear_all();
	}
	virtual void anim(const float t)
	{
		if (KEY->is_pressed(DIK_ESCAPE)) _exit=true;

		_sparks->anim(t);
		_mouse_cursor->anim(t);
		_exit_button->anim(t);
		for (size_t i=0; i<_explosions.size(); i++) _explosions[i]->anim(t);

		if (MOUSE->button_clicked(0))
		{
			if (_exit_button->is_inside()) _exit=true;
			else
			{
				_sparks->new_explosion(Vec3f((float)(rand()%64),1.0f,(float)(rand()%64)),0.00005f,100,2000.0f,0xFFFFFF00,0xFFFF0000);
			}
		}
		if (MOUSE->button_clicked(1))
		{
			Vec3f p((float)(rand()%64),1.0f,(float)(rand()%64));
			_explosions.push_back(new Explosion(p,Vec3f(0.0f,1.0f,0.0f),2.0f,true,p,NULL));
		}

		_camera_ang+=0.25f; if (_camera_ang>360.0f) _camera_ang-=360.0f;
		_ship_ang+=1.0f; if (_ship_ang>360.0f) _ship_ang-=360.0f;
	}

	virtual void render()
	{
		GFX->begin_scene();

		// 3d tests
		GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_ambient_light(Color(0.3f,0.3f,0.3f,1.0f));
		
		float dist=50.0f;

		_camera->_pos.x=32+(dist*sinf(deg_to_rad(_camera_ang)));
		_camera->_pos.y=16;
		_camera->_pos.z=32+(dist*cosf(deg_to_rad(_camera_ang)));
		_camera->set_lookat(_camera->_pos,Vec3f(32.0f,8.0,32.0f));

		_camera->set();
		GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));
		GFX->clear_lights();
		GFX->add_light(_light);

		Vec3f		ship_pos(32.0f+16.0f*sinf(deg_to_rad(_ship_ang)),10.0f,32.0f+16.0f*cosf(deg_to_rad(_ship_ang)));
		Quaternion	ship_ori1; ship_ori1.rotation_quaternion(0.0f,1.0f,0.0f,_ship_ang+90.0f);
		Quaternion	ship_ori2; ship_ori2.rotation_quaternion(0.0f,0.0f,1.0f,-20.0f);
		Quaternion	ship_ori=ship_ori1*ship_ori2;
		
		GFX->set_pass(PASS_SOLID);
		_voxels->render(_camera);
		_sparks->draw();
		for (size_t j=0; j<_explosions.size(); j++) _explosions[j]->draw();
		MODEL_MANAGER->draw(MODEL_MANAGER->find_by_name("lander"),ship_pos,ship_ori,Vec3f(1.0f,1.0f,1.0f),_camera);
		GFX->set_pass(PASS_ALPHA);
		_voxels->render(_camera);
		_sparks->draw();
		for (size_t j=0; j<_explosions.size(); j++) _explosions[j]->draw();
		MODEL_MANAGER->draw(MODEL_MANAGER->find_by_name("lander"),ship_pos,ship_ori,Vec3f(1.0f,1.0f,1.0f),_camera);

		BLIT3D->flush();

		GFX->clear_lights();

		GFX->set_2d_render();
		GFX->set_camera2d(GFX->get_virtual_resolution().x*0.5f,GFX->get_virtual_resolution().y*0.5f,0.0f);
		GFX->clear(D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_ambient_light(Color(1.0f,1.0f,1.0f,1.0f));
		
		_exit_button->draw();
		_mouse_cursor->draw();

		if (_exit_button->is_inside())
		{
			std::string tooltip=_exit_button->get_tooltip();

			_font->write_text(_mouse_cursor->_pos.x,_mouse_cursor->_pos.y,0.05f,1.0f,1.0f,1.0f,1.0f,tooltip.c_str(),TEXT_SHADOW);
		}

		BLIT->flush();

		GFX->end_scene();
	}

public:
	Camera*					_camera;
	float					_camera_ang;
	float					_ship_ang;
	VoxelWorld*				_voxels;
	MouseCursor*			_mouse_cursor;
	UIButton*				_exit_button;
	Light*					_light;
	Sparks*					_sparks;
	std::vector<Explosion*>	_explosions;
};


INT WINAPI WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,INT)
{
	options_load();
	if (options_show_launcher())
	{
		launch_launcher();
		options_save();
	}

	if (!init(hInst)) return -1;
	if (!init_input(hInst)) return -1;

	Screen::load_assets("assets.dat");

	Screen* tmp=new TestScreen;
	tmp->run();

	delete tmp;

	shutdown();

    return 0;
}
