
#include "Defs.h"

#include "Gfx.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "Keyboard.h"
#include "Timers.h"
#include "Sound.h"
#include "Utils.h"
#include "Screen.h"
#include "ParticleSystem.h"
#include "Mesh.h"
#include "Camera.h"
#include "Light.h"
#include "Sprite3d.h"
#include "Sprite.h"
#include "Tilemap3d.h"
#include "ModelObj3d.h"
#include "NavMesh.h"
#include "NavPath.h"
#include "ModelManager.h"
#include "MouseCursor.h"
#include "Mouse.h"
#include "lua/LUAMain.h"
#include "XML.h"

#include "lua/LUAEnv.h"

bool init_gfx(HINSTANCE hInst)
{
	new Gfx;

	GFX->set_window_mode(true);
	if (!GFX->init(hInst)) return false;
	
	new TextureManager();
	new BlitManager();
	new BlitManager3d();
	new ModelManager;

	return true;
}

bool init_input(HINSTANCE hInst)
{
	new Keyboard;
	if (!KEY->init(hInst)) return false;
	new Mouse;
	if (!MOUSE->init()) return false;

	return true;
}

bool init_sound()
{
	new Sound;

	if (!SOUND->init()) 
	{
		ERROR_MSG("Couldn't start sound... Swarm will be silent...");
	}

	return true;
}

bool init(HINSTANCE hInst)
{
	if (!init_gfx(hInst)) return false;
	if (!init_input(hInst)) return false;
	if (!init_sound()) return false;

	new Timers;

	return true;
}

void shutdown()
{
	delete SOUND;
	delete TIMERS;
	delete KEY;
	delete BLIT;
	delete TEXTURES;
	delete GFX;
}

class TestScreen : public Screen
{
public:
	TestScreen()
	{
		create_font();

		_camera=new Camera(NULL);
		_camera->_pos=Vec3f(0.0f,50.0f,-50.0f);
		_camera->_ori.rotation_quaternion(Vec3f(0.0f,1.0f,0.0f),0.0f);

		_ang=45.0f;

		_tilemap=Tilemap3d::load(NULL,"map.xml",false);
		
		float px=(_tilemap->get_map_size_x()*_tilemap->get_tile_size_x())*0.5f;
		float py=(_tilemap->get_map_size_y()*_tilemap->get_tile_size_y())*0.5f;
		_tilemap->_pos.set(-px,0.0f,-py);

		_lander=new ModelObj3d("lander","lander");
		_lander->set_mesh(MODEL_MANAGER->find_by_name("lander"));
		_lander->_pos.set(5.0f,0.1f,7.0f);

		_light=new Light(NULL);
		_light->set_type(LIGHT_POINT);
		_light->_pos.set(0.0f,20.0f,20.0f);
		_light->set_range(30.0f,35.0f);
		_light->_ori.rotation_quaternion(Vec3f(1.0f,0.0f,0.0f),30.0f);
		_light->set_color(1.0f,1.0f,1.0f,1.0f);

		_navmesh=new NavMesh;
		_navmesh->add_tilemap3d(_tilemap);
		_navmesh->add_modelobj3d(_lander);
		_navmesh->init();

		_mouse_cursor=new MouseCursor("mouse_cursor");
		_click_idx=0;
		_no_compute=true;
		_path=NULL;
	}
	virtual ~TestScreen()
	{
		TEXTURES->clear_all();
	}
	virtual void anim(const float t)
	{
		if (KEY->is_pressed(DIK_ESCAPE)) _exit=true;

		_mouse_cursor->anim(t);

		if (MOUSE->button_clicked(0))
		{
			Vec3f o,d;
			_camera->extract_ray((unsigned long)_mouse_cursor->_pos.x,(unsigned long)_mouse_cursor->_pos.y,o,d);
			float t=-o.y/d.y;
			Vec3f p(o.x+d.x*t,0.0f,o.z+d.z*t);
			_click_pos[_click_idx]=p;
			if (_click_idx==0) _click_idx=1;
			else _no_compute=false;
/*			_click_idx=(_click_idx+1)%2;
			if (_click_idx==0) _no_compute=false;*/

			if (!_no_compute)
			{
				// Compute path
				_path=_navmesh->get_path(_click_pos[0],_click_pos[1],_path);
			}
		}
	}

	virtual void render()
	{
		GFX->begin_scene();

		// 3d tests
		GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_ambient_light(Color(0.1f,0.1f,0.1f,1.0f));

//		_ang+=0.005f; if (_ang>360.0f) _ang-=360.0f;
		
		float dist=20.0f;

		_camera->_pos.x=dist*sinf(_ang);
		_camera->_pos.y=dist;
		_camera->_pos.z=dist*cosf(_ang);
		_camera->set_lookat(_camera->_pos,Vec3f(0.0f,0.0f,0.0f));

		_camera->set();
		GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));
		GFX->clear_lights();
		GFX->add_light(_light);

		GFX->set_pass(PASS_SOLID);
		_tilemap->draw();
		_lander->draw();

		GFX->set_pass(PASS_ALPHA);
		_navmesh->draw();
		if (_path) _path->draw();

		BLIT3D->add_triangle(_click_pos[0],_click_pos[0]+Vec3f(-1.0f,2.0f,0.0f),_click_pos[0]+Vec3f(1.0f,2.0f,0.0f),Color(1.0f,0.0f,0.0f,1.0f));
		BLIT3D->add_triangle(_click_pos[0],_click_pos[0]+Vec3f(0.0f,2.0f,-1.0f),_click_pos[0]+Vec3f(0.0f,2.0f,1.0f),Color(1.0f,0.0f,0.0f,1.0f));
		BLIT3D->add_triangle(_click_pos[1],_click_pos[1]+Vec3f(-1.0f,2.0f,0.0f),_click_pos[1]+Vec3f(1.0f,2.0f,0.0f),Color(0.0f,0.0f,1.0f,1.0f));
		BLIT3D->add_triangle(_click_pos[1],_click_pos[1]+Vec3f(0.0f,2.0f,-1.0f),_click_pos[1]+Vec3f(0.0f,2.0f,1.0f),Color(0.0f,0.0f,1.0f,1.0f));

		BLIT3D->flush();

		GFX->clear_lights();

		GFX->set_2d_render();
		GFX->set_camera2d(GFX->get_virtual_resolution().x,GFX->get_virtual_resolution().y,0.0f);
		GFX->set_ambient_light(Color(1.0f,1.0f,1.0f,1.0f));
		GFX->clear(D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);

		_mouse_cursor->draw();

		BLIT->flush();

		GFX->end_scene();
	}

public:
	Camera*					_camera;
	float					_ang;
	Tilemap3d*				_tilemap;
	Light*					_light;
	NavMesh*				_navmesh;
	NavPath*				_path;
	ModelObj3d*				_lander;
	MouseCursor*			_mouse_cursor;
	Vec3f					_click_pos[2];
	int						_click_idx;
	bool					_no_compute;
};

INT WINAPI WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,INT)
{
	if (!init(hInst)) return -1;
	if (!init_input(hInst)) return -1;

	Screen::load_assets("assets.dat");

	Screen* tmp=new TestScreen;
	tmp->run();

	delete tmp;

	shutdown();

    return 0;
}
