
#include "Defs.h"

#include "Gfx.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "Keyboard.h"
#include "Timers.h"
#include "Sound.h"
#include "Utils.h"
#include "Screen.h"
#include "ParticleSystem.h"
#include "Mesh.h"
#include "Camera.h"
#include "Light.h"
#include "Sprite3d.h"
#include "Sprite.h"
#include "Tilemap3d.h"
#include "ModelObj3d.h"
#include "ModelManager.h"
#include "MouseCursor.h"
#include "Mouse.h"
#include "FSEffect.h"
#include "lua/LUAMain.h"
#include "XML.h"

#include "lua/LUAEnv.h"

bool init_gfx(HINSTANCE hInst)
{
	new Gfx;

	GFX->set_window_mode(true);
	if (!GFX->init(hInst)) return false;
	
	new TextureManager();
	new BlitManager();
	new BlitManager3d();
	new ModelManager;

	return true;
}

bool init_input(HINSTANCE hInst)
{
	new Keyboard;
	if (!KEY->init(hInst)) return false;
	new Mouse;
	if (!MOUSE->init()) return false;

	return true;
}

bool init_sound()
{
	new Sound;

	if (!SOUND->init()) 
	{
		ERROR_MSG("Couldn't start sound... Swarm will be silent...");
	}

	return true;
}

bool init(HINSTANCE hInst)
{
	if (!init_gfx(hInst)) return false;
	if (!init_input(hInst)) return false;
	if (!init_sound()) return false;

	new Timers;

	return true;
}

void shutdown()
{
	delete SOUND;
	delete TIMERS;
	delete KEY;
	delete BLIT;
	delete TEXTURES;
	delete GFX;
}

class TestScreen : public Screen
{
public:
	TestScreen()
	{
		create_font();

		_camera=new Camera(NULL);
		_camera->_pos=Vec3f(0.0f,50.0f,-50.0f);
		_camera->_ori.rotation_quaternion(Vec3f(0.0f,1.0f,0.0f),0.0f);

		_ang=45.0f;

		_tilemap=Tilemap3d::load(NULL,"map.xml",false);
		
		float px=(_tilemap->get_map_size_x()*_tilemap->get_tile_size_x())*0.5f;
		float py=(_tilemap->get_map_size_y()*_tilemap->get_tile_size_y())*0.5f;
		_tilemap->_pos.set(-px,0.0f,-py);

		_lander=new ModelObj3d("lander","lander");
		_lander->set_mesh(MODEL_MANAGER->find_by_name("lander"));
		_lander->_pos.set(5.0f,0.1f,7.0f);

		_light=new Light(NULL);
		_light->set_type(LIGHT_POINT);
		_light->_pos.set(0.0f,20.0f,20.0f);
		_light->set_range(30.0f,35.0f);
		_light->_ori.rotation_quaternion(Vec3f(1.0f,0.0f,0.0f),30.0f);
		_light->set_color(1.0f,1.0f,1.0f,1.0f);

		_mouse_cursor=new MouseCursor("mouse_cursor");

		_effect=new FSEffect("main",2,false,0.25f);
	}
	virtual ~TestScreen()
	{
		delete _effect;
		_effect=NULL;

		TEXTURES->clear_all();
	}
	virtual void anim(const float t)
	{
		if (KEY->is_pressed(DIK_ESCAPE)) _exit=true;

		_mouse_cursor->anim(t);
	}

	virtual void render()
	{
		GFX->begin_scene();

		// 3d tests
		GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_ambient_light(Color(0.1f,0.1f,0.1f,1.0f));

//		_ang+=0.005f; if (_ang>360.0f) _ang-=360.0f;
		
		float dist=20.0f;

		_camera->_pos.x=dist*sinf(_ang);
		_camera->_pos.y=dist;
		_camera->_pos.z=dist*cosf(_ang);
		_camera->set_lookat(_camera->_pos,Vec3f(0.0f,0.0f,0.0f));

		_camera->set();
		GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));
		GFX->clear_lights();
		GFX->add_light(_light);

		GFX->set_pass(PASS_SOLID_NO_EMISSIVE);

		_tilemap->draw();
		_lander->draw();

		_effect->begin();
		GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0.0f,0.0f,0.0f,0.0f,1.0f,0);
		GFX->set_pass(PASS_SOLID_ONLY_EMISSIVE);

		_tilemap->draw();
		_lander->draw();

		// Do glow effect
		Vec2f	r=GFX->get_real_resolution();

		float inv_x=1.0f/(r.x*0.5f);
		float inv_y=1.0f/(r.y*0.5f);
		
		_effect->set_constant("v_offset0",Vec4f(0.0f,1.3366f*inv_y,0.0f,0.0f));
		_effect->set_constant("v_offset1",Vec4f(0.0f,3.4295f*inv_y,0.0f,0.0f));
		_effect->set_constant("v_offset2",Vec4f(0.0f,5.4264f*inv_y,0.0f,0.0f));
		_effect->set_constant("p_offset0",Vec4f(0.0f,7.4359f*inv_y,0.0f,0.0f));
		_effect->set_constant("p_offset1",Vec4f(0.0f,9.4436f*inv_y,0.0f,0.0f));
		_effect->set_constant("p_offset2",Vec4f(0.0f,11.4401f*inv_y,0.0f,0.0f));
		_effect->set_constant("glow_str",Vec4f(1.5f,0.0f,0.0f,0.0f));
		_effect->run("fx_blur");

		_effect->set_constant("v_offset0",Vec4f(1.3366f*inv_y,0.0f,0.0f,0.0f));
		_effect->set_constant("v_offset1",Vec4f(3.4295f*inv_y,0.0f,0.0f,0.0f));
		_effect->set_constant("v_offset2",Vec4f(5.4264f*inv_y,0.0f,0.0f,0.0f));
		_effect->set_constant("p_offset0",Vec4f(7.4359f*inv_y,0.0f,0.0f,0.0f));
		_effect->set_constant("p_offset1",Vec4f(9.4436f*inv_y,0.0f,0.0f,0.0f));
		_effect->set_constant("p_offset2",Vec4f(11.4401f*inv_y,0.0f,0.0f,0.0f));
		_effect->set_constant("glow_str",Vec4f(1.5f,0.0f,0.0f,0.0f));
		_effect->run("fx_blur");

		_effect->end(D3DBLEND_ONE,D3DBLEND_ONE);

		GFX->clear_lights();

		GFX->set_2d_render();
		GFX->set_camera2d(GFX->get_virtual_resolution().x,GFX->get_virtual_resolution().y,0.0f);
		GFX->set_ambient_light(Color(1.0f,1.0f,1.0f,1.0f));
		GFX->clear(D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_pass(PASS_SOLID);

		_mouse_cursor->draw();

		BLIT->flush();

		GFX->end_scene();
	}

public:
	Camera*					_camera;
	float					_ang;
	Tilemap3d*				_tilemap;
	Light*					_light;
	ModelObj3d*				_lander;
	MouseCursor*			_mouse_cursor;
	FSEffect*				_effect;
};

INT WINAPI WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,INT)
{
	if (!init(hInst)) return -1;
	if (!init_input(hInst)) return -1;

	Screen::load_assets("assets.dat");

	Screen* tmp=new TestScreen; 
	tmp->run(); 
	 
	delete tmp;

	shutdown();

    return 0;
}
